static int setup_icon_w = 128;
static int setup_icon_h = 128;

static const unsigned int setup_icon_data[] = {
    0xce9b35ff, 0xca982eff, 0xcb972bff, 0xd4a648ff, 0xd7ab5bff, 0xd7ac5cff, 
    0xd7ac5eff, 0xd5a448ff, 0xc18a20ff, 0xb37a0cff, 0xb37a0cff, 0xb37606ff, 
    0xb97609ff, 0xdca946ff, 0xe9c16dff, 0xedc97bff, 0xf3d78fff, 0xf7db92ff, 
    0xeac068ff, 0xe0ae4eff, 0xd99a27ff, 0xd99617ff, 0xdea12bff, 0xe1a735ff, 
    0xe6b142ff, 0xe8b444ff, 0xeab647ff, 0xebb746ff, 0xebb849ff, 0xecb84aff, 
    0xecb94aff, 0xebb948ff, 0xeab843ff, 0xeab946ff, 0xecb63eff, 0xecb63cff, 
    0xecbe5dff, 0xecc571ff, 0xedc776ff, 0xeac060ff, 0xebc268ff, 0xebc679ff, 
    0xecc882ff, 0xe9c579ff, 0xe7bf68ff, 0xebc26eff, 0xebc374ff, 0xebc579ff, 
    0xeac376ff, 0xe9bd66ff, 0xe9be64ff, 0xebc47aff, 0xecc479ff, 0xeac06dff, 
    0xeabe66ff, 0xecc26cff, 0xebc26dff, 0xedc675ff, 0xeec67bff, 0xeec77cff, 
    0xeec97eff, 0xf1d391ff, 0xf0cf89ff, 0xefcb7fff, 0xeec36aff, 0xedc36fff, 
    0xecc36dff, 0xf3d495ff, 0xf3d69dff, 0xf1d194ff, 0xedc880ff, 0xf3d295ff, 
    0xf6d69dff, 0xe8b96aff, 0xe1ab48ff, 0xe3b157ff, 0xe6bd75ff, 0xe7c485ff, 
    0xe8c585ff, 0xe9c684ff, 0xeac98bff, 0xecc990ff, 0xeac98eff, 0xe6c688ff, 
    0xe8c78bff, 0xe9c98dff, 0xe6c37eff, 0xe8c686ff, 0xebc995ff, 0xedcc95ff, 
    0xedcc93ff, 0xefd2a2ff, 0xebcd9dff, 0xe9c98fff, 0xecd1a5ff, 0xebd2abff, 
    0xe9cea2ff, 0xeacfa0ff, 0xe8cea0ff, 0xe4c28aff, 0xe1ba77ff, 0xdcb268ff, 
    0xd3a556ff, 0xc48f44ff, 0xc68f43ff, 0xe3ae5bff, 0xeabd7bff, 0xeecd99ff, 
    0xe8c892ff, 0xe2c080ff, 0xe0bd78ff, 0xe3c27fff, 0xe6c792ff, 0xddba7cff, 
    0xd7b169ff, 0xd3ac61ff, 0xd1b070ff, 0xcead6eff, 0xc5a35eff, 0xb98b33ff, 
    0xa9710eff, 0xac7417ff, 0xb38124ff, 0xb58424ff, 0xb88930ff, 0xbb8e36ff, 
    0xbc8d2fff, 0xbd8c2bff, 0xcb972dff, 0xc9982eff, 0xcd9a2eff, 0xd4a649ff, 
    0xd7ab57ff, 0xd8ab5aff, 0xd8aa59ff, 0xd8a851ff, 0xca952eff, 0xb3790bff, 
    0xb27607ff, 0xb17305ff, 0xb46d07ff, 0xd7a13dff, 0xe3b967ff, 0xe7c075ff, 
    0xeac578ff, 0xf2d388ff, 0xfae1a4ff, 0xfbe7b1ff, 0xf1d28bff, 0xdfa73fff, 
    0xda9c26ff, 0xdc9f29ff, 0xdfa32cff, 0xe1a72fff, 0xe3ab37ff, 0xe6b13dff, 
    0xe8b344ff, 0xe7b44aff, 0xe9b547ff, 0xe7b544ff, 0xe7b23eff, 0xe7b23fff, 
    0xe8b23aff, 0xe7b034ff, 0xe8b74dff, 0xecc571ff, 0xecc87dff, 0xebc269ff, 
    0xeac065ff, 0xeac372ff, 0xeac372ff, 0xe7bc62ff, 0xe8bc61ff, 0xeac06dff, 
    0xe8bc65ff, 0xe9bd64ff, 0xe8bd65ff, 0xe7b95bff, 0xe9c06bff, 0xe9c173ff, 
    0xe9be69ff, 0xe8bb65ff, 0xe7b958ff, 0xeabf67ff, 0xedc473ff, 0xeec676ff, 
    0xefc87bff, 0xefcc84ff, 0xeec77eff, 0xefcc8aff, 0xefce85ff, 0xf0cb81ff, 
    0xefc978ff, 0xf1ce85ff, 0xefc97aff, 0xf1d18aff, 0xf2d28eff, 0xf2d18bff, 
    0xf4d694ff, 0xf9e0aaff, 0xfbe3b2ff, 0xf2cd8fff, 0xe4b053ff, 0xe1af4fff, 
    0xe4b864ff, 0xe7c17dff, 0xebc98fff, 0xebc991ff, 0xe9c788ff, 0xeaca8bff, 
    0xeaca8fff, 0xe7c589ff, 0xe6c486ff, 0xe6c481ff, 0xe5c27cff, 0xe9c785ff, 
    0xeac98dff, 0xeacc8fff, 0xefd099ff, 0xeccd98ff, 0xe8c78bff, 0xeaca8fff, 
    0xebd1a3ff, 0xead0a7ff, 0xe9cc9eff, 0xe9cea1ff, 0xe8cfa5ff, 0xe7c896ff, 
    0xe3c189ff, 0xdfb97fff, 0xd5a75fff, 0xc1893dff, 0xa0652bff, 0xc79044ff, 
    0xe8b667ff, 0xe9c080ff, 0xe7c385ff, 0xe2bf7eff, 0xdeba72ff, 0xdebb75ff, 
    0xe0c28bff, 0xdcbb7eff, 0xd9b777ff, 0xd3b16cff, 0xd0af6fff, 0xcead70ff, 
    0xc5a35eff, 0xba913dff, 0xab7413ff, 0xa5680aff, 0xac761bff, 0xaf7d1eff, 
    0xb27f21ff, 0xb17f1fff, 0xb8882cff, 0xbb8c30ff, 0xce9d36ff, 0xcb9a31ff, 
    0xce992cff, 0xd8a951ff, 0xd9ad5bff, 0xd8ac5dff, 0xd5a855ff, 0xd6aa53ff, 
    0xd5a447ff, 0xbf861dff, 0xae7107ff, 0xad6c05ff, 0xb36905ff, 0xce922bff, 
    0xdba949ff, 0xdeb155ff, 0xe4b862ff, 0xe9c277ff, 0xf1d493ff, 0xf7e1aaff, 
    0xfceab4ff, 0xfae2a4ff, 0xedc772ff, 0xdea537ff, 0xda9b21ff, 0xdc9e21ff, 
    0xdea227ff, 0xe0a529ff, 0xe3aa32ff, 0xe3ad38ff, 0xe3ab37ff, 0xe4ab34ff, 
    0xe3a92fff, 0xe3a931ff, 0xe2ab35ff, 0xe2a92fff, 0xe7b750ff, 0xebc470ff, 
    0xeac673ff, 0xebc472ff, 0xeac06cff, 0xe8bb5eff, 0xe6b857ff, 0xe6b854ff, 
    0xe7b857ff, 0xe9bf69ff, 0xe7bb63ff, 0xe7ba5eff, 0xe7ba5dff, 0xe7ba5bff, 
    0xe8bd65ff, 0xeac379ff, 0xe7bc66ff, 0xe5b759ff, 0xe7ba5cff, 0xecc473ff, 
    0xecc26dff, 0xecc36fff, 0xeec97dff, 0xefcd88ff, 0xefca81ff, 0xeec87bff, 
    0xefc875ff, 0xefc977ff, 0xf1ca7aff, 0xf6db9fff, 0xf6d89aff, 0xf5d795ff, 
    0xf9dfa2ff, 0xf9dfa0ff, 0xf8dd9bff, 0xfce6b2ff, 0xfce7b7ff, 0xf7d9a1ff, 
    0xe9bd6dff, 0xe4b358ff, 0xe7bd6fff, 0xeac483ff, 0xe9c486ff, 0xe9c486ff, 
    0xe7c481ff, 0xe8c681ff, 0xe8c98aff, 0xe7c786ff, 0xe4c079ff, 0xe6c380ff, 
    0xe9c98aff, 0xedcc92ff, 0xe8c88aff, 0xe9c989ff, 0xebcb91ff, 0xeacb91ff, 
    0xe9ca8fff, 0xe9c890ff, 0xe9cb96ff, 0xe8c994ff, 0xe9ca9aff, 0xe9cda1ff, 
    0xe7ca9eff, 0xe3c088ff, 0xdeb671ff, 0xddb475ff, 0xd6ab68ff, 0xc8954cff, 
    0xa76b30ff, 0xac7533ff, 0xe4b15eff, 0xeabf7aff, 0xebc58dff, 0xe8c692ff, 
    0xe3c186ff, 0xdeb977ff, 0xdab877ff, 0xdebc80ff, 0xdcba7fff, 0xd2b06cff, 
    0xcead6cff, 0xcbaa6eff, 0xc09c53ff, 0xba9140ff, 0xae7f28ff, 0xa36d14ff, 
    0xa36a11ff, 0xa97415ff, 0xac7719ff, 0xaa7614ff, 0xb07f20ff, 0xb38220ff, 
    0xcd9c37ff, 0xce9d39ff, 0xd3a242ff, 0xd7aa57ff, 0xd7ab5bff, 0xd5aa57ff, 
    0xd7ab58ff, 0xd9ad59ff, 0xd7a751ff, 0xcd9a35ff, 0xb07414ff, 0xa56004ff, 
    0xb26904ff, 0xc17e11ff, 0xd5a13dff, 0xdaa748ff, 0xddad51ff, 0xe1b65fff, 
    0xe9c173ff, 0xeecc83ff, 0xf4da96ff, 0xfae4a5ff, 0xfce7a5ff, 0xf8df97ff, 
    0xe6ba58ff, 0xd8981cff, 0xd99a19ff, 0xdc9d20ff, 0xdea025ff, 0xdfa223ff, 
    0xe0a328ff, 0xe1a52dff, 0xe0a42bff, 0xdea123ff, 0xdea122ff, 0xdfa62bff, 
    0xe8bb58ff, 0xecc57aff, 0xeac372ff, 0xeac274ff, 0xeac275ff, 0xe9bd66ff, 
    0xe8bc62ff, 0xe7bb60ff, 0xe6b95cff, 0xe6bb60ff, 0xe7ba60ff, 0xe7ba5dff, 
    0xe7b95bff, 0xe7ba63ff, 0xe7ba62ff, 0xe8bb66ff, 0xe6b656ff, 0xe5b654ff, 
    0xe8be66ff, 0xecc475ff, 0xebc36dff, 0xedc97eff, 0xeeca81ff, 0xedc87bff, 
    0xf0cf89ff, 0xefcc7eff, 0xeec876ff, 0xefc97bff, 0xf4d48fff, 0xf6db9fff, 
    0xf6d893ff, 0xfbe2a8ff, 0xfce7b1ff, 0xfceab3ff, 0xfce9b2ff, 0xfce8b9ff, 
    0xfce5aeff, 0xfbe3abff, 0xefc981ff, 0xe3b258ff, 0xe6ba69ff, 0xe8bf75ff, 
    0xe6bb6cff, 0xe7c079ff, 0xe8c37eff, 0xe6be72ff, 0xe4bd73ff, 0xe5bf76ff, 
    0xe8c37cff, 0xeecf93ff, 0xf0d4a0ff, 0xeed19cff, 0xebcd95ff, 0xebcc92ff, 
    0xeed39fff, 0xedd4a4ff, 0xe7c78cff, 0xe9cb95ff, 0xe9ce9cff, 0xe8cb97ff, 
    0xe9ca99ff, 0xe9cca0ff, 0xe5c492ff, 0xdfba7aff, 0xddb46dff, 0xdbaf67ff, 
    0xd6aa60ff, 0xcd9b4eff, 0xb87e39ff, 0xa46727ff, 0xd99d42ff, 0xe7b769ff, 
    0xe8be7dff, 0xe6bd7fff, 0xe4bf86ff, 0xe3c189ff, 0xdcb879ff, 0xdcbb7fff, 
    0xd8b77cff, 0xd4b273ff, 0xcaa866ff, 0xc5a561ff, 0xbe9848ff, 0xbb9142ff, 
    0xab7f31ff, 0x9b722aff, 0x9b691aff, 0x9d640aff, 0x9f6a0dff, 0xa67416ff, 
    0xaa7617ff, 0xaf7c1eff, 0xd3a54fff, 0xd4a74eff, 0xd6ab57ff, 0xd8ac5bff, 
    0xd5aa57ff, 0xd5aa56ff, 0xd8ad5cff, 0xd8ae5aff, 0xd6a851ff, 0xd3a142ff, 
    0xba8420ff, 0x9d5b04ff, 0xac6404ff, 0xbd7d12ff, 0xd09d3fff, 0xd5a347ff, 
    0xd6a541ff, 0xddad51ff, 0xe3b967ff, 0xe8bf6dff, 0xecc776ff, 0xf1d58eff, 
    0xf6de9cff, 0xfae6acff, 0xfae4a3ff, 0xe8bf63ff, 0xd99c26ff, 0xd79718ff, 
    0xd99818ff, 0xdb9c1bff, 0xdc9e1dff, 0xdc9e1eff, 0xdc9e1dff, 0xdb9c1aff, 
    0xdb9b16ff, 0xdea327ff, 0xe6b750ff, 0xebc274ff, 0xeac376ff, 0xe7bf6cff, 
    0xe8bd65ff, 0xe9be69ff, 0xe7ba61ff, 0xe5b552ff, 0xe6b85bff, 0xe7ba5eff, 
    0xe7ba5fff, 0xe7bb60ff, 0xe8bc63ff, 0xe7bc67ff, 0xe5b85eff, 0xe4b351ff, 
    0xe5b658ff, 0xe4b24eff, 0xe7b95eff, 0xeac06aff, 0xebc270ff, 0xecc77aff, 
    0xecc778ff, 0xedc776ff, 0xf0d18bff, 0xeec97cff, 0xeec775ff, 0xf0ce82ff, 
    0xf7dda0ff, 0xf5d894ff, 0xf7dd98ff, 0xfce6acff, 0xfdeab6ff, 0xfdecbbff, 
    0xfdebbbff, 0xfdefc4ff, 0xfdecbaff, 0xfdeab3ff, 0xf7dda0ff, 0xe7b860ff, 
    0xe3b459ff, 0xe4b55dff, 0xe4b660ff, 0xe5b869ff, 0xe5bc6fff, 0xe2b561ff, 
    0xe4b967ff, 0xe8bf71ff, 0xefcf8dff, 0xf3daa4ff, 0xf2dcabff, 0xf1d8a9ff, 
    0xeed5a2ff, 0xefd29cff, 0xefd8acff, 0xecd3a4ff, 0xe6c485ff, 0xe9cb93ff, 
    0xebd0a0ff, 0xe9cf9dff, 0xeacfa2ff, 0xe9cea2ff, 0xe6c492ff, 0xe3c08aff, 
    0xe2bc85ff, 0xddb677ff, 0xd8aa63ff, 0xd19b49ff, 0xc58939ff, 0xbb7a32ff, 
    0xcf8920ff, 0xe1a94bff, 0xe7ba6dff, 0xeac17fff, 0xe6bf80ff, 0xdfbc81ff, 
    0xd9b573ff, 0xd9b776ff, 0xdab980ff, 0xd0ac68ff, 0xc69e51ff, 0xc29d4fff, 
    0xbd9645ff, 0xb58b3cff, 0xa57930ff, 0x9d7840ff, 0x9c763eff, 0x96661cff, 
    0x98630fff, 0xa06b12ff, 0xa6751bff, 0xa8781bff, 0xd8ab59ff, 0xd7aa55ff, 
    0xd5a952ff, 0xd7ac55ff, 0xd9b062ff, 0xd6ad5cff, 0xd7ac5bff, 0xd8ac5aff, 
    0xd6a952ff, 0xd2a245ff, 0xca9533ff, 0xaf7112ff, 0xa96004ff, 0xba7a11ff, 
    0xcd9835ff, 0xd3a246ff, 0xd6a345ff, 0xdbaa51ff, 0xddb057ff, 0xe0b459ff, 
    0xe5bb63ff, 0xeac476ff, 0xefce83ff, 0xf2d892ff, 0xf8e09fff, 0xfae39dff, 
    0xf6da91ff, 0xebc56fff, 0xdfa63aff, 0xd7971aff, 0xd79513ff, 0xd89511ff, 
    0xd79610ff, 0xd79611ff, 0xd7960fff, 0xdea326ff, 0xe4b247ff, 0xe5b85cff, 
    0xe8bc65ff, 0xe7bb61ff, 0xe5b758ff, 0xe7bc63ff, 0xe5b759ff, 0xe3b34bff, 
    0xe4b554ff, 0xe5b657ff, 0xe6b95fff, 0xe7bd68ff, 0xe6b85bff, 0xe5b658ff, 
    0xe4b65aff, 0xe1b14aff, 0xe3b252ff, 0xe2b04cff, 0xe4b451ff, 0xe9bc60ff, 
    0xebc06bff, 0xecc26dff, 0xebc370ff, 0xecc36dff, 0xefcc80ff, 0xf1d38dff, 
    0xf3d48dff, 0xf5d894ff, 0xf6da96ff, 0xf8dd9aff, 0xfbe5a8ff, 0xfce8aeff, 
    0xfcebb9ff, 0xfdeec0ff, 0xfdf0bfff, 0xfdf4c8ff, 0xfdf2c8ff, 0xfdefc0ff, 
    0xfbe6b0ff, 0xefc87cff, 0xe1b151ff, 0xe3b458ff, 0xe4b75fff, 0xe4b760ff, 
    0xe3b55dff, 0xe2b55fff, 0xe7be70ff, 0xe9c274ff, 0xedcd8bff, 0xf1d497ff, 
    0xf1d9a2ff, 0xf0d7a3ff, 0xeed199ff, 0xeecf96ff, 0xeccf9aff, 0xe8ca91ff, 
    0xebcd94ff, 0xecd29eff, 0xeacd98ff, 0xe8ca93ff, 0xe8cc98ff, 0xe8ca97ff, 
    0xe5c08aff, 0xe6c290ff, 0xe4bf8dff, 0xe0b87fff, 0xdbaf6cff, 0xd4a04fff, 
    0xd0963eff, 0xc27e2eff, 0xca8420ff, 0xdc9d33ff, 0xe4b055ff, 0xe7ba6dff, 
    0xe7be74ff, 0xe3bd7bff, 0xdab574ff, 0xdbb87aff, 0xd9b77cff, 0xcea559ff, 
    0xc29742ff, 0xbe9542ff, 0xb48831ff, 0xaa7720ff, 0xa06e1dff, 0x9e7e4fff, 
    0x9f835bff, 0x9d7e4cff, 0x976a24ff, 0x955e09ff, 0xa16f19ff, 0xaa7c26ff, 
    0xd6aa54ff, 0xd4a74dff, 0xd4a646ff, 0xd6a94cff, 0xd7ad58ff, 0xd6ab5aff, 
    0xd8ad61ff, 0xd4a954ff, 0xd4a64dff, 0xd1a24aff, 0xd0a049ff, 0xc6902bff, 
    0xac6406ff, 0xb36c08ff, 0xc89029ff, 0xcd9937ff, 0xd1a041ff, 0xd3a13eff, 
    0xd9a948ff, 0xdcae49ff, 0xe2b65dff, 0xe4ba5eff, 0xe9c064ff, 0xedcc7fff, 
    0xf1d695ff, 0xf6df9cff, 0xfae6a4ff, 0xfdebb1ff, 0xfae4aaff, 0xf0ce82ff, 
    0xdca438ff, 0xd4900dff, 0xd38c05ff, 0xd48e08ff, 0xd5920eff, 0xdca227ff, 
    0xe2af3fff, 0xe3b34eff, 0xe6b95fff, 0xe7bc68ff, 0xe4b65aff, 0xe5b75bff, 
    0xe5b75cff, 0xe2b248ff, 0xe4b654ff, 0xe5b657ff, 0xe4b453ff, 0xe5b960ff, 
    0xe4b557ff, 0xe2b351ff, 0xe4b65cff, 0xe4b659ff, 0xe3b251ff, 0xe2b04aff, 
    0xe2b046ff, 0xe7b959ff, 0xebc171ff, 0xedc779ff, 0xecc571ff, 0xecc369ff, 
    0xf0cd81ff, 0xf1d38eff, 0xf4d691ff, 0xf6d993ff, 0xf7d992ff, 0xfae2a2ff, 
    0xfceab5ff, 0xfdeab0ff, 0xfce9acff, 0xfdebb1ff, 0xfdf2bfff, 0xfdf5c8ff, 
    0xfdf7d1ff, 0xfdf9d6ff, 0xfdf4cfff, 0xf6d99bff, 0xe3b254ff, 0xe4b35aff, 
    0xe5b75eff, 0xe2b354ff, 0xe2b355ff, 0xe2b357ff, 0xe2b459ff, 0xe4b85eff, 
    0xeac578ff, 0xeccd8bff, 0xeed197ff, 0xf0d39cff, 0xeccd90ff, 0xe8c683ff, 
    0xe6c583ff, 0xe8c68cff, 0xebcc94ff, 0xeacb92ff, 0xeacd94ff, 0xe9ca93ff, 
    0xe8c890ff, 0xe8c790ff, 0xe5c089ff, 0xeac79aff, 0xe7c497ff, 0xe3bb86ff, 
    0xddb06cff, 0xd9a558ff, 0xd5a051ff, 0xcf9852ff, 0xcf9548ff, 0xd99b35ff, 
    0xdfa642ff, 0xe5b25bff, 0xe6b964ff, 0xe8c37aff, 0xdfb879ff, 0xd6b06cff, 
    0xcfa658ff, 0xcb9e4aff, 0xc1963eff, 0xbd9241ff, 0xb2822cff, 0xa56e12ff, 
    0x9f6b17ff, 0x9e8057ff, 0xa18d6fff, 0xa18a68ff, 0x9f7f52ff, 0x986a22ff, 
    0x9d6d19ff, 0xac8237ff, 0xd6a84fff, 0xcfa041ff, 0xd4a443ff, 0xd7a94fff, 
    0xd6a952ff, 0xd6ab59ff, 0xd6ab5bff, 0xd1a449ff, 0xd1a144ff, 0xd5a752ff, 
    0xd2a553ff, 0xcb9633ff, 0xb4710cff, 0xb56f08ff, 0xc2881cff, 0xc79129ff, 
    0xc99228ff, 0xcf9b35ff, 0xd8a84bff, 0xdcad51ff, 0xe0b45cff, 0xe4ba63ff, 
    0xe7c06cff, 0xeac575ff, 0xeecd88ff, 0xf3d9a0ff, 0xf5dea5ff, 0xf7e1a5ff, 
    0xfae6b0ff, 0xfdebb7ff, 0xf6dc95ff, 0xdda93eff, 0xd49318ff, 0xd28e0aff, 
    0xd48f0aff, 0xdaa027ff, 0xdfac42ff, 0xe1af48ff, 0xe4b356ff, 0xe5b762ff, 
    0xe4b762ff, 0xe2b256ff, 0xe0af48ff, 0xe0ad3eff, 0xe3b24fff, 0xe3b251ff, 
    0xe2b14eff, 0xe3b556ff, 0xe2b350ff, 0xe2b452ff, 0xe4b65dff, 0xe3b457ff, 
    0xe3b252ff, 0xe3b24eff, 0xe5b655ff, 0xe7b95bff, 0xeabf6cff, 0xebc36fff, 
    0xebc370ff, 0xeec876ff, 0xf1d086ff, 0xf3d58fff, 0xf6db98ff, 0xf9e09fff, 
    0xf9df96ff, 0xfbe6a5ff, 0xfceab1ff, 0xfceaadff, 0xfdecb1ff, 0xfdf1bdff, 
    0xfdf4c7ff, 0xfdf7cbff, 0xfdf9d2ff, 0xfdf7cdff, 0xfdf0c1ff, 0xf6d993ff, 
    0xe5b658ff, 0xe1b154ff, 0xe2b050ff, 0xe0b14aff, 0xe2b556ff, 0xe3b659ff, 
    0xe1b257ff, 0xe1b355ff, 0xe7bf6bff, 0xe9c77fff, 0xebcb8bff, 0xedcf94ff, 
    0xeaca8aff, 0xe8c783ff, 0xe9c888ff, 0xe8c68bff, 0xe9c88eff, 0xeccf9bff, 
    0xecd29fff, 0xe9c991ff, 0xe9cb94ff, 0xe9c995ff, 0xe7c48fff, 0xe9c89aff, 
    0xe9c89fff, 0xe5be8fff, 0xe0b373ff, 0xdfb170ff, 0xdcad6aff, 0xd4a15cff, 
    0xd09c56ff, 0xd89c3fff, 0xe1a74fff, 0xe3b05eff, 0xe6b869ff, 0xe8c480ff, 
    0xe6c17dff, 0xd7ac5fff, 0xcc9c44ff, 0xca9d46ff, 0xc79c48ff, 0xc49d52ff, 
    0xba8e41ff, 0xa46b10ff, 0x9d6710ff, 0x9d783fff, 0xa58759ff, 0xa38c65ff, 
    0xa28b68ff, 0x9f8354ff, 0x9f7936ff, 0xa77f35ff, 0xe4bd6aff, 0xd8ab4fff, 
    0xd2a243ff, 0xd3a341ff, 0xd3a443ff, 0xd2a548ff, 0xd3a64cff, 0xd1a349ff, 
    0xd0a148ff, 0xcf9e41ff, 0xd1a047ff, 0xcb9733ff, 0xb97811ff, 0xae6704ff, 
    0xbe8114ff, 0xc58c23ff, 0xc79026ff, 0xcd9a3aff, 0xd4a750ff, 0xd9ac54ff, 
    0xdbae51ff, 0xe1b966ff, 0xe6c178ff, 0xe8c57bff, 0xeac77eff, 0xeecd8eff, 
    0xf2d49aff, 0xf3d89aff, 0xf4db9aff, 0xf7dfa0ff, 0xf9e29cff, 0xf9de90ff, 
    0xf2d07dff, 0xe7b651ff, 0xd89b24ff, 0xd99a1cff, 0xdca634ff, 0xdeac44ff, 
    0xe1ae4fff, 0xe3b358ff, 0xe4b561ff, 0xe1b056ff, 0xe0ad4aff, 0xddaa39ff, 
    0xdeab3eff, 0xe0af47ff, 0xe2b04bff, 0xe3b250ff, 0xe4b45aff, 0xe3b456ff, 
    0xe3b65aff, 0xe4b355ff, 0xe2b250ff, 0xe2b450ff, 0xe4b54fff, 0xe4b44dff, 
    0xe8bb5eff, 0xeabf62ff, 0xecc267ff, 0xf0ca79ff, 0xf3d389ff, 0xf4d489ff, 
    0xf6d88fff, 0xfae29dff, 0xfbe6a2ff, 0xfce7a4ff, 0xfce8a6ff, 0xfce8a4ff, 
    0xfceaa9ff, 0xfdefbaff, 0xfdf0bfff, 0xfdf5c9ff, 0xfdf9d2ff, 0xfdf9d2ff, 
    0xfdf0c0ff, 0xf1d082ff, 0xdfb04cff, 0xdfaf4cff, 0xe0af4dff, 0xe1b14dff, 
    0xe4b453ff, 0xe4b455ff, 0xe3b354ff, 0xe6bb67ff, 0xe8bf6eff, 0xe6be6aff, 
    0xe9c57bff, 0xe9c885ff, 0xe8c884ff, 0xe9c987ff, 0xebca8fff, 0xe9c88eff, 
    0xeacb90ff, 0xedd19dff, 0xeacd99ff, 0xe7c68cff, 0xe9c790ff, 0xe9c791ff, 
    0xe8c691ff, 0xe9c798ff, 0xe8c79cff, 0xe7c295ff, 0xe4bc89ff, 0xe7c395ff, 
    0xe2be8bff, 0xd8aa68ff, 0xcf994eff, 0xd89a40ff, 0xe5af63ff, 0xe6b771ff, 
    0xe5b86eff, 0xe6be75ff, 0xe7c683ff, 0xdeb66eff, 0xc99439ff, 0xc59236ff, 
    0xc69643ff, 0xc09140ff, 0xb28028ff, 0xa26f15ff, 0xa37424ff, 0xa77e38ff, 
    0xaa833fff, 0xa8833dff, 0xa9884dff, 0xa68857ff, 0xa18457ff, 0xa07d44ff, 
    0xe7c377ff, 0xe2b762ff, 0xd9ab52ff, 0xd7a74cff, 0xd7ab50ff, 0xd4a749ff, 
    0xd2a447ff, 0xd1a34aff, 0xd1a148ff, 0xd09e40ff, 0xcf9d3fff, 0xce9a38ff, 
    0xc08620ff, 0xaf6704ff, 0xb77609ff, 0xbd8116ff, 0xc08719ff, 0xc58e25ff, 
    0xcf9e45ff, 0xd3a54cff, 0xd7a745ff, 0xddaf56ff, 0xe3bc71ff, 0xe6c27dff, 
    0xe8c378ff, 0xeac578ff, 0xecc572ff, 0xedca7bff, 0xefd18bff, 0xf1d792ff, 
    0xf4da95ff, 0xf5dc95ff, 0xf5d88eff, 0xf6d27eff, 0xecbc59ff, 0xcf9234ff, 
    0xd3992eff, 0xd8a432ff, 0xdeac45ff, 0xdfad4bff, 0xe1af53ff, 0xdfae50ff, 
    0xddaa45ff, 0xdca93eff, 0xdca734ff, 0xdcaa3dff, 0xdfaf4dff, 0xe2b252ff, 
    0xe5b862ff, 0xe4b55aff, 0xe3b353ff, 0xe4b456ff, 0xe3b352ff, 0xe6b95eff, 
    0xe4b54fff, 0xe4b347ff, 0xe5b54bff, 0xe6ba55ff, 0xecc265ff, 0xefc874ff, 
    0xf1d184ff, 0xf4d88dff, 0xf7de98ff, 0xfae3a0ff, 0xf8df96ff, 0xfae49eff, 
    0xfce9abff, 0xfcecb5ff, 0xfcecb2ff, 0xfcedb4ff, 0xfdf1baff, 0xfdf5caff, 
    0xfdf5cbff, 0xfdefbfff, 0xf8e1a3ff, 0xe7bb5dff, 0xddaa3fff, 0xe2b454ff, 
    0xe5b75cff, 0xe5b456ff, 0xe6b759ff, 0xe3b151ff, 0xe4b354ff, 0xe9be6aff, 
    0xe8bd6cff, 0xe5ba64ff, 0xe9c579ff, 0xe8c47eff, 0xe6c27cff, 0xebca8eff, 
    0xedce96ff, 0xeac98eff, 0xeac989ff, 0xebcc92ff, 0xeccd95ff, 0xeccd98ff, 
    0xecd1a2ff, 0xe9cb9dff, 0xe4bf88ff, 0xe6c390ff, 0xe5c191ff, 0xe5bf8eff, 
    0xe4bc89ff, 0xe5be8dff, 0xe7c598ff, 0xe6c495ff, 0xdfb67fff, 0xdda75cff, 
    0xe4af62ff, 0xe6bb7bff, 0xe5bb75ff, 0xe5bb70ff, 0xe9ca8fff, 0xe7c88dff, 
    0xca953dff, 0xb97e1dff, 0xbd872eff, 0xaf7714ff, 0xad7d25ff, 0xb0883eff, 
    0xb79557ff, 0xb69554ff, 0xb3904dff, 0xb08c41ff, 0xb09148ff, 0xae8b44ff, 
    0xa88443ff, 0xa78344ff, 0xe7c27cff, 0xe4bb6bff, 0xe5bf76ff, 0xe5c17fff, 
    0xe2bc72ff, 0xe2ba6eff, 0xdeb363ff, 0xd9ac5bff, 0xd09e41ff, 0xcc9834ff, 
    0xd1a248ff, 0xcf9e46ff, 0xc48f2bff, 0xad6706ff, 0xb16b04ff, 0xb77609ff, 
    0xb97b0aff, 0xbd7e0bff, 0xc58c21ff, 0xce9b3dff, 0xd3a241ff, 0xd8a94aff, 
    0xddb15cff, 0xe4bd72ff, 0xe7be6eff, 0xe8c06dff, 0xe8c06cff, 0xebc269ff, 
    0xedc977ff, 0xeecf85ff, 0xf1d590ff, 0xf3d894ff, 0xf3d790ff, 0xf2d388ff, 
    0xedc46cff, 0xc88a3bff, 0xcc882bff, 0xce9432ff, 0xd6a237ff, 0xdaa73bff, 
    0xdaa63aff, 0xd9a639ff, 0xdaa636ff, 0xdba636ff, 0xdca737ff, 0xddab43ff, 
    0xdfad48ff, 0xe1b14fff, 0xe4b65dff, 0xe3b555ff, 0xe3b555ff, 0xe4b75bff, 
    0xe4b659ff, 0xe3b350ff, 0xe3b245ff, 0xe4b448ff, 0xe6b750ff, 0xe6b84eff, 
    0xe9bd57ff, 0xeec66aff, 0xf1d07eff, 0xf5d990ff, 0xf9e09dff, 0xf9df98ff, 
    0xf9e197ff, 0xfbe5a1ff, 0xfce8a7ff, 0xfdecb4ff, 0xfdedb3ff, 0xfdeeb0ff, 
    0xfdedb3ff, 0xfdeeb8ff, 0xfce9adff, 0xf9e19dff, 0xeecc78ff, 0xdcab3eff, 
    0xdfac42ff, 0xe3b459ff, 0xe3b457ff, 0xe6b757ff, 0xe8ba5eff, 0xe4b353ff, 
    0xe7b95eff, 0xe8bb66ff, 0xe7bb67ff, 0xe5ba63ff, 0xe6bd69ff, 0xe7be6fff, 
    0xe6be73ff, 0xeac98cff, 0xebcc93ff, 0xeaca8eff, 0xeccc8fff, 0xedd09aff, 
    0xecd2a1ff, 0xe9d4a9ff, 0xe8d3b0ff, 0xe1c9a3ff, 0xe2c393ff, 0xe3c292ff, 
    0xe5c597ff, 0xe8c89dff, 0xeacaa1ff, 0xe9c699ff, 0xe3bd85ff, 0xe2bb82ff, 
    0xe1bb86ff, 0xd9a047ff, 0xdea748ff, 0xe4b871ff, 0xe5bb77ff, 0xe4bd73ff, 
    0xe5c587ff, 0xe8cb93ff, 0xd9b370ff, 0xb27b1dff, 0xac7b1dff, 0xad812cff, 
    0xb28b3eff, 0xb7944fff, 0xc1a26cff, 0xc1a369ff, 0xc2a468ff, 0xbf9f59ff, 
    0xc1a058ff, 0xc6a65fff, 0xc8a45aff, 0xc8a156ff, 0xe3bd6bff, 0xe2b963ff, 
    0xe4bd72ff, 0xe3be7aff, 0xe5c27fff, 0xe8c688ff, 0xe5c17cff, 0xe6c487ff, 
    0xddb366ff, 0xcb952cff, 0xcd9936ff, 0xcc9739ff, 0xc38920ff, 0xae6605ff, 
    0xa95c04ff, 0xb16904ff, 0xb77305ff, 0xba7a07ff, 0xbf8411ff, 0xc79126ff, 
    0xd09e3dff, 0xd8a850ff, 0xdaa94dff, 0xdfb564ff, 0xe3b868ff, 0xe5bc6bff, 
    0xe6be6fff, 0xe9c072ff, 0xebc370ff, 0xecc36cff, 0xeec775ff, 0xefcc7eff, 
    0xf0cd7fff, 0xf1ce7fff, 0xf0ce7eff, 0xecbf5eff, 0xda9b2bff, 0xc28223ff, 
    0xb17a22ff, 0xd0972aff, 0xd69d29ff, 0xd79d28ff, 0xd6a12fff, 0xd8a230ff, 
    0xdaa636ff, 0xdcaa43ff, 0xe0ad4dff, 0xe0b052ff, 0xe3b35cff, 0xe5b65eff, 
    0xe3b559ff, 0xe4b557ff, 0xe4b85bff, 0xe3b455ff, 0xe3b146ff, 0xe4b347ff, 
    0xe4b345ff, 0xe5b546ff, 0xe8bc53ff, 0xeec76fff, 0xefcc72ff, 0xf5da8fff, 
    0xf7dc93ff, 0xfae19aff, 0xfce6a3ff, 0xfce8a4ff, 0xfce7a2ff, 0xfceaadff, 
    0xfcebaeff, 0xfdedb3ff, 0xfcefbaff, 0xfce7a7ff, 0xfae093ff, 0xf2cf7aff, 
    0xddaa3cff, 0xdaa42cff, 0xdeab3dff, 0xe1b14dff, 0xe5b658ff, 0xe5b653ff, 
    0xe7b959ff, 0xe6b654ff, 0xe8bb61ff, 0xebc172ff, 0xe9be6eff, 0xe8bd68ff, 
    0xe4b75bff, 0xe4b75dff, 0xe8c278ff, 0xe9c382ff, 0xe6c27dff, 0xe6c380ff, 
    0xe9ca8bff, 0xe8ca94ff, 0xe2cb9eff, 0xd5bd97ff, 0xa29177ff, 0x7d7463ff, 
    0xb49c75ff, 0xdaba8dff, 0xe0c39aff, 0xe7c9a4ff, 0xe9caa3ff, 0xe9c99cff, 
    0xe3bc83ff, 0xe4bf87ff, 0xe5c28cff, 0xdaa249ff, 0xdca343ff, 0xe0b05cff, 
    0xe4bb72ff, 0xe3bb70ff, 0xe5c384ff, 0xe6c88fff, 0xe4c792ff, 0xbc8c3bff, 
    0xad7f25ff, 0xb18a39ff, 0xb6934aff, 0xba9b55ff, 0xc0a466ff, 0xc5a664ff, 
    0xcba657ff, 0xd4ae61ff, 0xdcbd77ff, 0xdfc385ff, 0xdfc387ff, 0xe1c48dff, 
    0xe1b760ff, 0xdfb359ff, 0xe3b969ff, 0xe3bd75ff, 0xe6c58aff, 0xe8c792ff, 
    0xe3bf7eff, 0xe6c68dff, 0xe4c181ff, 0xdeb25eff, 0xcf9c36ff, 0xc79022ff, 
    0xc48a18ff, 0xb77309ff, 0xa75b04ff, 0xa35804ff, 0xb16a04ff, 0xb77507ff, 
    0xbb7b09ff, 0xc0830eff, 0xcf9b3cff, 0xd8a958ff, 0xd8a84cff, 0xdcad57ff, 
    0xdfb25dff, 0xe1b562ff, 0xe5ba65ff, 0xe7bb64ff, 0xe8be64ff, 0xeac16bff, 
    0xecc470ff, 0xefc66fff, 0xefc66fff, 0xefc871ff, 0xf1c96eff, 0xf0c561ff, 
    0xe6b145ff, 0xe2a836ff, 0xd39020ff, 0xca852aff, 0xd09431ff, 0xd29722ff, 
    0xd59e2cff, 0xd7a236ff, 0xd69e28ff, 0xd9a330ff, 0xdca942ff, 0xdca941ff, 
    0xdfad4eff, 0xe1b153ff, 0xe2b253ff, 0xe2b253ff, 0xe3b352ff, 0xe2b04cff, 
    0xe1af44ff, 0xe2b246ff, 0xe4b241ff, 0xe4b440ff, 0xe7ba50ff, 0xedc261ff, 
    0xefc96cff, 0xf1d07cff, 0xf4d78aff, 0xf7dc90ff, 0xf8e29cff, 0xfbe7a3ff, 
    0xfce7a1ff, 0xfceaa7ff, 0xfcebaaff, 0xfdf0bbff, 0xfdefbaff, 0xf9e098ff, 
    0xf1ce76ff, 0xe1b24aff, 0xd8a22bff, 0xdba42dff, 0xdfab39ff, 0xe8b857ff, 
    0xefc677ff, 0xeec471ff, 0xecc16aff, 0xe9ba59ff, 0xe9ba5aff, 0xe9bd68ff, 
    0xebc273ff, 0xebc476ff, 0xe9bf6eff, 0xe6b964ff, 0xe9c177ff, 0xe7bf74ff, 
    0xe8c27dff, 0xe6c688ff, 0xe0c385ff, 0xdbc08dff, 0xbaa47dff, 0x7c715fff, 
    0x8d8c89ff, 0xafb0b0ff, 0x6f695eff, 0xa8916eff, 0xd4b991ff, 0xdfc39dff, 
    0xe5c79eff, 0xe8cda0ff, 0xe7c490ff, 0xe7c38eff, 0xe6c48fff, 0xdeae64ff, 
    0xdda64dff, 0xe0b05aff, 0xe4bd78ff, 0xe5c07eff, 0xead09eff, 0xecd6aaff, 
    0xe9d0a1ff, 0xcba35eff, 0xb28631ff, 0xb68f3fff, 0xb9984cff, 0xbd9d55ff, 
    0xc49e4cff, 0xcfa34dff, 0xd9b161ff, 0xe0c282ff, 0xe1ca93ff, 0xe3ca96ff, 
    0xe2cb99ff, 0xe2cd9eff, 0xe3b862ff, 0xe2b75cff, 0xe1b762ff, 0xdfb55fff, 
    0xe3bc73ff, 0xe5c180ff, 0xdfb768ff, 0xe1bc77ff, 0xe0b970ff, 0xdfb76dff, 
    0xdcb264ff, 0xd2a042ff, 0xc89125ff, 0xbe8111ff, 0xa75f05ff, 0x9a4c04ff, 
    0xae6504ff, 0xb46f04ff, 0xb67305ff, 0xbe800eff, 0xcd9b3dff, 0xd6a959ff, 
    0xdaac5cff, 0xd9a94fff, 0xdaa847ff, 0xddad51ff, 0xe3b761ff, 0xe6bd68ff, 
    0xe7bc63ff, 0xe9be69ff, 0xe9bf67ff, 0xebc067ff, 0xecc36fff, 0xefc671ff, 
    0xefc66dff, 0xf0c76eff, 0xf1c566ff, 0xefbe4cff, 0xeeb63bff, 0xd3912cff, 
    0xb16824ff, 0xc17d22ff, 0xd1972aff, 0xd29a27ff, 0xd3991fff, 0xd59e26ff, 
    0xd9a336ff, 0xdaa537ff, 0xdba73aff, 0xdbaa3eff, 0xdeac43ff, 0xdfac43ff, 
    0xdfaf48ff, 0xdfae44ff, 0xdfac3bff, 0xe0ad3dff, 0xe3b141ff, 0xe4b342ff, 
    0xe7b84bff, 0xedc364ff, 0xefc667ff, 0xf1ce77ff, 0xf4d88aff, 0xf8df96ff, 
    0xf7dd90ff, 0xf9e3a0ff, 0xfcedb8ff, 0xfcefbaff, 0xfdf2c2ff, 0xfdf2c5ff, 
    0xfceaaeff, 0xf8df98ff, 0xdfb14cff, 0xd8a22bff, 0xdaa52dff, 0xdea933ff, 
    0xe8b953ff, 0xf0c771ff, 0xf1cc7eff, 0xf1ce83ff, 0xf1ce85ff, 0xedc269ff, 
    0xe8b855ff, 0xe9bb5eff, 0xebc16cff, 0xecc473ff, 0xecc680ff, 0xeac277ff, 
    0xe8bd6dff, 0xe7be70ff, 0xe4be79ff, 0xdbb975ff, 0xcdaf75ff, 0x9b8764ff, 
    0x716b5fff, 0xabababff, 0xe0e1e0ff, 0xe5e6e4ff, 0xbcbdbcff, 0x6a645aff, 
    0xa08764ff, 0xd0b085ff, 0xdebf90ff, 0xe7ca9aff, 0xeacc99ff, 0xeac995ff, 
    0xe6c690ff, 0xddac5eff, 0xdfab57ff, 0xe4bd7cff, 0xe6c58dff, 0xe1bd7fff, 
    0xe2c285ff, 0xead6acff, 0xead6acff, 0xd2ad6cff, 0xb88e3eff, 0xba9445ff, 
    0xbd9a4cff, 0xc8a049ff, 0xd3a240ff, 0xd8ad5aff, 0xddbb75ff, 0xe0c48cff, 
    0xe0c897ff, 0xe0cb9aff, 0xe1cda0ff, 0xe0ca9eff, 0xe6bb68ff, 0xe3b964ff, 
    0xe0b459ff, 0xdfb050ff, 0xe4bb6dff, 0xe4bd73ff, 0xe4ba6bff, 0xe0b665ff, 
    0xddb25dff, 0xdeb667ff, 0xe1bc7aff, 0xdcb264ff, 0xd29f41ff, 0xc89029ff, 
    0xa8630eff, 0x954504ff, 0xab5f04ff, 0xaf6905ff, 0xb57104ff, 0xbb7a07ff, 
    0xc6902aff, 0xd1a24dff, 0xd7ab5cff, 0xd8ab55ff, 0xdbaa4bff, 0xddad51ff, 
    0xe0b564ff, 0xe2b766ff, 0xe5ba64ff, 0xe7bc68ff, 0xe8bd62ff, 0xe7bb5cff, 
    0xe9bd5eff, 0xecbf5fff, 0xeec160ff, 0xeec362ff, 0xefc463ff, 0xf0c45eff, 
    0xf0c25dff, 0xedbc54ff, 0xd1912eff, 0xb67a2cff, 0xd59f48ff, 0xd59b2fff, 
    0xd39822ff, 0xd19821ff, 0xd49d27ff, 0xd79f2fff, 0xd7a12bff, 0xdaa637ff, 
    0xdcab40ff, 0xddab3fff, 0xdca839ff, 0xdda834ff, 0xdeaa38ff, 0xdfac3dff, 
    0xe1af3fff, 0xe4b344ff, 0xe6b74dff, 0xedc366ff, 0xf0ca72ff, 0xf0cb72ff, 
    0xf2d280ff, 0xf7e19bff, 0xfae49fff, 0xfbe7a8ff, 0xfdedb9ff, 0xfcedb8ff, 
    0xfcedb8ff, 0xfbe6aaff, 0xf8dd93ff, 0xe8c46cff, 0xd7a129ff, 0xdaa329ff, 
    0xdea832ff, 0xe6b64bff, 0xeec366ff, 0xf0cc7dff, 0xf2d087ff, 0xf3d38cff, 
    0xf1cf8aff, 0xedc26dff, 0xe9b959ff, 0xedc16bff, 0xecc16eff, 0xeabe66ff, 
    0xecc475ff, 0xecc780ff, 0xe7bd72ff, 0xe1bc77ff, 0xd5b16bff, 0xb99a5cff, 
    0x796a4bff, 0x7e7b75ff, 0xcacbcaff, 0xe4e5e3ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe4e4e3ff, 0xbbbcbbff, 0x696258ff, 0xa18a68ff, 0xd0b281ff, 0xdfc08fff, 
    0xe6ca97ff, 0xeacf9eff, 0xe8ca98ff, 0xddac61ff, 0xe0b369ff, 0xe5c187ff, 
    0xe0bc81ff, 0xdbb56fff, 0xddb972ff, 0xe7cfa0ff, 0xe7cf9fff, 0xd6b16fff, 
    0xb88e3fff, 0xb48d3bff, 0xc39b44ff, 0xd6ab4eff, 0xdcb35cff, 0xe1bf80ff, 
    0xe2c68fff, 0xdfc28aff, 0xe0c694ff, 0xdec792ff, 0xdfc895ff, 0xe1c896ff, 
    0xe5b960ff, 0xe5bc69ff, 0xe4b960ff, 0xe4b85eff, 0xe6bd71ff, 0xe5bf76ff, 
    0xe7c582ff, 0xe5c17aff, 0xdfb466ff, 0xd9aa55ff, 0xdbae60ff, 0xddb165ff, 
    0xdaad5aff, 0xd8aa55ff, 0xc48e32ff, 0x8f4308ff, 0xa45704ff, 0xae6605ff, 
    0xb26c04ff, 0xb97907ff, 0xc08517ff, 0xca973bff, 0xd3a451ff, 0xd5a74dff, 
    0xdbaf59ff, 0xdcae57ff, 0xdbae54ff, 0xddb15aff, 0xe2b65eff, 0xe3b761ff, 
    0xe3b456ff, 0xe4b554ff, 0xe6b856ff, 0xe8b955ff, 0xebbd5dff, 0xecbf5eff, 
    0xecbe57ff, 0xedbd57ff, 0xeebf5aff, 0xf0c362ff, 0xf0c055ff, 0xeaba50ff, 
    0xebbe5fff, 0xe5b255ff, 0xe4b453ff, 0xcf901bff, 0xcc8c0fff, 0xcf8e10ff, 
    0xd49a20ff, 0xd8a030ff, 0xdaa538ff, 0xdca83bff, 0xdba632ff, 0xdca632ff, 
    0xdda62fff, 0xdfaa39ff, 0xe1ae3eff, 0xe2b143ff, 0xe5b74eff, 0xe9bd5aff, 
    0xedc463ff, 0xf0cb70ff, 0xf3d485ff, 0xf4da8eff, 0xf8e29aff, 0xfbe7a7ff, 
    0xfae7a8ff, 0xf9e4a0ff, 0xf8de95ff, 0xf6d582ff, 0xebc467ff, 0xdaa537ff, 
    0xd9a127ff, 0xdda52bff, 0xe4b445ff, 0xeabc57ff, 0xefc66cff, 0xf0cb7bff, 
    0xf2d086ff, 0xf3d48dff, 0xf2d494ff, 0xf0ca80ff, 0xecbe62ff, 0xecbf65ff, 
    0xecbf6aff, 0xeabc65ff, 0xeac16fff, 0xe6bf72ff, 0xdfba70ff, 0xd1b171ff, 
    0xa1885aff, 0x6a614dff, 0x9c9c9bff, 0xdddddcff, 0xe3e4e2ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe3e4e3ff, 0xbbbcbbff, 0x696359ff, 
    0xa18964ff, 0xd1b484ff, 0xdebf8dff, 0xe6cb9aff, 0xe9cd9fff, 0xdfb36dff, 
    0xe0b36bff, 0xe5c188ff, 0xe5c28cff, 0xe3c189ff, 0xe4c68dff, 0xe7cb96ff, 
    0xe6cb91ff, 0xdab775ff, 0xbd9242ff, 0xbc903aff, 0xd7ae5aff, 0xe3bf79ff, 
    0xe3c48aff, 0xe3c99aff, 0xe2c89aff, 0xe0c388ff, 0xe1c895ff, 0xdcc28bff, 
    0xdec187ff, 0xdcbe7cff, 0xe8bf70ff, 0xe7c278ff, 0xe4bb67ff, 0xe5bd6fff, 
    0xe7c278ff, 0xe9c785ff, 0xe8c586ff, 0xe6c383ff, 0xe3bd78ff, 0xe0b873ff, 
    0xd39f49ff, 0xd3a14eff, 0xdbb167ff, 0xdcb46cff, 0xd8ad5aff, 0xac7027ff, 
    0x9f5104ff, 0xad6304ff, 0xb06c04ff, 0xb67305ff, 0xbc7c0aff, 0xc58d27ff, 
    0xce9a3eff, 0xd1a242ff, 0xd6a951ff, 0xdaac56ff, 0xd9a849ff, 0xdbad53ff, 
    0xddb058ff, 0xdeaf50ff, 0xdfad46ff, 0xdfae46ff, 0xe2b14aff, 0xe3b348ff, 
    0xe4b44bff, 0xe6b54cff, 0xe8b549ff, 0xe8b648ff, 0xeab950ff, 0xebbb53ff, 
    0xedbc54ff, 0xeebe57ff, 0xf2c769ff, 0xf4d17bff, 0xf4d07bff, 0xeabd63ff, 
    0xddac4bff, 0xcd8d15ff, 0xcf9219ff, 0xd39b2aff, 0xd69f2eff, 0xd7a32fff, 
    0xd8a22cff, 0xd9a129ff, 0xdaa52eff, 0xdda83aff, 0xe1ae4aff, 0xe2b147ff, 
    0xe4b64eff, 0xe7bb55ff, 0xebc05aff, 0xefc86aff, 0xf5d992ff, 0xf5db91ff, 
    0xf5dc90ff, 0xf7df9aff, 0xfbe7abff, 0xf8e19dff, 0xf4d480ff, 0xeac160ff, 
    0xdeb358ff, 0xdaa73cff, 0xdda730ff, 0xe2ad35ff, 0xe6b74aff, 0xebbd5aff, 
    0xefc770ff, 0xf0ce83ff, 0xf3d592ff, 0xf4d794ff, 0xf2d393ff, 0xefc87cff, 
    0xedc166ff, 0xebbc61ff, 0xecc16cff, 0xebc371ff, 0xe5c074ff, 0xdbb567ff, 
    0xc4a462ff, 0x82714dff, 0x737069ff, 0xbfc0bfff, 0xe3e4e3ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0x9e9f9dff, 0x615b4fff, 0xa68d65ff, 0xd5b481ff, 0xe0c391ff, 
    0xe4c593ff, 0xdcaf64ff, 0xdeb064ff, 0xe3be83ff, 0xe4c28aff, 0xe6c892ff, 
    0xe5c890ff, 0xe6c88cff, 0xe8cc90ff, 0xdfbd79ff, 0xc59745ff, 0xd0a756ff, 
    0xe5c485ff, 0xe7c999ff, 0xe7cfa5ff, 0xe5cea1ff, 0xe0c48aff, 0xe2c68aff, 
    0xe0c793ff, 0xddc392ff, 0xdcc18cff, 0xdec283ff, 0xe7c37cff, 0xe6bf70ff, 
    0xe6bd68ff, 0xe6bb68ff, 0xe8c277ff, 0xe9c785ff, 0xe7c587ff, 0xe4c180ff, 
    0xe1b96fff, 0xe2bb78ff, 0xdaae62ff, 0xd29b40ff, 0xdaae62ff, 0xdab26bff, 
    0xdab164ff, 0xd2a450ff, 0xa96410ff, 0xa95f04ff, 0xad6604ff, 0xb26c04ff, 
    0xb87706ff, 0xc48b24ff, 0xce993aff, 0xd1a243ff, 0xd3a54bff, 0xd6a547ff, 
    0xd7a642ff, 0xd7a53eff, 0xd8a843ff, 0xdca947ff, 0xddaa44ff, 0xdca93fff, 
    0xdbaa3fff, 0xdfac40ff, 0xdfab3aff, 0xe1ad40ff, 0xe2ac39ff, 0xe3ad3bff, 
    0xe5af40ff, 0xe7b03fff, 0xe7b03cff, 0xe9b549ff, 0xecbb5aff, 0xf0c86fff, 
    0xf4d486ff, 0xf6d485ff, 0xf9db8cff, 0xeeca74ff, 0xd49c34ff, 0xcb8e17ff, 
    0xd09521ff, 0xd29720ff, 0xd49a20ff, 0xd69d24ff, 0xd9a435ff, 0xdaa434ff, 
    0xdba635ff, 0xe0ac42ff, 0xe5b451ff, 0xe8bc5cff, 0xebc160ff, 0xedc465ff, 
    0xf0cf7bff, 0xf5da92ff, 0xf5dc92ff, 0xf8e3a7ff, 0xf9e6b2ff, 0xf3d589ff, 
    0xf0cb70ff, 0xdcac43ff, 0xd59d28ff, 0xdba431ff, 0xe0ab35ff, 0xe3b242ff, 
    0xe6b74dff, 0xeabe5bff, 0xeec56cff, 0xf0d18bff, 0xf1d38eff, 0xf4d798ff, 
    0xf4d898ff, 0xf1ce87ff, 0xf1c977ff, 0xf0c977ff, 0xecca7fff, 0xe4c079ff, 
    0xd7b56eff, 0xab9056ff, 0x6d6249ff, 0x8e8e8cff, 0xd7d8d7ff, 0xe4e4e3ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xdfe0deff, 0xb6b7b4ff, 0x92948fff, 0x737573ff, 0x6d604eff, 
    0xcdb082ff, 0xdbbb83ff, 0xdeb97aff, 0xd9a650ff, 0xdcaa50ff, 0xe0b871ff, 
    0xe0bd7cff, 0xe3c388ff, 0xdfbf7eff, 0xe3c684ff, 0xe4c684ff, 0xd9b061ff, 
    0xcd9b3fff, 0xdfb771ff, 0xe5c48bff, 0xe5c896ff, 0xe7cfa3ff, 0xe5ca94ff, 
    0xe1c383ff, 0xe4ca95ff, 0xe0c68fff, 0xd9b97dff, 0xdbbc81ff, 0xdebf86ff, 
    0xe4b964ff, 0xe4b861ff, 0xe7c06fff, 0xe5be6bff, 0xe6bf71ff, 0xe6c17bff, 
    0xe5c383ff, 0xe2bc74ff, 0xe1b86dff, 0xe0b870ff, 0xdeb66cff, 0xdbaf5aff, 
    0xdaaf60ff, 0xd9b167ff, 0xd9ad63ff, 0xd6ab5bff, 0xbe872dff, 0xa85f05ff, 
    0xab6304ff, 0xaf6904ff, 0xb46f04ff, 0xc0851eff, 0xcc9a3bff, 0xd2a64eff, 
    0xd3a652ff, 0xd5a751ff, 0xd5a543ff, 0xd7a43aff, 0xdaa743ff, 0xdaa845ff, 
    0xdaa847ff, 0xdaa73fff, 0xd9a637ff, 0xdaa83dff, 0xdca83dff, 0xdca839ff, 
    0xdba42bff, 0xdda328ff, 0xdea229ff, 0xdea022ff, 0xdda126ff, 0xe1a838ff, 
    0xe5af41ff, 0xe9b852ff, 0xeec46bff, 0xf3cd73ff, 0xf7d67eff, 0xfae195ff, 
    0xf5dc97ff, 0xe6bb5fff, 0xe1b04eff, 0xd59b2eff, 0xcd9018ff, 0xcf9317ff, 
    0xd39721ff, 0xd59c23ff, 0xd89f26ff, 0xdca533ff, 0xdfac3dff, 0xe4b54eff, 
    0xe9be5bff, 0xedc465ff, 0xeec76aff, 0xf4d68dff, 0xf5da93ff, 0xf5dd9bff, 
    0xf5dc9aff, 0xf1d07aff, 0xe5b857ff, 0xd59d2bff, 0xd7a02bff, 0xdca52fff, 
    0xe0ac34ff, 0xe2b344ff, 0xe6b953ff, 0xe9bc59ff, 0xecc369ff, 0xeecc80ff, 
    0xf0d18eff, 0xefcf8dff, 0xf2d393ff, 0xefce86ff, 0xefcc7bff, 0xebca82ff, 
    0xe0c07aff, 0xcba862ff, 0x8b7548ff, 0x6b655aff, 0xb2b3b2ff, 0xe2e3e1ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xdfe0deff, 0xb5b6b3ff, 0x8d8f8aff, 0x828480ff, 
    0x535047ff, 0x8e7a5bff, 0xcaaf82ff, 0xd9b87eff, 0xdab269ff, 0xd8a343ff, 
    0xdeaf5cff, 0xe0b976ff, 0xdeb566ff, 0xdfbc70ff, 0xe0bf77ff, 0xe2c27eff, 
    0xddba70ff, 0xd2a146ff, 0xd39c35ff, 0xdaa854ff, 0xe7c489ff, 0xe1c288ff, 
    0xe4c893ff, 0xe5ca91ff, 0xe5c98cff, 0xe2c487ff, 0xddbf83ff, 0xdfc288ff, 
    0xe1c590ff, 0xdfbe80ff, 0xe4b75eff, 0xe5ba65ff, 0xe5be6dff, 0xe6be6cff, 
    0xe8c788ff, 0xe7c487ff, 0xe5c282ff, 0xe1ba70ff, 0xdfb565ff, 0xddb260ff, 
    0xddb464ff, 0xddb466ff, 0xdbb368ff, 0xdab26aff, 0xd9af66ff, 0xd5aa5cff, 
    0xca9a3cff, 0xac680aff, 0xa96104ff, 0xae6804ff, 0xb16d04ff, 0xbd831cff, 
    0xc89432ff, 0xd0a24aff, 0xd6ab5fff, 0xd2a557ff, 0xd1a248ff, 0xd7a64aff, 
    0xd8a94aff, 0xd7a743ff, 0xd8a541ff, 0xd9a640ff, 0xd7a333ff, 0xd7a12fff, 
    0xd7a029ff, 0xd79f29ff, 0xd69b1dff, 0xd89b21ff, 0xd9981dff, 0xd89615ff, 
    0xd69310ff, 0xd89719ff, 0xdba12aff, 0xe0a42fff, 0xe4ae3eff, 0xe9b546ff, 
    0xefc35bff, 0xf6d47eff, 0xf9e39cff, 0xfce6a1ff, 0xfdebaeff, 0xf7e0a4ff, 
    0xe9c06fff, 0xcf9224ff, 0xca860aff, 0xce8f13ff, 0xd19316ff, 0xd5981aff, 
    0xdba42dff, 0xe1ae3fff, 0xe3b74bff, 0xe9c15bff, 0xedc562ff, 0xf0ce7dff, 
    0xf0cf80ff, 0xefce7dff, 0xeecb77ff, 0xe6bc5eff, 0xdaa742ff, 0xd49c27ff, 
    0xdaa332ff, 0xdba329ff, 0xe1ae3bff, 0xe5b54dff, 0xe6b753ff, 0xe9bb5aff, 
    0xedc66fff, 0xedcb7fff, 0xefce87ff, 0xf1d18bff, 0xeecc84ff, 0xecc87bff, 
    0xe7c270ff, 0xdbb86eff, 0xb69a5eff, 0x716344ff, 0x807f7cff, 0xcfd0cfff, 
    0xe4e5e3ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xdfe0ddff, 0xb4b5b2ff, 0x8d8f8aff, 
    0x828481ff, 0x4d4b44ff, 0x6f5d3aff, 0xaf9569ff, 0xccaf7cff, 0xd8b370ff, 
    0xd5a751ff, 0xd7a242ff, 0xdcac58ff, 0xdeb266ff, 0xd7a744ff, 0xd9b04dff, 
    0xe1be70ff, 0xe4c687ff, 0xdab36bff, 0xd49c3eff, 0xdaa23bff, 0xdfb056ff, 
    0xe7c485ff, 0xe9c993ff, 0xe6c78cff, 0xe4c68bff, 0xe5c88eff, 0xe4c68dff, 
    0xd8b879ff, 0xd8b776ff, 0xd9b879ff, 0xdbbb7aff, 0xe6c072ff, 0xe6c279ff, 
    0xe4ba67ff, 0xe5bf72ff, 0xe5c383ff, 0xe2bc73ff, 0xe1b969ff, 0xe2bc72ff, 
    0xe0b769ff, 0xdbaf56ff, 0xdaad55ff, 0xdeb466ff, 0xdeb670ff, 0xdab066ff, 
    0xd6ac5eff, 0xd4a959ff, 0xc79435ff, 0xa7610cff, 0xa55a04ff, 0xac6304ff, 
    0xb16d04ff, 0xb7790dff, 0xc38c23ff, 0xcd9c3eff, 0xd5aa5aff, 0xd3a857ff, 
    0xcf9f47ff, 0xd3a54eff, 0xd8a74aff, 0xd7a641ff, 0xd8a544ff, 0xd8a542ff, 
    0xd5a134ff, 0xd59c25ff, 0xd3981bff, 0xd39518ff, 0xd39416ff, 0xd39112ff, 
    0xd39115ff, 0xd38e10ff, 0xd38e0cff, 0xd18d0dff, 0xd49110ff, 0xd79413ff, 
    0xda9611ff, 0xdb9915ff, 0xde9f21ff, 0xebb94cff, 0xf3cf79ff, 0xf9de94ff, 
    0xfce6a5ff, 0xfdeeb7ff, 0xfdedb7ff, 0xf1d188ff, 0xd1962aff, 0xc58105ff, 
    0xcb8b14ff, 0xd0941dff, 0xd49922ff, 0xdaa533ff, 0xe1af41ff, 0xe7bc54ff, 
    0xeac161ff, 0xedc66bff, 0xefca75ff, 0xecc364ff, 0xe6b954ff, 0xd9a642ff, 
    0xd9a43bff, 0xdaa538ff, 0xdaa530ff, 0xdca72cff, 0xe0b03eff, 0xe2b349ff, 
    0xe6b856ff, 0xe9bf64ff, 0xebc267ff, 0xedc978ff, 0xefce86ff, 0xeed08cff, 
    0xe9ca84ff, 0xe2c47fff, 0xd2b36dff, 0x978250ff, 0x676152ff, 0xa4a5a4ff, 
    0xdfe0deff, 0xe3e4e2ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xdedfddff, 0xb3b5b1ff, 
    0x8d8f8aff, 0x828481ff, 0x4b4a45ff, 0x655943ff, 0x9c865dff, 0xbca070ff, 
    0xd1b179ff, 0xd9b672ff, 0xd5a64eff, 0xd5a244ff, 0xd8a850ff, 0xdbac56ff, 
    0xdaac4cff, 0xdfb865ff, 0xe5c784ff, 0xe9cf97ff, 0xe0bb7dff, 0xdba751ff, 
    0xe2b15dff, 0xe5bd6dff, 0xecca8dff, 0xefd19cff, 0xeacc93ff, 0xe1c283ff, 
    0xdfc080ff, 0xdfc081ff, 0xdebf7eff, 0xdab976ff, 0xdab875ff, 0xdcbb77ff, 
    0xe5be72ff, 0xe5be71ff, 0xe3ba65ff, 0xe4be72ff, 0xe2bb70ff, 0xe0b664ff, 
    0xdfb561ff, 0xdfb564ff, 0xdcaf58ff, 0xdcaf57ff, 0xe0ba74ff, 0xe1bc7dff, 
    0xdeb670ff, 0xdab064ff, 0xd8ad62ff, 0xd3a855ff, 0xce9d40ff, 0xc58d2eff, 
    0xb17012ff, 0xa96205ff, 0xaf6c05ff, 0xb27104ff, 0xbe8315ff, 0xcc9a3aff, 
    0xd1a34bff, 0xd3a653ff, 0xd0a34fff, 0xd1a452ff, 0xd3a348ff, 0xd7a545ff, 
    0xd5a23dff, 0xd39e2dff, 0xd29c29ff, 0xd3971eff, 0xd59b24ff, 0xda9d25ff, 
    0xdb9d24ff, 0xd69518ff, 0xd49116ff, 0xd28b0eff, 0xd08a0fff, 0xcf8a0dff, 
    0xd0880bff, 0xd08806ff, 0xd38a06ff, 0xd38906ff, 0xd18805ff, 0xd9961cff, 
    0xe9b750ff, 0xf2c969ff, 0xf6d37bff, 0xf8df96ff, 0xfae8abff, 0xfdecb2ff, 
    0xf1d791ff, 0xd8a84eff, 0xd59f3cff, 0xcc8c1eff, 0xcb8a11ff, 0xd29620ff, 
    0xdaa22aff, 0xe2b245ff, 0xeac265ff, 0xedc972ff, 0xeec86dff, 0xeac25cff, 
    0xe4b445ff, 0xdba942ff, 0xd9a83fff, 0xdba83bff, 0xdaa52fff, 0xdba62cff, 
    0xe0ae39ff, 0xe2b246ff, 0xe6b958ff, 0xe9be64ff, 0xecc672ff, 0xedcc82ff, 
    0xeacb85ff, 0xe5c57fff, 0xdbb96eff, 0xbc9e5cff, 0x766848ff, 0x75736fff, 
    0xc5c6c6ff, 0xe4e4e3ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xdedfddff, 
    0xb3b4b1ff, 0x8d8f8aff, 0x828481ff, 0x4c4b45ff, 0x645d49ff, 0x9a8e74ff, 
    0xbbab89ff, 0xd1b88dff, 0xdfc293ff, 0xdfbc7cff, 0xd7a852ff, 0xd19d37ff, 
    0xd5a341ff, 0xdaa84bff, 0xdbad50ff, 0xe2bf71ff, 0xe7c889ff, 0xebd39dff, 
    0xe3c07cff, 0xdda74dff, 0xe2b25dff, 0xe7c27eff, 0xeacb8fff, 0xefd29fff, 
    0xedd09bff, 0xe1bf79ff, 0xddbc76ff, 0xe3c584ff, 0xe5c88bff, 0xe4c78dff, 
    0xdfbe7bff, 0xdcb86eff, 0xe5bd70ff, 0xe2b966ff, 0xe2b660ff, 0xe2b869ff, 
    0xe1b566ff, 0xdfb462ff, 0xe1b769ff, 0xddae56ff, 0xdaab50ff, 0xdeb568ff, 
    0xdfbc79ff, 0xddb872ff, 0xdab265ff, 0xd8af5fff, 0xdab46dff, 0xd6ad5fff, 
    0xcd9e43ff, 0xcc9839ff, 0xc38d21ff, 0xb2700cff, 0xaa6304ff, 0xb06c04ff, 
    0xb87a0cff, 0xc79530ff, 0xcea042ff, 0xd2a755ff, 0xd1a558ff, 0xd3a85dff, 
    0xd0a24dff, 0xd2a142ff, 0xd29e31ff, 0xd29c2dff, 0xd19924ff, 0xd4981fff, 
    0xe1a631ff, 0xe6ad3cff, 0xe6a831ff, 0xe8ad3dff, 0xe7aa38ff, 0xebb449ff, 
    0xe8b54fff, 0xe5ad45ff, 0xe1a231ff, 0xdd9c21ff, 0xde9e25ff, 0xd89315ff, 
    0xd69017ff, 0xd58f18ff, 0xdfa234ff, 0xe4b249ff, 0xe8b74bff, 0xeec05bff, 
    0xf2d178ff, 0xf7db93ff, 0xfae39dff, 0xfbe79fff, 0xf9e09cff, 0xeeca7dff, 
    0xd9a343ff, 0xcd8f22ff, 0xd0921eff, 0xdaa42fff, 0xe6b854ff, 0xedc66cff, 
    0xeec974ff, 0xecc873ff, 0xe9c062ff, 0xddac47ff, 0xddaa43ff, 0xdeab41ff, 
    0xdaa42eff, 0xdaa429ff, 0xe0ad3aff, 0xe3b54bff, 0xe7bb5aff, 0xeac168ff, 
    0xeac472ff, 0xe8c881ff, 0xe2c481ff, 0xd6ba7aff, 0x9e8650ff, 0x645b46ff, 
    0x969795ff, 0xdadbdaff, 0xe3e4e2ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xdedfddff, 0xb2b3b0ff, 0x8d8f8aff, 0x828481ff, 0x4c4b47ff, 0x645d4bff, 
    0x9b9179ff, 0xbbb097ff, 0xd5c6a8ff, 0xe1cca4ff, 0xe9d3acff, 0xe1bd81ff, 
    0xd39e42ff, 0xd19b39ff, 0xd2a042ff, 0xdbae54ff, 0xe0b868ff, 0xe4c47fff, 
    0xe4c47fff, 0xe5c784ff, 0xe3bc6fff, 0xe1b056ff, 0xe4ba6cff, 0xeccc90ff, 
    0xedcd93ff, 0xeacb93ff, 0xe8c98fff, 0xe6c684ff, 0xe6c684ff, 0xe7c88bff, 
    0xe4c887ff, 0xe1c381ff, 0xdeba6fff, 0xdeb96dff, 0xe5bc6fff, 0xe2b96aff, 
    0xe3b868ff, 0xe2b868ff, 0xe1b562ff, 0xe0b665ff, 0xe0b665ff, 0xe0b669ff, 
    0xdfb975ff, 0xddb871ff, 0xddbb75ff, 0xdcb770ff, 0xdab366ff, 0xd8af60ff, 
    0xdab269ff, 0xd9b068ff, 0xd2a653ff, 0xcb9838ff, 0xc48e21ff, 0xb97e10ff, 
    0xa86005ff, 0xab6304ff, 0xb57508ff, 0xc28c23ff, 0xcd9d3fff, 0xd2a856ff, 
    0xd5ac63ff, 0xd4aa61ff, 0xd0a454ff, 0xcd9e42ff, 0xd19e35ff, 0xd19b2bff, 
    0xcf951dff, 0xcf941aff, 0xd5971cff, 0xd7971aff, 0xd68d09ff, 0xdb951aff, 
    0xdf9d2aff, 0xe3a740ff, 0xe7b246ff, 0xedbd63ff, 0xe7b14eff, 0xe1a334ff, 
    0xd99018ff, 0xdc9211ff, 0xe3a328ff, 0xe7af3dff, 0xeec265ff, 0xefc86dff, 
    0xedc461ff, 0xefc55eff, 0xefc55dff, 0xeabd53ff, 0xe6b74fff, 0xedc563ff, 
    0xf3d383ff, 0xf6d88fff, 0xf2ce7aff, 0xe6b454ff, 0xcf8f29ff, 0xc98712ff, 
    0xdda834ff, 0xe5b74fff, 0xebc46bff, 0xedc873ff, 0xe9c162ff, 0xdeac46ff, 
    0xdfae4bff, 0xdead45ff, 0xdba93bff, 0xdca939ff, 0xdfaf3eff, 0xe4b651ff, 
    0xe6ba5aff, 0xe7c16cff, 0xe3c379ff, 0xddbf7aff, 0xc4ab6eff, 0x7d6f4cff, 
    0x6b6961ff, 0xbabbbbff, 0xe3e3e2ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xdedfddff, 0xb1b3afff, 0x8d8e8aff, 0x828481ff, 0x4c4b46ff, 
    0x656051ff, 0x9b9178ff, 0xbcb093ff, 0xd5c8acff, 0xe3d3b1ff, 0xead6acff, 
    0xecd7afff, 0xe2c085ff, 0xd6a451ff, 0xd5a249ff, 0xd8a74dff, 0xdfb569ff, 
    0xe6c58aff, 0xe4c17dff, 0xe1b969ff, 0xe0ba6aff, 0xe5ba67ff, 0xe3b864ff, 
    0xe6c07bff, 0xe7c589ff, 0xe7c889ff, 0xe4c280ff, 0xe0be75ff, 0xe1be76ff, 
    0xe4c37dff, 0xe8c885ff, 0xe7c782ff, 0xe5c681ff, 0xdeba70ff, 0xe1bc72ff, 
    0xe5bf78ff, 0xe1b561ff, 0xdfb25aff, 0xe2b869ff, 0xe0b668ff, 0xdeb25bff, 
    0xdcaf55ff, 0xe2bd7dff, 0xe0bc7fff, 0xdcb366ff, 0xdcb266ff, 0xdcb46bff, 
    0xdab36dff, 0xd8b067ff, 0xd7af68ff, 0xd4ac5fff, 0xd2a750ff, 0xca9835ff, 
    0xc58f26ff, 0xbe8718ff, 0xab6506ff, 0xa86004ff, 0xb16f05ff, 0xbd831aff, 
    0xcb993cff, 0xcea047ff, 0xd3a858ff, 0xd2a654ff, 0xcfa24eff, 0xcc9c42ff, 
    0xd09c37ff, 0xd19b2dff, 0xce9623ff, 0xcd931cff, 0xcf9626ff, 0xcf9528ff, 
    0xcb8a0fff, 0xce8809ff, 0xce8206ff, 0xce7e04ff, 0xd1830cff, 0xd89423ff, 
    0xe1a32fff, 0xe5ab3dff, 0xe4ab45ff, 0xe0a23dff, 0xe7b557ff, 0xe9bb60ff, 
    0xe7b557ff, 0xe6b043ff, 0xe7b342ff, 0xebc164ff, 0xeec974ff, 0xeec96cff, 
    0xebc263ff, 0xe7b64cff, 0xe7b953ff, 0xecc66aff, 0xedc464ff, 0xe9bd5bff, 
    0xe4b44eff, 0xd99f30ff, 0xd89e29ff, 0xdca630ff, 0xe3b54dff, 0xe5b452ff, 
    0xe5b852ff, 0xdead48ff, 0xdfaf4dff, 0xe2b450ff, 0xe3b456ff, 0xe3b350ff, 
    0xe1b347ff, 0xe5b753ff, 0xe7be5dff, 0xe7c578ff, 0xdcbf7aff, 0xac965fff, 
    0x665e48ff, 0x888887ff, 0xd4d5d4ff, 0xe3e4e2ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xdddedcff, 0xb0b2aeff, 0x8c8e89ff, 0x828481ff, 
    0x4c4b46ff, 0x655e4cff, 0x9c937aff, 0xbfb397ff, 0xd8ccb0ff, 0xe7dabdff, 
    0xeddebaff, 0xeddcb3ff, 0xe6cb9bff, 0xd8ad68ff, 0xd8aa63ff, 0xdcb068ff, 
    0xddb365ff, 0xdeb86cff, 0xe0bb76ff, 0xe2c181ff, 0xe1bd75ff, 0xdfb462ff, 
    0xe4b869ff, 0xe6bf75ff, 0xe5c179ff, 0xe1bb73ff, 0xdeb96dff, 0xe0b968ff, 
    0xe4bf71ff, 0xe4bf72ff, 0xe5c27cff, 0xe3c075ff, 0xe6c278ff, 0xe7c680ff, 
    0xe1bc6fff, 0xdeb663ff, 0xe4c07cff, 0xe1b768ff, 0xe0b15cff, 0xe2b870ff, 
    0xdfb464ff, 0xdeb364ff, 0xdeb56fff, 0xe0bc83ff, 0xe1bb80ff, 0xdbb060ff, 
    0xd9af5eff, 0xdab065ff, 0xd7ad60ff, 0xd5ab5fff, 0xd3a85cff, 0xd4ab5eff, 
    0xd0a34bff, 0xc99839ff, 0xc5912aff, 0xbe8618ff, 0xae6b05ff, 0xa75e05ff, 
    0xad6903ff, 0xb87c12ff, 0xc89534ff, 0xd0a249ff, 0xd3a853ff, 0xd2a752ff, 
    0xd2a756ff, 0xcea04dff, 0xce9c40ff, 0xd09b32ff, 0xcd9626ff, 0xcf9727ff, 
    0xce9727ff, 0xcd9426ff, 0xc88810ff, 0xc8860bff, 0xc78308ff, 0xc78307ff, 
    0xc67f05ff, 0xc57803ff, 0xc67804ff, 0xca7c07ff, 0xd18912ff, 0xd59424ff, 
    0xd9992bff, 0xd79521ff, 0xd99c2fff, 0xd7961eff, 0xdb9d21ff, 0xdfa733ff, 
    0xe4b44dff, 0xe9bf65ff, 0xecc572ff, 0xecc774ff, 0xecc66cff, 0xecc971ff, 
    0xeecf7fff, 0xedca7aff, 0xe7b95cff, 0xdfab41ff, 0xdeaa45ff, 0xd79f36ff, 
    0xc88e21ff, 0xd59e32ff, 0xdba73eff, 0xdca944ff, 0xe1b452ff, 0xe6bc62ff, 
    0xe7be68ff, 0xe7bc5eff, 0xe9be5eff, 0xecc366ff, 0xedce83ff, 0xe5ce91ff, 
    0xb2a175ff, 0x5d5a4eff, 0x8f908fff, 0xdedfddff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xdddedcff, 0xafb1adff, 0x8c8e89ff, 
    0x828481ff, 0x4c4c48ff, 0x655f4fff, 0x9d957fff, 0xbfb497ff, 0xdaceb2ff, 
    0xe9dbc0ff, 0xf3e7cfff, 0xf4e5c7ff, 0xefdbb1ff, 0xdfbb7fff, 0xd2a151ff, 
    0xdab16fff, 0xdfba80ff, 0xdfbb7dff, 0xdebb77ff, 0xdaaf5bff, 0xdbb05bff, 
    0xdcb461ff, 0xe0bb6fff, 0xe7c280ff, 0xe7c17eff, 0xe5c27dff, 0xe3bb6cff, 
    0xe1b965ff, 0xe3bc6aff, 0xe7c57eff, 0xe8c67cff, 0xeac881ff, 0xebcd87ff, 
    0xe7c679ff, 0xe6bf6eff, 0xdfb35bff, 0xd8a948ff, 0xdeaf55ff, 0xdfb057ff, 
    0xe2b86bff, 0xe1bb79ff, 0xe0b66aff, 0xdeb369ff, 0xdbb16bff, 0xdcb473ff, 
    0xdab16bff, 0xd8ac5aff, 0xd8ac5aff, 0xd8ad60ff, 0xd5aa5aff, 0xd4a856ff, 
    0xd2a552ff, 0xd1a653ff, 0xce9f45ff, 0xc7932eff, 0xc5912aff, 0xbf881dff, 
    0xb87b0bff, 0xad6805ff, 0xa75d04ff, 0xb3720bff, 0xc38e29ff, 0xcc9e44ff, 
    0xd4a959ff, 0xd4aa57ff, 0xd5aa5aff, 0xcfa251ff, 0xcc9a40ff, 0xd09c3dff, 
    0xcf9b37ff, 0xd0982cff, 0xce9424ff, 0xcb8f19ff, 0xc98a11ff, 0xc9860bff, 
    0xc58005ff, 0xc88510ff, 0xca8a19ff, 0xc8820fff, 0xc88217ff, 0xce8d2bff, 
    0xdca54cff, 0xd49a3fff, 0xd29531ff, 0xcf8e25ff, 0xd9a344ff, 0xc67b18ff, 
    0xc17208ff, 0xc88012ff, 0xc98317ff, 0xcb8721ff, 0xd29229ff, 0xd99e34ff, 
    0xdfab3fff, 0xe4b95aff, 0xe8c16dff, 0xe9c677ff, 0xedcd8aff, 0xebc985ff, 
    0xefcf8dff, 0xe7bf73ff, 0xcc902dff, 0xd09225ff, 0xdca641ff, 0xdfae4eff, 
    0xe5ba62ff, 0xeac171ff, 0xebc475ff, 0xedc776ff, 0xeec872ff, 0xf0d288ff, 
    0xedd69bff, 0xddc990ff, 0x8e8261ff, 0x696a67ff, 0x9d9f9aff, 0xd9dad7ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xdddedcff, 0xafb0adff, 
    0x8c8e89ff, 0x828481ff, 0x4c4c48ff, 0x676255ff, 0x9f9784ff, 0xbeb39aff, 
    0xd7caadff, 0xe8dcc1ff, 0xf2e5cdff, 0xf6ebd3ff, 0xf5e8caff, 0xf2e0baff, 
    0xe3c38dff, 0xd2a152ff, 0xdbb16aff, 0xd9b066ff, 0xdbb468ff, 0xdcb56cff, 
    0xd9af5dff, 0xd5a94fff, 0xd6ae5bff, 0xdfbf7aff, 0xe4c384ff, 0xe6c788ff, 
    0xe9c989ff, 0xe6c37cff, 0xe4be74ff, 0xe7c27bff, 0xe8c377ff, 0xebca80ff, 
    0xedcd87ff, 0xedcb83ff, 0xe5bd66ff, 0xe3b85eff, 0xdfb250ff, 0xd7a236ff, 
    0xdfb154ff, 0xe0b35bff, 0xe0b566ff, 0xe0b66eff, 0xddb165ff, 0xdbae5dff, 
    0xdcb165ff, 0xdbb164ff, 0xdaaf63ff, 0xd6a852ff, 0xd2a449ff, 0xd0a249ff, 
    0xd0a44eff, 0xd1a34eff, 0xce9f43ff, 0xc9983cff, 0xc89637ff, 0xc89637ff, 
    0xc5912eff, 0xc28c26ff, 0xba7f10ff, 0xb27309ff, 0xa55c04ff, 0xac6705ff, 
    0xbd871dff, 0xc89739ff, 0xd1a54fff, 0xd5ab59ff, 0xd6ab5aff, 0xce9f45ff, 
    0xcc973aff, 0xd09d3eff, 0xd09d39ff, 0xd09930ff, 0xcc9322ff, 0xc98b14ff, 
    0xcb901eff, 0xcf9428ff, 0xd59d37ff, 0xe7bc67ff, 0xecc97dff, 0xe8bd6dff, 
    0xeac376ff, 0xebc577ff, 0xebc77eff, 0xe5ba6aff, 0xdfb058ff, 0xd9a644ff, 
    0xdba84eff, 0xce922dff, 0xb76609ff, 0xb15904ff, 0xb15b03ff, 0xb05703ff, 
    0xb15703ff, 0xb75f03ff, 0xbe6c09ff, 0xcc8c29ff, 0xd49c39ff, 0xd49c36ff, 
    0xdbab4cff, 0xe5bd69ff, 0xebc87dff, 0xebc981ff, 0xcc9033ff, 0xce8f24ff, 
    0xdfab4eff, 0xe2b55dff, 0xe8be6dff, 0xeac375ff, 0xecc87fff, 0xedca7bff, 
    0xecc773ff, 0xeed18aff, 0xecd499ff, 0xd7c38cff, 0x968967ff, 0x5f5f5bff, 
    0x92938fff, 0xcdceccff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xdddedbff, 
    0xaeb0acff, 0x8c8e89ff, 0x838481ff, 0x4c4c47ff, 0x676254ff, 0x9f9785ff, 
    0xc0b7a0ff, 0xd8cbaeff, 0xe7d9bbff, 0xf0e2c7ff, 0xf1e2c6ff, 0xf1e1bfff, 
    0xf1e0bcff, 0xedd8afff, 0xdab36eff, 0xd3a452ff, 0xdbaf65ff, 0xd9ae5fff, 
    0xd6a952ff, 0xd7ae5eff, 0xd4b069ff, 0xcea75aff, 0xb3914fff, 0x82704dff, 
    0xb29860ff, 0xd8b871ff, 0xe5c98bff, 0xeacd8dff, 0xe6c076ff, 0xe6be71ff, 
    0xe8c378ff, 0xebca80ff, 0xedce87ff, 0xebc679ff, 0xe8bf68ff, 0xe0b358ff, 
    0xd9a941ff, 0xd7a232ff, 0xe3b767ff, 0xe1b562ff, 0xdeb15eff, 0xdcae5bff, 
    0xdcaf64ff, 0xd9aa5bff, 0xd8ab57ff, 0xd7a958ff, 0xd4a550ff, 0xd3a24bff, 
    0xd0a149ff, 0xd0a24dff, 0xce9f4aff, 0xcf9f49ff, 0xcd9d47ff, 0xca9a42ff, 
    0xc99a41ff, 0xc8973eff, 0xc48f30ff, 0xc08a23ff, 0xb5780aff, 0xb06c05ff, 
    0xa86204ff, 0xaa6605ff, 0xba8016ff, 0xc79333ff, 0xcea148ff, 0xd1a64fff, 
    0xd2a54fff, 0xd9ab51ff, 0xe7bd6cff, 0xe1b45bff, 0xe0b25aff, 0xe3b45aff, 
    0xe0b158ff, 0xe0b35aff, 0xe7c06eff, 0xf0d189ff, 0xf4d995ff, 0xf3d892ff, 
    0xf1d796ff, 0xefd390ff, 0xebc87eff, 0xeac577ff, 0xe8c278ff, 0xe4ba6fff, 
    0xdeb15eff, 0xdaaa52ff, 0xd7a44dff, 0xd09630ff, 0xbe750eff, 0xb25a04ff, 
    0xb45c04ff, 0xba6204ff, 0xb55c04ff, 0xb25605ff, 0xae5504ff, 0xb15605ff, 
    0xb66009ff, 0xbd6b0cff, 0xc6831dff, 0xd29833ff, 0xddac50ff, 0xe6ba65ff, 
    0xcb8929ff, 0xd29633ff, 0xe4b663ff, 0xe6bb67ff, 0xe9c171ff, 0xebc578ff, 
    0xedc97fff, 0xeecc80ff, 0xeecb7bff, 0xf3d795ff, 0xefdbaaff, 0xd6c79cff, 
    0x9f9475ff, 0x575650ff, 0x8b8d89ff, 0xbebfbcff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xdcdddbff, 0xaeafacff, 0x8b8d88ff, 0x838480ff, 0x4b4b48ff, 0x656050ff, 
    0x9d967dff, 0xc0b598ff, 0xdaceb2ff, 0xeadfc7ff, 0xf2e6cfff, 0xf2e4c8ff, 
    0xf1e1beff, 0xf1deb7ff, 0xefd9abff, 0xe3c188ff, 0xd09f4bff, 0xd0a14cff, 
    0xd2a352ff, 0xd7ac60ff, 0xd0a14aff, 0xcaa049ff, 0xc09845ff, 0x9f7c36ff, 
    0x61553dff, 0x6c6d6cff, 0x635c4aff, 0xa68b51ff, 0xd3b56fff, 0xe1c684ff, 
    0xe7c881ff, 0xe8c274ff, 0xe7c071ff, 0xe7c070ff, 0xeac77dff, 0xe8bf6cff, 
    0xe6ba61ff, 0xe0b353ff, 0xd8a535ff, 0xd59f29ff, 0xe1b86bff, 0xe0b460ff, 
    0xddb15fff, 0xdaaa53ff, 0xdaac56ff, 0xdaab5aff, 0xd8aa5bff, 0xd6a859ff, 
    0xcf9f47ff, 0xcf993dff, 0xce9b42ff, 0xcb993cff, 0xcb9738ff, 0xc9973dff, 
    0xc8953aff, 0xc6933aff, 0xc59239ff, 0xc49139ff, 0xc0892aff, 0xbc841bff, 
    0xb37306ff, 0xaf6a04ff, 0xa96404ff, 0xa96704ff, 0xb17307ff, 0xc28b26ff, 
    0xd0a148ff, 0xd7ac57ff, 0xe2b966ff, 0xedc97aff, 0xf2d38aff, 0xf2d48bff, 
    0xf3d68fff, 0xf3d58fff, 0xf4d68fff, 0xf4d78fff, 0xf3d690ff, 0xf3d898ff, 
    0xf3d79aff, 0xf0d596ff, 0xefd295ff, 0xebcb8aff, 0xe6bf6fff, 0xe5bc67ff, 
    0xe4ba6fff, 0xdfb464ff, 0xd7a54aff, 0xd29d3bff, 0xcc9229ff, 0xc58313ff, 
    0xbe7105ff, 0xbe6d05ff, 0xc3780aff, 0xc57f0fff, 0xc0780cff, 0xb96806ff, 
    0xac5405ff, 0xa14104ff, 0xab5204ff, 0xb45e04ff, 0xb66104ff, 0xc07410ff, 
    0xc9841dff, 0xcb821aff, 0xc3780fff, 0xd49834ff, 0xe4b864ff, 0xe8c06fff, 
    0xecc87eff, 0xeecd86ff, 0xf2d48eff, 0xf4d893ff, 0xf5d78eff, 0xf8e0a7ff, 
    0xf1e3bdff, 0xd8ceadff, 0xaaa180ff, 0x55544aff, 0x848582ff, 0xaeafacff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xd2d3d1ff, 0x91928eff, 0x898b87ff, 0x4c4c49ff, 
    0x5d584cff, 0x99917aff, 0xbfb597ff, 0xdacdabff, 0xeadcbcff, 0xf1e5caff, 
    0xf2e3c6ff, 0xf0debbff, 0xf2e2beff, 0xf2e0b4ff, 0xecd5a2ff, 0xd3a75bff, 
    0xcc973fff, 0xd2a355ff, 0xd5a862ff, 0xd0a452ff, 0xc5953aff, 0xb6872cff, 
    0x8b6212ff, 0x594e34ff, 0x7a7c79ff, 0x8f918cff, 0x80827fff, 0x635d4fff, 
    0xa38c58ff, 0xd3b570ff, 0xe2c57fff, 0xe7c67aff, 0xe8c372ff, 0xe7c06dff, 
    0xe9c273ff, 0xe5b85cff, 0xe1b44fff, 0xddaf45ff, 0xdaa93dff, 0xd6a12cff, 
    0xe0b563ff, 0xdfb15dff, 0xddb061ff, 0xdcaf60ff, 0xdbad5bff, 0xd9ab5cff, 
    0xd6a657ff, 0xd5a859ff, 0xd2a34eff, 0xd09e44ff, 0xcc983bff, 0xca953aff, 
    0xcb973bff, 0xca9840ff, 0xc59032ff, 0xc28d2fff, 0xc38e36ff, 0xc0892fff, 
    0xbc831fff, 0xb87d11ff, 0xb27207ff, 0xba7c19ff, 0xb5761aff, 0xae6e0eff, 
    0xbe8625ff, 0xdaad56ff, 0xeccc82ff, 0xf1d692ff, 0xf4da9aff, 0xf1d593ff, 
    0xefd087ff, 0xefd087ff, 0xefcf86ff, 0xefcd7fff, 0xeecd7fff, 0xeecc7fff, 
    0xecc978ff, 0xebc674ff, 0xebc87eff, 0xeac377ff, 0xe8c278ff, 0xe4ba67ff, 
    0xdfb051ff, 0xdcac4cff, 0xdaaa50ff, 0xd39d36ff, 0xcd9221ff, 0xc98c1bff, 
    0xc6840eff, 0xc88512ff, 0xc98514ff, 0xcd8e1fff, 0xcf952cff, 0xcb9025ff, 
    0xc58417ff, 0xbd7309ff, 0xb25f04ff, 0xa74a04ff, 0xa94f04ff, 0xaf5a05ff, 
    0xb96b05ff, 0xc88621ff, 0xcc9034ff, 0xc07408ff, 0xc68312ff, 0xdaa646ff, 
    0xe7be73ff, 0xecc880ff, 0xefd08bff, 0xf2d792ff, 0xf3d996ff, 0xf6dda0ff, 
    0xf8e0a1ff, 0xfae6b0ff, 0xf4eac7ff, 0xddd5afff, 0xb6b08dff, 0x5d5c4fff, 
    0x7a7b78ff, 0x9fa19dff, 0xdddedcff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xdddedcff, 0x9fa09cff, 
    0x878985ff, 0x4a4946ff, 0x807a6cff, 0xb5ad98ff, 0xd8cdb1ff, 0xeee3c5ff, 
    0xf6ead1ff, 0xf6e8caff, 0xf0dfb8ff, 0xf0dcb4ff, 0xf4e7c6ff, 0xf3e5c1ff, 
    0xeacf9cff, 0xd3a553ff, 0xce9c43ff, 0xd2a352ff, 0xcfa253ff, 0xc79844ff, 
    0xb98e3bff, 0x8e6721ff, 0x53452cff, 0x7a7c79ff, 0x8c8e89ff, 0x878984ff, 
    0xa6a8a4ff, 0xb5b6b5ff, 0x6c675aff, 0xa4915fff, 0xd5b873ff, 0xe1c37bff, 
    0xe9cb7eff, 0xe9c677ff, 0xe8c16eff, 0xe5b95dff, 0xe3b756ff, 0xdcad45ff, 
    0xd8a739ff, 0xd6a130ff, 0xe0b262ff, 0xddb05bff, 0xdbaa4fff, 0xdcae5eff, 
    0xdcaf65ff, 0xdbad65ff, 0xd7a859ff, 0xd2a04aff, 0xd3a34eff, 0xd29f48ff, 
    0xcf9d45ff, 0xcb9535ff, 0xc99636ff, 0xca963eff, 0xc58f31ff, 0xc28d2eff, 
    0xc08a2eff, 0xc28b2eff, 0xc79231ff, 0xd6a74dff, 0xdaae5dff, 0xe5bb68ff, 
    0xe5bb67ff, 0xe4c075ff, 0xefd38fff, 0xf2d690ff, 0xf2d58fff, 0xf0d189ff, 
    0xeece85ff, 0xedca7dff, 0xedc770ff, 0xeccb7dff, 0xecca7cff, 0xecc671ff, 
    0xebc571ff, 0xe9c372ff, 0xe9bf68ff, 0xe9bf69ff, 0xe8c06fff, 0xe5bb68ff, 
    0xe2b55fff, 0xdfb259ff, 0xdaab4eff, 0xd8a546ff, 0xd29d35ff, 0xcb8f1bff, 
    0xc8870fff, 0xc8850eff, 0xce9125ff, 0xd7a23fff, 0xd8a648ff, 0xd7a345ff, 
    0xd5a34bff, 0xd29b40ff, 0xca8d25ff, 0xbe770bff, 0xbc6f05ff, 0xbf7506ff, 
    0xbc7107ff, 0xb46104ff, 0xb25e04ff, 0xb86a0fff, 0xc07b17ff, 0xc0780aff, 
    0xc47f0eff, 0xdaa84aff, 0xecc885ff, 0xf4d9a1ff, 0xf4da99ff, 0xf5dc99ff, 
    0xf5dea1ff, 0xf9e4b2ff, 0xfbe7aeff, 0xfce8adff, 0xf5e5b1ff, 0xddc890ff, 
    0xb8a373ff, 0x685e49ff, 0x6d6e6bff, 0x959792ff, 0xd4d5d2ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xb4b5b2ff, 0x8c8e89ff, 0x5c5d5aff, 0x7f7a6dff, 0xc3b9a2ff, 
    0xe2d8b9ff, 0xf5ebcfff, 0xf9eed3ff, 0xfaefd5ff, 0xf7e9c8ff, 0xf5e4c1ff, 
    0xf2e2c0ff, 0xf0dfb7ff, 0xe5c48dff, 0xd0a14cff, 0xcc9c45ff, 0xcca053ff, 
    0xc6994bff, 0xb48430ff, 0x865a17ff, 0x534631ff, 0x7a7c7aff, 0x8c8e89ff, 
    0x878984ff, 0xa3a5a1ff, 0xd7d8d6ff, 0xe4e5e3ff, 0xbcbdbcff, 0x6b6659ff, 
    0xa69360ff, 0xd5b972ff, 0xe0c277ff, 0xe8c778ff, 0xeac574ff, 0xe9c36dff, 
    0xe5bd64ff, 0xdcad47ff, 0xd6a534ff, 0xd29d25ff, 0xdda948ff, 0xdca94dff, 
    0xdbaa55ff, 0xd9ab59ff, 0xdbae64ff, 0xdaae64ff, 0xd6a857ff, 0xd4a452ff, 
    0xd4a453ff, 0xd5a658ff, 0xd0a149ff, 0xcd9838ff, 0xcc9b44ff, 0xca953dff, 
    0xc38b28ff, 0xbb8118ff, 0xc08724ff, 0xd8ab5bff, 0xe7c174ff, 0xefd18cff, 
    0xf0d594ff, 0xeed28aff, 0xefd087ff, 0xf2d692ff, 0xf1d597ff, 0xedcd82ff, 
    0xe9c369ff, 0xeac36dff, 0xeac46fff, 0xeac16aff, 0xe9c26cff, 0xe8c373ff, 
    0xe9c271ff, 0xeac16fff, 0xe9c06cff, 0xe5b95fff, 0xe4b95fff, 0xe4b961ff, 
    0xe2b65dff, 0xdfb359ff, 0xdcad50ff, 0xdaa745ff, 0xd49f34ff, 0xd0992cff, 
    0xcb8f18ff, 0xca890dff, 0xd0921eff, 0xdaa441ff, 0xe2b566ff, 0xe2b767ff, 
    0xe1b569ff, 0xdeb163ff, 0xd9a954ff, 0xd5a14bff, 0xd19b45ff, 0xc5841bff, 
    0xca8a16ff, 0xce9023ff, 0xcb8e24ff, 0xc48214ff, 0xbc750dff, 0xb36204ff, 
    0xb05d03ff, 0xb66904ff, 0xbe7809ff, 0xd5a043ff, 0xefcc8fff, 0xf4db9cff, 
    0xf6dc99ff, 0xf9e4a9ff, 0xfbe8b3ff, 0xfbe9b5ff, 0xf8de9cff, 0xf4d184ff, 
    0xebbd66ff, 0xce993aff, 0xa77121ff, 0x694b18ff, 0x5f5e59ff, 0x8f918cff, 
    0xc6c8c5ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xcacbc8ff, 0x8f918cff, 0x70716fff, 
    0x6e6a5fff, 0xc5bca5ff, 0xe2d6b4ff, 0xf5ebcdff, 0xfcf2d7ff, 0xf6e9c4ff, 
    0xf6e9c5ff, 0xf6e8c4ff, 0xf1deb6ff, 0xe9cd99ff, 0xd5a95dff, 0xcb973bff, 
    0xc7953eff, 0xc49543ff, 0xb78a39ff, 0x83530fff, 0x4f3e29ff, 0x7a7b79ff, 
    0x8c8e89ff, 0x878984ff, 0xa3a4a0ff, 0xd7d8d6ff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe4e5e3ff, 0xbcbdbcff, 0x6b6657ff, 0xa79056ff, 0xd7ba6fff, 0xe2c172ff, 
    0xe8c676ff, 0xe9c673ff, 0xe2b95bff, 0xdcae4bff, 0xd7a739ff, 0xcf981aff, 
    0xd8a33fff, 0xdaa74eff, 0xd9a853ff, 0xdbad5eff, 0xd8ab5fff, 0xd7a95bff, 
    0xd8a85cff, 0xd5a754ff, 0xd09d47ff, 0xd3a45aff, 0xd4a755ff, 0xd5a554ff, 
    0xd09e4cff, 0xc99237ff, 0xc69035ff, 0xc0851fff, 0xd39e3cff, 0xe6bd71ff, 
    0xeac982ff, 0xe9c87eff, 0xe9c77dff, 0xe7c373ff, 0xe6c06dff, 0xe4bb5fff, 
    0xe4bb66ff, 0xe5bc64ff, 0xe4ba5dff, 0xe5bb64ff, 0xe5bc65ff, 0xe6bd69ff, 
    0xe2b65aff, 0xe2b658ff, 0xe3b75fff, 0xe3b75fff, 0xe2b55bff, 0xe1b55bff, 
    0xdfb156ff, 0xdbab4bff, 0xdaa947ff, 0xd7a43dff, 0xd4a036ff, 0xd19a29ff, 
    0xcd921aff, 0xcb8e16ff, 0xce911aff, 0xd9a239ff, 0xe1b155ff, 0xe3b762ff, 
    0xe6bf79ff, 0xe5bf7aff, 0xe2bc77ff, 0xdeb36aff, 0xd9aa55ff, 0xd6a44bff, 
    0xd3a14eff, 0xcb9132ff, 0xc98c1cff, 0xd59e3eff, 0xd8a64bff, 0xd69e38ff, 
    0xcf942cff, 0xc38014ff, 0xba7007ff, 0xb36504ff, 0xb36603ff, 0xc98d2fff, 
    0xedc67aff, 0xf2d287ff, 0xf6da95ff, 0xf5d58fff, 0xedc87eff, 0xe5b566ff, 
    0xe7b45bff, 0xe7b558ff, 0xdaa649ff, 0xb5761bff, 0x9f6916ff, 0x734f0cff, 
    0x524e44ff, 0x8b8d89ff, 0xb6b7b4ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xdadbd8ff, 
    0x9a9b97ff, 0x81837fff, 0x5c584eff, 0xb7ab8eff, 0xdccca6ff, 0xeedfb6ff, 
    0xf7eac5ff, 0xf8eac3ff, 0xf5e5bdff, 0xf1e0b5ff, 0xefd9abff, 0xddb571ff, 
    0xcb9943ff, 0xc7973eff, 0xc18f38ff, 0xb4832eff, 0x8b621cff, 0x52432cff, 
    0x7a7c79ff, 0x8c8e89ff, 0x878984ff, 0xa2a4a0ff, 0xd7d8d6ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe4e5e3ff, 0xbdbebdff, 0x6e6652ff, 
    0xb89e5eff, 0xddbc6aff, 0xe6c472ff, 0xe9c673ff, 0xe1b859ff, 0xdbad46ff, 
    0xd5a534ff, 0xd09b1eff, 0xd69f3fff, 0xd39c3bff, 0xd6a048ff, 0xdcad62ff, 
    0xd9ab5eff, 0xd4a14aff, 0xd5a24cff, 0xd3a24bff, 0xd29f4aff, 0xd2a152ff, 
    0xd5a75aff, 0xd3a556ff, 0xcc983bff, 0xc99130ff, 0xc99136ff, 0xd6a243ff, 
    0xe5ba65ff, 0xe7c276ff, 0xe5bf6fff, 0xe4bc69ff, 0xe3b861ff, 0xddb151ff, 
    0xdbae4eff, 0xd9aa47ff, 0xd8a744ff, 0xdaab4cff, 0xdeb156ff, 0xddb04eff, 
    0xddb04cff, 0xdbad50ff, 0xd8a847ff, 0xd7a742ff, 0xd6a43cff, 0xd5a032ff, 
    0xd39e34ff, 0xd29f36ff, 0xd19b2dff, 0xcf9927ff, 0xcf9623ff, 0xce941fff, 
    0xce901aff, 0xcb8d16ff, 0xc9880cff, 0xd2982dff, 0xe1b45bff, 0xe7be6aff, 
    0xe8c57fff, 0xe9c886ff, 0xe7c587ff, 0xe6c286ff, 0xe4c080ff, 0xdfb670ff, 
    0xd8a64eff, 0xd49e3bff, 0xd3a146ff, 0xcd973aff, 0xc08114ff, 0xc28325ff, 
    0xbc7a26ff, 0xc18331ff, 0xd5a146ff, 0xdaa746ff, 0xce942bff, 0xc58115ff, 
    0xc07a15ff, 0xbf7912ff, 0xd6a13cff, 0xe8bc5eff, 0xebc16aff, 0xe4b351ff, 
    0xd79a32ff, 0xcd8b25ff, 0xca8c28ff, 0xbf7a1cff, 0xb3670dff, 0xa65704ff, 
    0x944f03ff, 0x784803ff, 0x4b412fff, 0x868885ff, 0xa6a7a3ff, 0xe0e1dfff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xadaeaaff, 0x8a8c87ff, 0x595853ff, 0xa0957bff, 
    0xd5c69fff, 0xebdbaeff, 0xf3e1b3ff, 0xf8e9c0ff, 0xf6e7beff, 0xf1ddafff, 
    0xe8ca94ff, 0xcc9b4dff, 0xbf8b34ff, 0xb48024ff, 0xa6721aff, 0x80510cff, 
    0x50412bff, 0x7a7c79ff, 0x8c8e89ff, 0x878984ff, 0xa2a4a0ff, 0xd7d8d6ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe3e4e3ff, 0x817b70ff, 0x927129ff, 0xd5b057ff, 0xe4be66ff, 0xe5be64ff, 
    0xdfb553ff, 0xdbae47ff, 0xd5a432ff, 0xd09a1cff, 0xd29835ff, 0xd39a36ff, 
    0xd7a34fff, 0xd6a048ff, 0xd8a450ff, 0xd5a14cff, 0xd29c3fff, 0xd49f45ff, 
    0xd29d42ff, 0xd09d41ff, 0xce993cff, 0xcb9535ff, 0xcd9738ff, 0xc99233ff, 
    0xd29c3cff, 0xe4ba63ff, 0xe6bf71ff, 0xe4bd6fff, 0xe0b761ff, 0xddb256ff, 
    0xddb055ff, 0xd7a645ff, 0xd6a542ff, 0xd9a747ff, 0xd8a543ff, 0xd4a138ff, 
    0xd9a84aff, 0xd7a53bff, 0xd5a339ff, 0xd4a13cff, 0xd4a039ff, 0xd3a03aff, 
    0xd09a30ff, 0xd09a2eff, 0xcc9529ff, 0xc88e1cff, 0xc48410ff, 0xc38614ff, 
    0xc78811ff, 0xc9890cff, 0xc7850aff, 0xc8860aff, 0xd19624ff, 0xe3b866ff, 
    0xe8c275ff, 0xe8c67eff, 0xeaca8eff, 0xe8c783ff, 0xe7c584ff, 0xe8c68cff, 
    0xe8c58dff, 0xe1b873ff, 0xdaaa53ff, 0xd7a445ff, 0xd6a349ff, 0xcf993dff, 
    0xc08015ff, 0xb66d03ff, 0xa75303ff, 0x953904ff, 0x9e4d13ff, 0xbe8335ff, 
    0xd39f49ff, 0xdcab4eff, 0xdbaa4eff, 0xd1962fff, 0xcd9128ff, 0xdaa742ff, 
    0xdaa740ff, 0xcf962cff, 0xc8891fff, 0xbc7409ff, 0xb66804ff, 0xb56703ff, 
    0xb26404ff, 0xaa5d05ff, 0x974f04ff, 0x7a4102ff, 0x47331cff, 0x7d807dff, 
    0x999b96ff, 0xd9dad8ff, 0xe2e2e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xc3c4c1ff, 0x8e908bff, 
    0x676865ff, 0x837d6eff, 0xcdbf9bff, 0xe6d7acff, 0xf5e8c1ff, 0xf9ebc5ff, 
    0xf2deb1ff, 0xe4c383ff, 0xc99945ff, 0xb47819ff, 0xaf7411ff, 0xa56c0cff, 
    0x835509ff, 0x51412aff, 0x7a7c79ff, 0x8c8e89ff, 0x878984ff, 0xa2a39fff, 
    0xd7d8d6ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe4e5e3ff, 0xb3b3b1ff, 0x5a401eff, 0x834d06ff, 0xb98d35ff, 
    0xe0b856ff, 0xe4b956ff, 0xdeb24aff, 0xd8aa3cff, 0xd2a02cff, 0xd09b25ff, 
    0xcc8b1bff, 0xd0942aff, 0xd49d3cff, 0xd29a3aff, 0xd59d44ff, 0xd19a3aff, 
    0xd29b3cff, 0xd49d43ff, 0xd3a047ff, 0xd29e42ff, 0xcf9b41ff, 0xcd973cff, 
    0xcc9842ff, 0xce983bff, 0xdeb058ff, 0xe2b967ff, 0xe2b968ff, 0xdfb663ff, 
    0xdcb158ff, 0xd8a94aff, 0xdaac51ff, 0xdcb15dff, 0xd5a240ff, 0xd8a74aff, 
    0xd8a747ff, 0xd4a33eff, 0xd7a544ff, 0xd6a33eff, 0xd3a03aff, 0xd4a13cff, 
    0xd4a139ff, 0xd4a13aff, 0xd5a141ff, 0xd39f3fff, 0xcf9832ff, 0xcc922aff, 
    0xca8e29ff, 0xcb912dff, 0xc78a20ff, 0xc5840dff, 0xd49b2fff, 0xdda840ff, 
    0xe2b34eff, 0xe8be69ff, 0xe8c375ff, 0xe9c988ff, 0xedcf99ff, 0xebcb91ff, 
    0xeccc8dff, 0xe9c88cff, 0xe7c486ff, 0xe3bb74ff, 0xddad59ff, 0xd4a142ff, 
    0xd6a54eff, 0xd09b3eff, 0xc88e28ff, 0xbc7a10ff, 0xaf6205ff, 0x994004ff, 
    0x8c2f03ff, 0x903a03ff, 0x98510dff, 0xad7420ff, 0xc59030ff, 0xd49e31ff, 
    0xd6a136ff, 0xd5a138ff, 0xd4a035ff, 0xd29b2fff, 0xc48212ff, 0xbd7404ff, 
    0xbc7104ff, 0xba6f05ff, 0xb56b04ff, 0xaf6504ff, 0x9f5a04ff, 0x834802ff, 
    0x4b2e0fff, 0x717270ff, 0x91938eff, 0xcecfcdff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xd5d6d4ff, 0x969793ff, 0x797b77ff, 0x625f56ff, 0xbab095ff, 0xd9ceabff, 
    0xe9ddb9ff, 0xe8d4aaff, 0xd5ae6aff, 0xbd862aff, 0xb87f21ff, 0xb47d20ff, 
    0xa9771cff, 0x8a631aff, 0x564a32ff, 0x7a7c79ff, 0x8c8e89ff, 0x878984ff, 
    0xa1a39fff, 0xd6d7d5ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe3e4e2ff, 0xd1d2d2ff, 0x5e5344ff, 0x623601ff, 
    0x834703ff, 0x975909ff, 0xbf8d2eff, 0xdcaf44ff, 0xddaf44ff, 0xd8aa3dff, 
    0xd0a02bff, 0xce9d2dff, 0xca8612ff, 0xcc8c1cff, 0xce932bff, 0xce9431ff, 
    0xcf932dff, 0xcf932dff, 0xd2983bff, 0xd09731ff, 0xcf9837ff, 0xd29d44ff, 
    0xcf993aff, 0xcc9533ff, 0xca9230ff, 0xcc9734ff, 0xdaad55ff, 0xdfb462ff, 
    0xdfb568ff, 0xddb261ff, 0xd7a94dff, 0xd39f3bff, 0xd4a040ff, 0xd7a74eff, 
    0xd5a241ff, 0xd6a545ff, 0xd8a94cff, 0xd9ab51ff, 0xd9a847ff, 0xdaaa4dff, 
    0xd5a243ff, 0xd39e3bff, 0xd4a340ff, 0xd5a340ff, 0xd59f3cff, 0xd49f3bff, 
    0xd09c38ff, 0xd29d42ff, 0xd29d49ff, 0xce963eff, 0xcd9334ff, 0xdaa73dff, 
    0xeac368ff, 0xeec870ff, 0xecc874ff, 0xeac674ff, 0xe9c67aff, 0xeac989ff, 
    0xeac992ff, 0xe7c686ff, 0xe9c988ff, 0xe9c889ff, 0xe3bb6dff, 0xe1b55eff, 
    0xdcae52ff, 0xd7a342ff, 0xd7a54bff, 0xd4a145ff, 0xcf983bff, 0xc48820ff, 
    0xba770dff, 0xa95b04ff, 0x964004ff, 0x8a3504ff, 0x894803ff, 0x874602ff, 
    0x904706ff, 0xab6410ff, 0xc88f23ff, 0xd09b2cff, 0xd29e2cff, 0xd49f30ff, 
    0xd39e30ff, 0xc88916ff, 0xc17b05ff, 0xbe7604ff, 0xbb7004ff, 0xb46a04ff, 
    0xa45f04ff, 0x8b5003ff, 0x563307ff, 0x62615dff, 0x8d8f8aff, 0xbfc0bdff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xdfe0deff, 0xa6a7a3ff, 0x888985ff, 0x545450ff, 
    0x686252ff, 0x938568ff, 0xb19a6cff, 0xb28942ff, 0xab7619ff, 0xad7818ff, 
    0xaf7c22ff, 0xaa7c26ff, 0x8a6623ff, 0x554a32ff, 0x7a7c79ff, 0x8c8e89ff, 
    0x878984ff, 0xa1a39fff, 0xd6d7d5ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e1ff, 0x807d78ff, 
    0x4a2b06ff, 0x713f03ff, 0x874b04ff, 0x925103ff, 0x9e5d09ff, 0xc38d21ff, 
    0xd7a32cff, 0xd5a434ff, 0xd19f2eff, 0xcd9821ff, 0xcb8a20ff, 0xcb891cff, 
    0xd19534ff, 0xd1993eff, 0xcf9431ff, 0xd29838ff, 0xd19939ff, 0xd19836ff, 
    0xd49f43ff, 0xd29e46ff, 0xcd9635ff, 0xca922eff, 0xc78e29ff, 0xc68d2aff, 
    0xd09d3fff, 0xdbae5bff, 0xdbaf5dff, 0xd9ac58ff, 0xd6a64bff, 0xd29f3cff, 
    0xcf9837ff, 0xd09a38ff, 0xd5a448ff, 0xd5a445ff, 0xd9a84cff, 0xd7a746ff, 
    0xdaac52ff, 0xdbad58ff, 0xdaad5cff, 0xd9aa58ff, 0xdbae5aff, 0xd8a84cff, 
    0xd6a443ff, 0xd5a346ff, 0xd3a248ff, 0xd6a552ff, 0xd4a14fff, 0xd39c44ff, 
    0xd59e32ff, 0xe5b850ff, 0xf0cc78ff, 0xefd080ff, 0xeccb75ff, 0xeac773ff, 
    0xe8c575ff, 0xeac885ff, 0xe7c583ff, 0xe5c075ff, 0xe9c781ff, 0xe7c682ff, 
    0xe6bf74ff, 0xe2b763ff, 0xdaaa4cff, 0xd9a847ff, 0xd9a950ff, 0xd7a654ff, 
    0xd09c40ff, 0xc9902aff, 0xbe8013ff, 0xb16605ff, 0xa14d05ff, 0x913404ff, 
    0x8c3f04ff, 0x884204ff, 0x914104ff, 0x974503ff, 0x9e5206ff, 0xaf680eff, 
    0xc58716ff, 0xcd9219ff, 0xd49d29ff, 0xd29d2eff, 0xcb8d19ff, 0xc37d05ff, 
    0xc27905ff, 0xbc7705ff, 0xac6805ff, 0x945903ff, 0x643b03ff, 0x535048ff, 
    0x8b8d89ff, 0xaeb0acff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xbcbdbaff, 
    0x8a8b87ff, 0x868884ff, 0x757674ff, 0x63625eff, 0x575041ff, 0x594622ff, 
    0x6d4e16ff, 0x835d12ff, 0x916716ff, 0x846124ff, 0x564c39ff, 0x7a7c79ff, 
    0x8c8e89ff, 0x878984ff, 0xa1a39fff, 0xd6d7d5ff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe4e5e3ff, 
    0xaaabaaff, 0x4b3d24ff, 0x6a4004ff, 0x834b03ff, 0x965604ff, 0x9e5c04ff, 
    0x9e5904ff, 0xa66407ff, 0xcb9314ff, 0xd19c22ff, 0xd2a235ff, 0xd09e2dff, 
    0xc68213ff, 0xc9881eff, 0xcf9231ff, 0xd09638ff, 0xd09634ff, 0xd39f45ff, 
    0xd09732ff, 0xd19a35ff, 0xd6a34aff, 0xd09b3eff, 0xcd9739ff, 0xcc9432ff, 
    0xc8902aff, 0xc1871dff, 0xcb9230ff, 0xd7a74dff, 0xd7a84fff, 0xd8ab55ff, 
    0xd9ab55ff, 0xcf9c38ff, 0xd19e43ff, 0xd29d40ff, 0xd7a64fff, 0xd7a649ff, 
    0xd9a950ff, 0xd9a84fff, 0xdaad59ff, 0xd9ac57ff, 0xdaac5aff, 0xd9aa55ff, 
    0xdaaa52ff, 0xdbac58ff, 0xd8a952ff, 0xd8a953ff, 0xd6a755ff, 0xd6a450ff, 
    0xd8a656ff, 0xd39c3fff, 0xdaa031ff, 0xecc268ff, 0xf0d083ff, 0xeece78ff, 
    0xeecf7bff, 0xedcd7aff, 0xeac774ff, 0xe9c679ff, 0xe6c272ff, 0xe7c273ff, 
    0xe9c67dff, 0xe9c780ff, 0xe7c278ff, 0xdfb35bff, 0xdeae52ff, 0xdfb35eff, 
    0xdbad57ff, 0xd9ab5bff, 0xd3a045ff, 0xcc942fff, 0xc3881eff, 0xb77209ff, 
    0xaf6304ff, 0x9d4404ff, 0x943804ff, 0x8f3904ff, 0x974304ff, 0x9c4e04ff, 
    0xa05304ff, 0xbe7f20ff, 0xc78c24ff, 0xb16605ff, 0xbc7b10ff, 0xd09922ff, 
    0xd19a28ff, 0xcb8a0eff, 0xc78207ff, 0xc17e05ff, 0xb57405ff, 0x9d6204ff, 
    0x734601ff, 0x4a4132ff, 0x888a87ff, 0x9fa09cff, 0xdddedcff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xd7d8d6ff, 0xb1b2afff, 0x9c9e9aff, 0x91938eff, 0x8c8e8aff, 
    0x888b87ff, 0x7e807eff, 0x6c6c69ff, 0x5a564dff, 0x524731ff, 0x4f4738ff, 
    0x7b7c79ff, 0x8c8e89ff, 0x878984ff, 0xa1a29eff, 0xd6d7d5ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe3e4e2ff, 0xcdcdcdff, 0x58544bff, 0x6e582cff, 0x8f671dff, 0x965d05ff, 
    0x9f5e04ff, 0x9e5c04ff, 0x9f5c04ff, 0x9c5604ff, 0xb57a0cff, 0xcc9313ff, 
    0xd09c29ff, 0xddb65cff, 0xcb8e2bff, 0xc78922ff, 0xc8891fff, 0xcf9535ff, 
    0xd19938ff, 0xd7a551ff, 0xd29e41ff, 0xd29e3dff, 0xd4a145ff, 0xd19c41ff, 
    0xce963cff, 0xcd9536ff, 0xc8912eff, 0xc0851bff, 0xc28720ff, 0xcb9530ff, 
    0xcf9933ff, 0xd6a54bff, 0xd7a854ff, 0xd19d3bff, 0xd5a54eff, 0xd4a44aff, 
    0xd4a142ff, 0xd9a94fff, 0xdeb366ff, 0xdcb062ff, 0xdcaf5eff, 0xdbae5bff, 
    0xdbad5bff, 0xdaae5aff, 0xdcb061ff, 0xd9aa55ff, 0xd9aa56ff, 0xd7a852ff, 
    0xd5a659ff, 0xd3a14aff, 0xd4a046ff, 0xd7a041ff, 0xe2ac43ff, 0xf0cc7cff, 
    0xf2d488ff, 0xf1d481ff, 0xf3d992ff, 0xf0d386ff, 0xeed082ff, 0xeccc80ff, 
    0xe8c271ff, 0xe8c67eff, 0xeaca86ff, 0xebcc8dff, 0xe4bd6dff, 0xe0b153ff, 
    0xe4ba65ff, 0xe2b967ff, 0xddb059ff, 0xddb265ff, 0xd3a246ff, 0xcd9936ff, 
    0xc78e28ff, 0xba760bff, 0xb36d05ff, 0xab6005ff, 0x9d4304ff, 0x974004ff, 
    0x9b4804ff, 0xa35604ff, 0xb16b0cff, 0xd09c35ff, 0xd09b33ff, 0xb26708ff, 
    0x994b03ff, 0xab660cff, 0xcb921fff, 0xcf941cff, 0xcb8b0eff, 0xc68407ff, 
    0xbd7e07ff, 0xa76e06ff, 0x825302ff, 0x483920ff, 0x808380ff, 0x939591ff, 
    0xd5d6d3ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xdcdddaff, 
    0xcecfcdff, 0xbabcb8ff, 0xa5a7a3ff, 0x959792ff, 0x8e908bff, 0x8b8d89ff, 
    0x848683ff, 0x858784ff, 0x8c8d89ff, 0x878984ff, 0xa1a29eff, 0xd6d7d4ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xdfe0deff, 0x797875ff, 0x5a4b30ff, 0x947a49ff, 
    0xad863aff, 0xb7872aff, 0xb27815ff, 0xa15e04ff, 0xa05d04ff, 0x9b5704ff, 
    0xa26207ff, 0xbe800aff, 0xc78d15ff, 0xdcb65bff, 0xcc942fff, 0xcf9737ff, 
    0xce9739ff, 0xcf973cff, 0xd49f4aff, 0xd7a54eff, 0xd6a54fff, 0xd39f43ff, 
    0xd3a148ff, 0xd4a34dff, 0xcf993cff, 0xcd973aff, 0xcb9336ff, 0xc68c2cff, 
    0xbe801aff, 0xc68e27ff, 0xcf9a38ff, 0xd6a54fff, 0xd29c3aff, 0xd3a246ff, 
    0xd9ab5dff, 0xd9ab5bff, 0xd8a74fff, 0xdaab55ff, 0xdcb164ff, 0xdcb164ff, 
    0xdaae5eff, 0xdaae62ff, 0xdaad5eff, 0xd8ac58ff, 0xd8aa56ff, 0xd4a346ff, 
    0xd5a449ff, 0xd7a756ff, 0xd7a85dff, 0xd4a24dff, 0xd19d3dff, 0xd39b35ff, 
    0xe4b046ff, 0xf1cb77ff, 0xf3d68bff, 0xf4da8cff, 0xf4da90ff, 0xf3d68cff, 
    0xf3d993ff, 0xf0d58dff, 0xefd088ff, 0xedd08bff, 0xedcf8cff, 0xeac881ff, 
    0xe8c271ff, 0xe8c173ff, 0xe9c37cff, 0xe5bd70ff, 0xe0b562ff, 0xdbaf5eff, 
    0xd4a248ff, 0xcf9c39ff, 0xca9332ff, 0xbc7a10ff, 0xb46b05ff, 0xb06805ff, 
    0xa35005ff, 0x9c4704ff, 0xa35504ff, 0xab6405ff, 0xc38926ff, 0xddb156ff, 
    0xdbac4bff, 0xc28116ff, 0x9d4f03ff, 0x964804ff, 0xa25b09ff, 0xc68c1fff, 
    0xcf9619ff, 0xcb8e10ff, 0xc1850dff, 0xac750aff, 0x895a04ff, 0x4b3611ff, 
    0x747673ff, 0x8d8f8aff, 0xc7c9c6ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xdfe0deff, 0xd6d7d5ff, 
    0xc5c6c3ff, 0xafb1adff, 0x9c9e99ff, 0x8f918cff, 0x888a85ff, 0xa0a29eff, 
    0xd6d7d4ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe4e5e3ff, 0xa5a5a5ff, 0x4c4434ff, 
    0x836c41ff, 0xa98e57ff, 0xbc984cff, 0xc89f48ff, 0xc3912eff, 0xa76706ff, 
    0x9f5c04ff, 0x9a5704ff, 0x995604ff, 0xae6a04ff, 0xc08214ff, 0xddb65eff, 
    0xcc9430ff, 0xd29d3cff, 0xd3a04aff, 0xd39e4aff, 0xd5a452ff, 0xd6a651ff, 
    0xd8a956ff, 0xd29e44ff, 0xd09c41ff, 0xcf9c3eff, 0xcd9a39ff, 0xcf9f4eff, 
    0xcd9946ff, 0xc58d2fff, 0xbf8321ff, 0xc18622ff, 0xca922bff, 0xd4a44bff, 
    0xd3a346ff, 0xd5a449ff, 0xd6a54aff, 0xd8aa55ff, 0xdbaf5bff, 0xdaad5aff, 
    0xd7a64cff, 0xd8aa56ff, 0xd5a246ff, 0xd7a754ff, 0xdbae61ff, 0xd9ac5cff, 
    0xdbaf67ff, 0xd6a859ff, 0xd2a045ff, 0xd5a450ff, 0xd19f47ff, 0xcd973bff, 
    0xca9435ff, 0xcf952bff, 0xe2ac3aff, 0xeec76cff, 0xf7df98ff, 0xfae7a7ff, 
    0xf7e093ff, 0xf7dd91ff, 0xf7de9aff, 0xf4d991ff, 0xf5dfa0ff, 0xf3d999ff, 
    0xf1d591ff, 0xedcd82ff, 0xedcc84ff, 0xebca87ff, 0xebc985ff, 0xe6c37cff, 
    0xe3bc73ff, 0xdeb163ff, 0xd7a751ff, 0xd2a144ff, 0xca9331ff, 0xbf8012ff, 
    0xb76f05ff, 0xb26b05ff, 0xab5f04ff, 0xa35004ff, 0xa95f04ff, 0xae6a07ff, 
    0xcc983dff, 0xe6c177ff, 0xe5bd6cff, 0xd29e39ff, 0xa45b0aff, 0x974903ff, 
    0x974904ff, 0xa05506ff, 0xbc7d13ff, 0xcb921aff, 0xc18a17ff, 0xac7b15ff, 
    0x7f580cff, 0x45371dff, 0x787977ff, 0x8a8c87ff, 0xb9bab7ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xdbdcdaff, 0xcecfccff, 
    0xbebfbcff, 0xd5d6d4ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe3e4e2ff, 0xc9cac9ff, 
    0x55534dff, 0x716144ff, 0x9e844fff, 0xbb9d5fff, 0xc9a660ff, 0xcfa659ff, 
    0xca9c3fff, 0xbb831eff, 0xa36106ff, 0x9e5a04ff, 0x9b5804ff, 0xa75f02ff, 
    0xbf8427ff, 0xe0c071ff, 0xd29d41ff, 0xd5a34cff, 0xd3a249ff, 0xd19e42ff, 
    0xd5a54dff, 0xd3a246ff, 0xcd9836ff, 0xcd9735ff, 0xcc993fff, 0xcf9b41ff, 
    0xd4a34eff, 0xd2a551ff, 0xcb973aff, 0xc68f2eff, 0xc28824ff, 0xc18725ff, 
    0xc58b29ff, 0xd19f45ff, 0xd2a142ff, 0xd2a142ff, 0xd3a142ff, 0xd6a546ff, 
    0xd8a94fff, 0xd8a84eff, 0xd5a649ff, 0xd5a549ff, 0xcf9832ff, 0xcf9a3dff, 
    0xd7a856ff, 0xd9ac5bff, 0xd9ab5eff, 0xd19f4bff, 0xcf9a40ff, 0xd09d48ff, 
    0xc99136ff, 0xc28724ff, 0xc28422ff, 0xcb8f24ff, 0xe2a832ff, 0xefc668ff, 
    0xf9e2a2ff, 0xfbe8aeff, 0xfce9a7ff, 0xfbe8a6ff, 0xf9e29eff, 0xf8e1a0ff, 
    0xf6e0a6ff, 0xf4dc9eff, 0xf3d998ff, 0xf2d596ff, 0xeed18cff, 0xeecf8cff, 
    0xebca85ff, 0xe8c680ff, 0xe2bb72ff, 0xdeb15dff, 0xdcae5cff, 0xd6a753ff, 
    0xc9912bff, 0xc18314ff, 0xba7507ff, 0xb46d04ff, 0xae6404ff, 0xa65405ff, 
    0xaa5e04ff, 0xad6705ff, 0xc68e31ff, 0xe5be6fff, 0xe7c278ff, 0xdeb156ff, 
    0xb97b22ff, 0x984a03ff, 0x994b04ff, 0x994b04ff, 0x994d04ff, 0xaa690cff, 
    0xb98519ff, 0x926a13ff, 0x4e4123ff, 0x6d6e6cff, 0x8b8d89ff, 0x999b96ff, 
    0xcfd0ceff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xdddeddff, 0x767674ff, 0x554a34ff, 0x947e54ff, 0xbaa06eff, 0xccae74ff, 
    0xd2b070ff, 0xd6b372ff, 0xcca04aff, 0xc38f2aff, 0xb27511ff, 0xa46307ff, 
    0x9c5503ff, 0xa7600dff, 0xd2a758ff, 0xe3c880ff, 0xcf9937ff, 0xd3a145ff, 
    0xd3a248ff, 0xd19f42ff, 0xd7a64eff, 0xd4a44aff, 0xd09f43ff, 0xcd9b3dff, 
    0xce9c42ff, 0xcf9f4fff, 0xd09f4dff, 0xcfa14cff, 0xce9b44ff, 0xc99336ff, 
    0xc68e30ff, 0xc38929ff, 0xc38b2cff, 0xcc9b42ff, 0xd1a045ff, 0xd09c39ff, 
    0xd19f3fff, 0xd3a245ff, 0xd3a242ff, 0xd4a140ff, 0xd2a03dff, 0xd29e3aff, 
    0xd2a046ff, 0xc78f27ff, 0xcd9736ff, 0xd29f45ff, 0xd3a34eff, 0xd19e4bff, 
    0xcb9541ff, 0xc88f3aff, 0xca9540ff, 0xc18320ff, 0xbb7812ff, 0xc7871fff, 
    0xe6ad45ff, 0xf2c970ff, 0xfbe7a6ff, 0xfbe9aaff, 0xfcedb4ff, 0xfceeb9ff, 
    0xfbe9b1ff, 0xf8e19fff, 0xf6df9aff, 0xf6dfa2ff, 0xf4da97ff, 0xf3d893ff, 
    0xf3d995ff, 0xf3d89bff, 0xedcf91ff, 0xeac888ff, 0xe7c27eff, 0xe0b565ff, 
    0xdeb362ff, 0xd9ac5aff, 0xcc9434ff, 0xc08111ff, 0xbe7c0dff, 0xb9740aff, 
    0xb16604ff, 0xab5c05ff, 0xab6004ff, 0xad6704ff, 0xb5750fff, 0xdaad57ff, 
    0xe6c173ff, 0xe6bf6eff, 0xd6a648ff, 0xa45b0bff, 0x954903ff, 0x944804ff, 
    0x8e4404ff, 0x8d4704ff, 0x8b5c0eff, 0x554728ff, 0x717270ff, 0x8c8e89ff, 
    0x9c9d99ff, 0xd1d2cfff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe3e4e2ff, 0xa2a3a2ff, 0x484236ff, 0x816d48ff, 0xac9466ff, 
    0xc6a973ff, 0xd1af70ff, 0xd1ab62ff, 0xca9f49ff, 0xc49331ff, 0xc49230ff, 
    0xc38d2aff, 0xbf8720ff, 0xa9660bff, 0xbc8a43ff, 0xe7cf93ff, 0xe5ca8cff, 
    0xd4a043ff, 0xdaad5fff, 0xd9ac5dff, 0xd6a852ff, 0xd8ab54ff, 0xd4a64bff, 
    0xd4a852ff, 0xd3a54fff, 0xd3a551ff, 0xcfa14eff, 0xcfa24fff, 0xcd9e4aff, 
    0xcb9b49ff, 0xc79339ff, 0xc38b2cff, 0xc28a2cff, 0xc1892aff, 0xc9963dff, 
    0xd1a34fff, 0xcd9835ff, 0xd1a045ff, 0xd3a552ff, 0xd3a54eff, 0xd29f41ff, 
    0xd2a040ff, 0xd2a042ff, 0xd09e3fff, 0xc9932dff, 0xc18313ff, 0xc68b24ff, 
    0xcc973bff, 0xce9b42ff, 0xcc994aff, 0xca9343ff, 0xc99243ff, 0xbd7d21ff, 
    0xad6305ff, 0xc47e18ff, 0xecb54cff, 0xf7d482ff, 0xfce8abff, 0xfbe9b0ff, 
    0xfbeab4ff, 0xfcebbaff, 0xfbeab6ff, 0xfae4a4ff, 0xf9e39fff, 0xf8e1a2ff, 
    0xf5de9cff, 0xf4dc97ff, 0xf4d99aff, 0xf3d79aff, 0xf1d49bff, 0xeecd92ff, 
    0xe9c689ff, 0xdfb86cff, 0xe2b86cff, 0xdbaf62ff, 0xcb9532ff, 0xc38715ff, 
    0xc68b27ff, 0xbf811eff, 0xb56e07ff, 0xaf6105ff, 0xaf6704ff, 0xb16e06ff, 
    0xb27108ff, 0xc18722ff, 0xdbae52ff, 0xe4ba5fff, 0xe5bb64ff, 0xba7d2aff, 
    0x954903ff, 0x914703ff, 0x8a4303ff, 0x723a03ff, 0x4e3d27ff, 0x757774ff, 
    0x8c8e89ff, 0x9ea09cff, 0xd3d4d2ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe3e4e2ff, 0xc7c8c7ff, 0x535049ff, 0x67552eff, 
    0x9a7e46ff, 0xb8995dff, 0xcdad70ff, 0xd5b26fff, 0xcfa856ff, 0xca9d45ff, 
    0xbf8f2dff, 0xc18e29ff, 0xc28e26ff, 0xc39030ff, 0xc18c29ff, 0xdbbb75ff, 
    0xe6d097ff, 0xe4c98cff, 0xd7a64cff, 0xdeb56eff, 0xddb36bff, 0xd9af5fff, 
    0xd2a348ff, 0xd3a54cff, 0xd6a855ff, 0xd3a751ff, 0xd0a249ff, 0xd0a24eff, 
    0xca993cff, 0xc9973aff, 0xcc9a45ff, 0xc9943aff, 0xc58f33ff, 0xc38c30ff, 
    0xbe8523ff, 0xc08726ff, 0xd0a152ff, 0xd3a85aff, 0xd3a857ff, 0xd3a757ff, 
    0xd1a551ff, 0xd2a44eff, 0xd1a34bff, 0xd1a34cff, 0xcfa045ff, 0xca9738ff, 
    0xc48a1fff, 0xc48a21ff, 0xbe801aff, 0xc08421ff, 0xc38933ff, 0xbf842dff, 
    0xbd8028ff, 0xb67517ff, 0xaf6404ff, 0xc27c0eff, 0xebb852ff, 0xf9db91ff, 
    0xfce9b0ff, 0xfbe8b3ff, 0xfae7b2ff, 0xfbe9b4ff, 0xfae7adff, 0xfae5aaff, 
    0xf8e4a7ff, 0xf6de98ff, 0xf5dd99ff, 0xf6dd98ff, 0xf7dea1ff, 0xf5da9fff, 
    0xefd191ff, 0xeaca88ff, 0xe6c383ff, 0xe2bc74ff, 0xdfb567ff, 0xd7a752ff, 
    0xcd9731ff, 0xcb942dff, 0xcc9539ff, 0xc3892aff, 0xba760bff, 0xb06404ff, 
    0xae6205ff, 0xb26f08ff, 0xb3720aff, 0xb6760cff, 0xbe8016ff, 0xd5a33dff, 
    0xe2b75dff, 0xdcb05dff, 0xbc853aff, 0x9d611cff, 0x6d3908ff, 0x4b3b2aff, 
    0x797b78ff, 0x8c8e89ff, 0xa1a39fff, 0xd5d6d4ff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xdbdcdbff, 0x757573ff, 
    0x504329ff, 0x8e7442ff, 0xad8c4bff, 0xbd9647ff, 0xcba453ff, 0xd2ab5aff, 
    0xd2af63ff, 0xcea553ff, 0xc39436ff, 0xc39230ff, 0xc39233ff, 0xc3933bff, 
    0xca9d48ff, 0xe3c989ff, 0xe4cd94ff, 0xe5cb8fff, 0xd8aa50ff, 0xd9ae5cff, 
    0xdbb164ff, 0xd9b063ff, 0xd2a34bff, 0xd09f45ff, 0xd4a54fff, 0xd4a855ff, 
    0xd1a34cff, 0xcc9c40ff, 0xca983dff, 0xcc9d46ff, 0xca9843ff, 0xc8943bff, 
    0xc68f32ff, 0xc38c30ff, 0xbf8627ff, 0xbb801dff, 0xcb9844ff, 0xd4a960ff, 
    0xd3a858ff, 0xd4a959ff, 0xd5ab60ff, 0xd3a95cff, 0xd1a655ff, 0xd3a95bff, 
    0xd0a554ff, 0xd0a24eff, 0xca983aff, 0xc7922dff, 0xc2881fff, 0xba7d13ff, 
    0xb06c0aff, 0xae670bff, 0xbb7c14ff, 0xbe8317ff, 0xc2851aff, 0xc4881eff, 
    0xe2ad49ff, 0xf6d381ff, 0xfce8abff, 0xfdeab8ff, 0xfbeab8ff, 0xfbeab6ff, 
    0xfae7afff, 0xf8e3a6ff, 0xf7e3a5ff, 0xf7df9cff, 0xf7de98ff, 0xf5d890ff, 
    0xf5db9eff, 0xf2d598ff, 0xf0d291ff, 0xeccb89ff, 0xe7c480ff, 0xe5be77ff, 
    0xe1b86eff, 0xdcae5cff, 0xd7a54fff, 0xd2a048ff, 0xcd983bff, 0xc68c2dff, 
    0xbc7a0eff, 0xb26804ff, 0xad6205ff, 0xb26f08ff, 0xb26f07ff, 0xb37006ff, 
    0xb47206ff, 0xbf8216ff, 0xd3a442ff, 0xd6ac54ff, 0xcba453ff, 0x97783fff, 
    0x51493cff, 0x7b7d7bff, 0x8c8e89ff, 0xa5a6a2ff, 0xd7d9d6ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xa1a3a1ff, 0x443d2dff, 0x7a6438ff, 0xa58b57ff, 0xbd9d59ff, 0xc2993cff, 
    0xc79a37ff, 0xcfa54bff, 0xcea44cff, 0xca9f45ff, 0xcca14aff, 0xc99c3eff, 
    0xc2902cff, 0xbe8924ff, 0xd3ac57ff, 0xe4cb8bff, 0xe2ca8aff, 0xe5cc8eff, 
    0xd8a851ff, 0xd9ab58ff, 0xdcb367ff, 0xdeb973ff, 0xd4a855ff, 0xcfa044ff, 
    0xd3a853ff, 0xd5aa5aff, 0xd1a54eff, 0xcea14bff, 0xd0a454ff, 0xd0a455ff, 
    0xc99946ff, 0xc48f34ff, 0xc38c2cff, 0xc59138ff, 0xc48f36ff, 0xc08628ff, 
    0xc38c31ff, 0xcfa255ff, 0xd2a554ff, 0xd4a95dff, 0xd5ab60ff, 0xd4aa5eff, 
    0xd4aa5fff, 0xd7af67ff, 0xd5ad64ff, 0xd5aa60ff, 0xd3a653ff, 0xcea044ff, 
    0xc99633ff, 0xc99331ff, 0xc38b2bff, 0xc58b2bff, 0xc99234ff, 0xcc9a42ff, 
    0xcc9a42ff, 0xca9336ff, 0xd39d38ff, 0xefc56bff, 0xf9e3a0ff, 0xfae6aaff, 
    0xfbe8b7ff, 0xfae8b9ff, 0xfbeabaff, 0xfbe9b3ff, 0xf9e4a7ff, 0xf7e0a3ff, 
    0xf7dea0ff, 0xf5db96ff, 0xf2d692ff, 0xefd28eff, 0xf1d392ff, 0xf0d297ff, 
    0xe9c680ff, 0xe4be74ff, 0xdcb261ff, 0xddb05cff, 0xdcb066ff, 0xd7a857ff, 
    0xd09a3dff, 0xc78f2aff, 0xbc7a0fff, 0xb36804ff, 0xaf6404ff, 0xb37209ff, 
    0xb4740aff, 0xb5760aff, 0xb7780fff, 0xb5760cff, 0xba8318ff, 0xbd9233ff, 
    0x90722fff, 0x554e3fff, 0x7e807eff, 0x8c8e89ff, 0xa8aaa6ff, 0xd9dad8ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xc5c6c4ff, 0x53514aff, 0x614e26ff, 0x95783bff, 0xb38f46ff, 
    0xc19942ff, 0xcca245ff, 0xcfa54dff, 0xcb9e41ff, 0xcc9f3dff, 0xcda141ff, 
    0xca9f41ff, 0xc6983bff, 0xbd8b24ff, 0xc3912eff, 0xdfc073ff, 0xe4cc8bff, 
    0xe3cd8dff, 0xe5ce95ff, 0xdaad5cff, 0xd8aa55ff, 0xdbb160ff, 0xddb56cff, 
    0xd3a855ff, 0xce9e3fff, 0xd4a95aff, 0xd5ac64ff, 0xd6ab5eff, 0xcfa355ff, 
    0xcea252ff, 0xcea250ff, 0xc89844ff, 0xc6923bff, 0xc38c2dff, 0xc38e36ff, 
    0xc38e37ff, 0xc28a2eff, 0xbe8324ff, 0xc89846ff, 0xd1a656ff, 0xd3a85bff, 
    0xd4aa5eff, 0xd6ac61ff, 0xd8b069ff, 0xd9b16bff, 0xd9b16dff, 0xd8af66ff, 
    0xd5ab5aff, 0xd0a44dff, 0xcc9b3eff, 0xce9d42ff, 0xd0a24dff, 0xcfa04cff, 
    0xd2a452ff, 0xd2a554ff, 0xd2a34fff, 0xd1a049ff, 0xd29c3dff, 0xe7b655ff, 
    0xf1cf7eff, 0xf7e1a0ff, 0xf7e3a9ff, 0xf8e4afff, 0xf7e3b3ff, 0xf9e4b0ff, 
    0xf8e3a8ff, 0xf8e2a5ff, 0xf5dd9fff, 0xf3d791ff, 0xf4d899ff, 0xf2d694ff, 
    0xf1d390ff, 0xebc982ff, 0xe8c170ff, 0xe4bb69ff, 0xdaad53ff, 0xddb05cff, 
    0xdfb56bff, 0xdaad5dff, 0xd09c3eff, 0xc58b1eff, 0xbf8011ff, 0xb77007ff, 
    0xaf6504ff, 0xb26f05ff, 0xb5750aff, 0xb87b12ff, 0xba801fff, 0xb07714ff, 
    0xa36a07ff, 0x815d17ff, 0x575140ff, 0x818380ff, 0x8c8e89ff, 0xabada9ff, 
    0xdbdcdaff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xdadbd9ff, 0xb7b8b5ff, 0xbbbcb9ff, 0xcbcdcaff, 0xd8d9d6ff, 
    0xdfe0deff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xd8d9d7ff, 0x767775ff, 0x4d4026ff, 0x8b713cff, 
    0xad8d4cff, 0xc19a4dff, 0xc89e47ff, 0xd6b16cff, 0xdbb87bff, 0xd0a34cff, 
    0xc5921eff, 0xc59429ff, 0xc3912eff, 0xc1902dff, 0xc18f2cff, 0xd1a751ff, 
    0xe3c67fff, 0xe3c889ff, 0xe4cb8fff, 0xe5cb93ff, 0xdaae5bff, 0xd6a953ff, 
    0xd8ab52ff, 0xd7ac5aff, 0xd2a54eff, 0xd2a54dff, 0xd4a757ff, 0xd3a758ff, 
    0xd6ac61ff, 0xcfa554ff, 0xcb9f48ff, 0xcfa556ff, 0xcfa55dff, 0xcda054ff, 
    0xc89740ff, 0xc39139ff, 0xbd8427ff, 0xbc8224ff, 0xb77b16ff, 0xc59340ff, 
    0xd2a75fff, 0xd4aa5fff, 0xd7ae64ff, 0xd9af68ff, 0xd9b16aff, 0xd6ad63ff, 
    0xd7ad65ff, 0xd4a95cff, 0xd5a957ff, 0xd1a54eff, 0xcfa34cff, 0xd3a656ff, 
    0xd4a95bff, 0xd5aa5eff, 0xd7ab5fff, 0xd5a857ff, 0xd5a856ff, 0xd8ac5fff, 
    0xd4a24fff, 0xd4992dff, 0xe5b553ff, 0xf7e0a0ff, 0xfae7afff, 0xf9e7b5ff, 
    0xf5e0aaff, 0xf8e5afff, 0xf7e2a5ff, 0xf7e2a2ff, 0xf6df9cff, 0xf1d386ff, 
    0xefcd7fff, 0xf3d99bff, 0xf2d79eff, 0xe9c67bff, 0xe9c478ff, 0xe4bb6aff, 
    0xdfb45eff, 0xdcaf5bff, 0xdeb469ff, 0xd7a74fff, 0xd09f40ff, 0xc9922bff, 
    0xc88f28ff, 0xbb790cff, 0xaf6604ff, 0xae6904ff, 0xb6780eff, 0xb67a13ff, 
    0xb17817ff, 0xa26e12ff, 0x6d4809ff, 0x544d40ff, 0x848683ff, 0x8c8e89ff, 
    0xafb0adff, 0xdcdddbff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xdedfddff, 0xb0b1aeff, 0x8c8e8aff, 0x8c8e89ff, 
    0x8f918cff, 0x969894ff, 0xa3a5a1ff, 0xb4b5b2ff, 0xc5c6c3ff, 0xd3d4d2ff, 
    0xdcdddbff, 0xe1e2e0ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe2e2e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe0e1dfff, 0xa0a19fff, 0x453f33ff, 
    0x79653aff, 0xaa915fff, 0xc4a569ff, 0xcca55aff, 0xcea451ff, 0xd6b06bff, 
    0xd5b06cff, 0xcea145ff, 0xc7952bff, 0xc79733ff, 0xc08e23ff, 0xbc891fff, 
    0xc49536ff, 0xd8b15aff, 0xdfbe6eff, 0xe1c57cff, 0xe3c988ff, 0xe2c685ff, 
    0xdbad5aff, 0xd9aa54ff, 0xd8ab57ff, 0xd9ad5eff, 0xd5a955ff, 0xd3a551ff, 
    0xd4a755ff, 0xd2a655ff, 0xd0a454ff, 0xd1a653ff, 0xd4aa5fff, 0xd4ae6cff, 
    0xd1aa6aff, 0xd0a863ff, 0xcca154ff, 0xc79546ff, 0xbd8428ff, 0xb67610ff, 
    0xb87a1aff, 0xbc8122ff, 0xc89744ff, 0xd1a65aff, 0xd6ab62ff, 0xd5aa5dff, 
    0xd5ab5eff, 0xd4ab60ff, 0xd4a95bff, 0xd2a757ff, 0xcfa24cff, 0xd1a551ff, 
    0xd6ae66ff, 0xdab270ff, 0xd8b16bff, 0xd7ac61ff, 0xdab068ff, 0xdab067ff, 
    0xdbb064ff, 0xdbaf64ff, 0xd6a757ff, 0xcc8d25ff, 0xe4b34cff, 0xf2d58cff, 
    0xf7e2a8ff, 0xf8e5b1ff, 0xf6e3aaff, 0xf6e4b1ff, 0xf7e4b0ff, 0xf7e2a9ff, 
    0xf7e1a2ff, 0xf4d994ff, 0xefcb71ff, 0xefd189ff, 0xf2d699ff, 0xedcc85ff, 
    0xebc575ff, 0xe8c06eff, 0xe6bf71ff, 0xddb25dff, 0xdeb264ff, 0xd8a850ff, 
    0xd3a241ff, 0xcb952dff, 0xc8902cff, 0xbe7d11ff, 0xb26b06ff, 0xae6706ff, 
    0xb57a16ff, 0xb0791aff, 0xa36f17ff, 0x6c4b14ff, 0x534f45ff, 0x858885ff, 
    0x8d8f8aff, 0xb3b4b1ff, 0xdedfddff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe0e1dfff, 0xa8a9a7ff, 0x4c4b47ff, 
    0x4e4d48ff, 0x646562ff, 0x767875ff, 0x838582ff, 0x8a8c88ff, 0x8c8e89ff, 
    0x8d8f8aff, 0x929490ff, 0x9d9f9bff, 0xadaeabff, 0xbebfbcff, 0xcecfccff, 
    0xd9dad8ff, 0xe0e1dfff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xc1c2c0ff, 
    0x55544fff, 0x5f4f31ff, 0x9c8557ff, 0xc1a775ff, 0xcead6fff, 0xcba14fff, 
    0xcea24cff, 0xd2a759ff, 0xd0a44eff, 0xcea143ff, 0xca9b39ff, 0xc89832ff, 
    0xc18f26ff, 0xc1902dff, 0xcca144ff, 0xdcb55cff, 0xdfbc6eff, 0xe0c37dff, 
    0xe1c37aff, 0xe0be6eff, 0xe1b86dff, 0xdfb873ff, 0xddb469ff, 0xdaaf62ff, 
    0xdbb370ff, 0xd6ab60ff, 0xd1a24bff, 0xd5aa5eff, 0xd2a657ff, 0xd5aa5eff, 
    0xd6ae67ff, 0xd7b171ff, 0xd5ac6bff, 0xd1a862ff, 0xcb9f55ff, 0xc79344ff, 
    0xbe852bff, 0xb3720aff, 0xb4710aff, 0xb6710eff, 0xba7f21ff, 0xcb9b4dff, 
    0xd3a760ff, 0xd3a85dff, 0xd5a95cff, 0xd3a85bff, 0xd2a654ff, 0xd0a555ff, 
    0xcea04bff, 0xd4a95cff, 0xdbb473ff, 0xdeb97aff, 0xdcb876ff, 0xddb571ff, 
    0xddb671ff, 0xddb46cff, 0xddb366ff, 0xdaae60ff, 0xd7a852ff, 0xdaa74aff, 
    0xe7b95bff, 0xf1cf7cff, 0xf3da94ff, 0xf3db96ff, 0xf5dfa0ff, 0xf6e0a9ff, 
    0xf6e3b3ff, 0xf6e0acff, 0xf6dfa4ff, 0xf5dca0ff, 0xf1d185ff, 0xebc56bff, 
    0xedcb7fff, 0xebc87aff, 0xebc46fff, 0xebc87eff, 0xe7c070ff, 0xe3ba6cff, 
    0xe5bb70ff, 0xdbad58ff, 0xd7a64dff, 0xcc9630ff, 0xc18415ff, 0xb87406ff, 
    0xb06a04ff, 0xae6b06ff, 0xae7512ff, 0xa06c12ff, 0x6b4d18ff, 0x58554dff, 
    0x878986ff, 0x8d8f8aff, 0xb6b8b4ff, 0xdfe0deff, 0xe1e2e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xaeb0aeff, 
    0x46423cff, 0x2e1c03ff, 0x412f0cff, 0x483d26ff, 0x443e30ff, 0x444139ff, 
    0x4c4b46ff, 0x5c5d5aff, 0x6f706eff, 0x7f807dff, 0x888a86ff, 0x8b8d89ff, 
    0x8c8e8aff, 0x90928dff, 0x989a95ff, 0xa6a7a3ff, 0xb7b8b5ff, 0xc8c9c7ff, 
    0xcbccc9ff, 0x777876ff, 0x4c422dff, 0x887043ff, 0xad935bff, 0xc6a666ff, 
    0xd2ac65ff, 0xcea450ff, 0xcfa34aff, 0xd0a34aff, 0xcea043ff, 0xd1a651ff, 
    0xd1a958ff, 0xcfa659ff, 0xcb9e4bff, 0xc5963cff, 0xcea44bff, 0xdcb45eff, 
    0xdebb6fff, 0xe0c279ff, 0xe0be6cff, 0xddb863ff, 0xdeb872ff, 0xd5a756ff, 
    0xdaaf66ff, 0xdbb169ff, 0xdab170ff, 0xd5aa5fff, 0xcf9f45ff, 0xd8af66ff, 
    0xd5ab60ff, 0xd4a95aff, 0xd9b170ff, 0xd9b176ff, 0xd6ac6bff, 0xd2a355ff, 
    0xce9b44ff, 0xcb9537ff, 0xc48722ff, 0xbf7f13ff, 0xbf7e13ff, 0xbb740aff, 
    0xb77312ff, 0xc58f3fff, 0xd3aa69ff, 0xd6af72ff, 0xd5ad6aff, 0xd3aa5fff, 
    0xd2a85dff, 0xd4aa65ff, 0xd4ab62ff, 0xd3a95dff, 0xdbb371ff, 0xddb573ff, 
    0xdeb66eff, 0xdeba75ff, 0xdebb79ff, 0xdeb670ff, 0xdbad5bff, 0xdaab55ff, 
    0xdbac55ff, 0xddaf59ff, 0xe9c06bff, 0xf1d180ff, 0xf5da92ff, 0xf3da92ff, 
    0xf4de9fff, 0xf5dfa3ff, 0xf6e0aaff, 0xf4dca6ff, 0xf1d697ff, 0xf3d99cff, 
    0xf2d692ff, 0xf0d17fff, 0xecca76ff, 0xeac573ff, 0xeac677ff, 0xeac77aff, 
    0xe8c474ff, 0xebc985ff, 0xe7c27cff, 0xdeb25dff, 0xdaaa55ff, 0xcb9430ff, 
    0xb9780aff, 0xb06b05ff, 0xaa6704ff, 0xa56605ff, 0x976207ff, 0x674811ff, 
    0x5c5951ff, 0x898b87ff, 0x8e908bff, 0xbabbb8ff, 0xe0e1dfff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xb3b4b2ff, 0x4f4e49ff, 0x4b3c22ff, 0x573308ff, 0x633d08ff, 0x826a3bff, 
    0x857452ff, 0x786a4fff, 0x675b46ff, 0x564d3aff, 0x494132ff, 0x423e32ff, 
    0x47453fff, 0x555551ff, 0x676866ff, 0x797a78ff, 0x858783ff, 0x8b8c88ff, 
    0x8c8e89ff, 0x90928dff, 0x888986ff, 0x49443aff, 0x766136ff, 0xa18349ff, 
    0xbd9d5bff, 0xcfab65ff, 0xd6b36bff, 0xd2a958ff, 0xd1a651ff, 0xcca149ff, 
    0xcea34fff, 0xd1a757ff, 0xcda350ff, 0xca9e49ff, 0xcba04eff, 0xc7983eff, 
    0xd2a445ff, 0xdbaf56ff, 0xe1c27cff, 0xe2c37bff, 0xddb965ff, 0xd8b25aff, 
    0xd9af5bff, 0xd8ab54ff, 0xdaaf63ff, 0xd7ab5dff, 0xd5a756ff, 0xcd9b40ff, 
    0xd0a14aff, 0xd1a350ff, 0xd5a95bff, 0xd9ad65ff, 0xdbb172ff, 0xd8ab63ff, 
    0xd7a752ff, 0xdaa74eff, 0xd9a344ff, 0xd5a03bff, 0xd1992bff, 0xce952aff, 
    0xca9028ff, 0xc68820ff, 0xbb7b15ff, 0xc2862eff, 0xcc9f56ff, 0xd6af6dff, 
    0xdcb781ff, 0xdab47aff, 0xd8b172ff, 0xdcb578ff, 0xdeb97eff, 0xdab272ff, 
    0xddb574ff, 0xdcb370ff, 0xdeb66eff, 0xe0ba76ff, 0xe1bc7aff, 0xdfb46aff, 
    0xdcad5aff, 0xdcae59ff, 0xdeb15fff, 0xdeae5aff, 0xe6bb63ff, 0xf1d180ff, 
    0xf3d589ff, 0xf0cf7dff, 0xf1d183ff, 0xf1d58dff, 0xf3dc9eff, 0xf3da9eff, 
    0xf2d799ff, 0xf1d594ff, 0xf2d792ff, 0xf2d88eff, 0xf3d891ff, 0xeccb7dff, 
    0xeccb81ff, 0xf3d89aff, 0xf1d69aff, 0xedcf8bff, 0xebc988ff, 0xe2ba6bff, 
    0xdeb266ff, 0xd09d44ff, 0xb6740cff, 0xa86504ff, 0xa36405ff, 0x945e06ff, 
    0x5d4114ff, 0x5f5d56ff, 0x8a8c88ff, 0x8f918cff, 0xbdbfbcff, 0xe0e1dfff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e2e0ff, 0xb8b9b7ff, 0x51514cff, 0x544930ff, 0x89764fff, 0x917038ff, 
    0x794409ff, 0x956c28ff, 0xb19764ff, 0xad9c77ff, 0xa29275ff, 0x98896cff, 
    0x8c7f60ff, 0x7e7152ff, 0x6f6449ff, 0x5c5038ff, 0x4a402bff, 0x413a2cff, 
    0x434037ff, 0x4e4d48ff, 0x5f605dff, 0x727471ff, 0x54534dff, 0x614f2bff, 
    0x98793fff, 0xb69451ff, 0xcaa660ff, 0xd6b370ff, 0xd9b770ff, 0xd6af5dff, 
    0xcfa44aff, 0xcba044ff, 0xd1a955ff, 0xcb9f42ff, 0xc5932eff, 0xc59430ff, 
    0xc39330ff, 0xc99936ff, 0xdcb254ff, 0xe0bd6dff, 0xe3c785ff, 0xdfc17aff, 
    0xd8b460ff, 0xd5af58ff, 0xd6ab52ff, 0xd4a74cff, 0xd5a956ff, 0xd0a047ff, 
    0xce9a3aff, 0xc8932dff, 0xce9e42ff, 0xd2a24aff, 0xdca951ff, 0xe1ac52ff, 
    0xe0aa47ff, 0xe2ad50ff, 0xe3b152ff, 0xdaa342ff, 0xd49937ff, 0xddaa50ff, 
    0xdaa74bff, 0xd29d39ff, 0xcf9a39ff, 0xcd9637ff, 0xc99036ff, 0xc3892dff, 
    0xc99545ff, 0xd8af6dff, 0xe2c08dff, 0xdebb87ff, 0xdfb981ff, 0xe4bf89ff, 
    0xe4bf8cff, 0xe4c08bff, 0xe5c18aff, 0xe0bb7eff, 0xdeb66fff, 0xe0b975ff, 
    0xe1bb7aff, 0xe4bd7aff, 0xe2b66dff, 0xe2b972ff, 0xe1b56cff, 0xd9a550ff, 
    0xddaa4aff, 0xefc772ff, 0xedc971ff, 0xe9c15dff, 0xf0cc74ff, 0xf1d387ff, 
    0xf3d997ff, 0xf3d99bff, 0xf0d68eff, 0xf3d993ff, 0xf5dd9cff, 0xf3d890ff, 
    0xf2d890ff, 0xf0d288ff, 0xf1d48fff, 0xf5deabff, 0xf2d7a0ff, 0xebca82ff, 
    0xe8c377ff, 0xe2b96aff, 0xdcae5bff, 0xcb9532ff, 0xaf6e0aff, 0xa26404ff, 
    0x905a03ff, 0x573c12ff, 0x605f5bff, 0x8b8d89ff, 0x91938eff, 0xc1c2bfff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xbdbebcff, 0x56554fff, 0x504224ff, 0x8a7445ff, 
    0xac935cff, 0xc1a466ff, 0xae803dff, 0x8f510bff, 0xba9042ff, 0xd0b781ff, 
    0xc8b287ff, 0xc0ae8aff, 0xbaaa89ff, 0xb1a07bff, 0xa89a75ff, 0x9c8c63ff, 
    0x8c7543ff, 0x7f6a40ff, 0x71613fff, 0x615234ff, 0x57492dff, 0x52472fff, 
    0x63522dff, 0x917135ff, 0xae8b47ff, 0xc49e56ff, 0xd1aa62ff, 0xd6b06bff, 
    0xd8b166ff, 0xd8b265ff, 0xd5ad5aff, 0xd5ab57ff, 0xcfa44dff, 0xc89739ff, 
    0xc18f2bff, 0xc08e25ff, 0xbe8c1eff, 0xd5aa4cff, 0xe1bc69ff, 0xe3c480ff, 
    0xe3c78cff, 0xddc281ff, 0xd6b66cff, 0xd7b66eff, 0xdcb36fff, 0xdaaf65ff, 
    0xd8ab5fff, 0xd09f46ff, 0xcf9a3eff, 0xca942cff, 0xd29d39ff, 0xe1ac4bff, 
    0xeab75dff, 0xecbc68ff, 0xeab75aff, 0xeab964ff, 0xe8b964ff, 0xe3b053ff, 
    0xe0ac54ff, 0xe4b663ff, 0xe0b25dff, 0xd9a951ff, 0xd3a144ff, 0xd29f48ff, 
    0xce9b49ff, 0xc48a2cff, 0xc9923bff, 0xdbb175ff, 0xe3c092ff, 0xe0bb85ff, 
    0xe0ba7bff, 0xe5c28bff, 0xe9c796ff, 0xe9c898ff, 0xebc897ff, 0xe7c38aff, 
    0xe2ba73ff, 0xe5bf7eff, 0xe8c287ff, 0xecc891ff, 0xe6c182ff, 0xdfb370ff, 
    0xd39e4cff, 0xc88c2cff, 0xc68b23ff, 0xdeae4cff, 0xe8ba55ff, 0xeabc59ff, 
    0xf0cb75ff, 0xf3d382ff, 0xf2d487ff, 0xf3d992ff, 0xf5e0a3ff, 0xf8e2acff, 
    0xf8e4afff, 0xf8e0a8ff, 0xf6dc9bff, 0xf1d080ff, 0xf4d48cff, 0xf3d99fff, 
    0xf1d597ff, 0xeece90ff, 0xe8c479ff, 0xe5bd72ff, 0xd9aa57ff, 0xc08a27ff, 
    0xa16506ff, 0x8d5702ff, 0x563d15ff, 0x646460ff, 0x8b8d89ff, 0x92948fff, 
    0xc4c5c3ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xc1c3c1ff, 0x5c5c58ff, 0x4e4022ff, 
    0x816528ff, 0xa28036ff, 0xc19d50ff, 0xd4b26aff, 0xd8b266ff, 0xbd8931ff, 
    0xbf8b2dff, 0xddb96aff, 0xe2c88dff, 0xdfca9bff, 0xdbcaa2ff, 0xd5c49aff, 
    0xcebd92ff, 0xc7b78eff, 0xbca874ff, 0xae904eff, 0xa38a54ff, 0x9b8456ff, 
    0x977f50ff, 0x9a7e4bff, 0xa4864cff, 0xb08d48ff, 0xc39f59ff, 0xcda65aff, 
    0xd3aa57ff, 0xd4ad5bff, 0xd6af61ff, 0xd8b062ff, 0xd4ac5cff, 0xd3ac5dff, 
    0xd3aa5aff, 0xcb9f47ff, 0xc49433ff, 0xc29027ff, 0xc4932bff, 0xdbb356ff, 
    0xe1bd68ff, 0xe4c785ff, 0xe3c98fff, 0xdabe7eff, 0xd2b367ff, 0xd5b56dff, 
    0xdbae64ff, 0xd9ab5fff, 0xd7ab5cff, 0xd1a146ff, 0xc9952eff, 0xd1992dff, 
    0xe5b04dff, 0xeec273ff, 0xefc477ff, 0xefc472ff, 0xedbd64ff, 0xe9b85fff, 
    0xe9b861ff, 0xe7b65eff, 0xe8bc70ff, 0xe8c07bff, 0xe4b969ff, 0xddaf58ff, 
    0xd7a54aff, 0xd9ab60ff, 0xd1a151ff, 0xc78f30ff, 0xc68d2cff, 0xdbaf6fff, 
    0xe2bc87ff, 0xe3bd86ff, 0xe6c18aff, 0xe9c692ff, 0xebcc9cff, 0xebcc9cff, 
    0xedca97ff, 0xebc791ff, 0xe9c585ff, 0xecc98eff, 0xeecd95ff, 0xe9c387ff, 
    0xdbac60ff, 0xcc943bff, 0xbe811bff, 0xb4760bff, 0xb17205ff, 0xbb7e10ff, 
    0xd59f32ff, 0xe3b144ff, 0xecbe58ff, 0xf1cb71ff, 0xf4d98fff, 0xf6df9fff, 
    0xf7e3adff, 0xf6dfa8ff, 0xf7e2b0ff, 0xf6e2b1ff, 0xf3da9bff, 0xf2d78cff, 
    0xf5dca0ff, 0xf0d594ff, 0xefd18aff, 0xefd290ff, 0xebca82ff, 0xe0b767ff, 
    0xc69432ff, 0xa76f16ff, 0x895507ff, 0x58411aff, 0x696965ff, 0x8c8e89ff, 
    0x949691ff, 0xc7c9c6ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xc6c7c5ff, 0x62625fff, 
    0x544c3bff, 0x8d7c57ff, 0xac9053ff, 0xbb9441ff, 0xc99d40ff, 0xd1a441ff, 
    0xd6a747ff, 0xd0a03dff, 0xcc9a30ff, 0xdaad4dff, 0xecd18fff, 0xf0dda8ff, 
    0xecd8a5ff, 0xe8d6aaff, 0xe5d2a4ff, 0xe2d0a1ff, 0xdbc697ff, 0xd1b26dff, 
    0xc7a55cff, 0xc1a46dff, 0xbb9d64ff, 0xba9857ff, 0xbd9851ff, 0xc49c4fff, 
    0xc89e4dff, 0xcfa451ff, 0xd1a651ff, 0xd5ac59ff, 0xd5af5fff, 0xd9b368ff, 
    0xd2a956ff, 0xcea552ff, 0xd0a758ff, 0xca9f46ff, 0xc79a3cff, 0xc2922cff, 
    0xcda13cff, 0xe0ba64ff, 0xe0bf6fff, 0xdec077ff, 0xddc180ff, 0xd9bb77ff, 
    0xd8ba78ff, 0xd6b877ff, 0xd29c3bff, 0xd19b3cff, 0xd4a247ff, 0xd39f40ff, 
    0xd49f3bff, 0xe3af4cff, 0xf0c87bff, 0xf0cd89ff, 0xf1ce88ff, 0xf0cc88ff, 
    0xedc880ff, 0xedc476ff, 0xedc57aff, 0xe9bc6bff, 0xe6b767ff, 0xe5b867ff, 
    0xe4b764ff, 0xdeaf57ff, 0xd4a03bff, 0xd09b37ff, 0xcf9b37ff, 0xc68d22ff, 
    0xbf8215ff, 0xd19f4dff, 0xdaae6bff, 0xdfb674ff, 0xeac892ff, 0xebc996ff, 
    0xecca96ff, 0xedcc98ff, 0xeecc95ff, 0xebc98dff, 0xedcc8eff, 0xedca8dff, 
    0xe8c383ff, 0xd8a858ff, 0xc28527ff, 0xb47311ff, 0xaf7004ff, 0xaf7004ff, 
    0xb17205ff, 0xb27405ff, 0xb37606ff, 0xcc9629ff, 0xebc05fff, 0xf7da94ff, 
    0xf7de9dff, 0xf5dd9aff, 0xf7e1a6ff, 0xf4dd9fff, 0xf4dda2ff, 0xf4dba5ff, 
    0xf3da9cff, 0xf2d58eff, 0xf4da99ff, 0xf2d58fff, 0xf0d182ff, 0xf2d894ff, 
    0xeac980ff, 0xd2a449ff, 0xb07a1cff, 0x936823ff, 0x57482dff, 0x6d6e6aff, 
    0x8c8e89ff, 0x969894ff, 0xcbccc9ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xc9cbc9ff, 
    0x656563ff, 0x4d4532ff, 0x897854ff, 0xb19d70ff, 0xcbb06fff, 0xddc17aff, 
    0xe7ca84ff, 0xe5c271ff, 0xd8ab45ff, 0xd3a339ff, 0xce9a30ff, 0xcc9623ff, 
    0xe6c577ff, 0xf1d9a0ff, 0xf0d69cff, 0xf0dcaaff, 0xf0deaeff, 0xefddadff, 
    0xebd7a7ff, 0xe9d39dff, 0xddba6cff, 0xcea455ff, 0xcda75aff, 0xd0aa5cff, 
    0xd1aa5dff, 0xcba14bff, 0xcb9e44ff, 0xcfa34cff, 0xd5ac5cff, 0xd4aa57ff, 
    0xd0a64fff, 0xd0a650ff, 0xd1a850ff, 0xd1a652ff, 0xd2a958ff, 0xcca044ff, 
    0xc59636ff, 0xc2932fff, 0xd9b35fff, 0xe0be71ff, 0xdebe70ff, 0xdaba6eff, 
    0xd9ba72ff, 0xdbbf7fff, 0xdbc085ff, 0xd5b673ff, 0xcf9528ff, 0xd39b39ff, 
    0xd9a54bff, 0xe2ae4dff, 0xeab85bff, 0xf0c577ff, 0xefcb86ff, 0xefcd8bff, 
    0xf1cf93ff, 0xf0cd91ff, 0xefcb88ff, 0xeecb86ff, 0xecc67fff, 0xe4b55eff, 
    0xe5b355ff, 0xebc174ff, 0xe4b75eff, 0xdca83eff, 0xd69f32ff, 0xd39e3bff, 
    0xd29c3aff, 0xce983aff, 0xc18317ff, 0xc48620ff, 0xd29e46ff, 0xddae62ff, 
    0xe9c38aff, 0xecca92ff, 0xedcb91ff, 0xe8c07bff, 0xe7bd71ff, 0xeac47cff, 
    0xe7bd76ff, 0xdca850ff, 0xce9330ff, 0xc17c12ff, 0xb87005ff, 0xb76f04ff, 
    0xb56e04ff, 0xb36f04ff, 0xb06f05ff, 0xb17205ff, 0xb07004ff, 0xb77c11ff, 
    0xd39f33ff, 0xefca75ff, 0xf9e3a5ff, 0xf8e3a2ff, 0xf9e4a2ff, 0xf7e3a0ff, 
    0xf8e4a7ff, 0xf6dca3ff, 0xf3d999ff, 0xf0d387ff, 0xf6de9cff, 0xf4d993ff, 
    0xf2d891ff, 0xf0d696ff, 0xe1bf77ff, 0xc09035ff, 0x9b7430ff, 0x594e37ff, 
    0x70716fff, 0x8c8e89ff, 0x999a96ff, 0xcdcfccff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xcdceccff, 0x6a6b69ff, 0x463e2dff, 0x826f4cff, 0xa89164ff, 0xc4a96eff, 
    0xdabc7aff, 0xe7cb87ff, 0xf0d598ff, 0xf4dda3ff, 0xf4da9cff, 0xeac97bff, 
    0xe2bc67ff, 0xcf9b2dff, 0xdab052ff, 0xedd190ff, 0xedd393ff, 0xf3dea7ff, 
    0xf6e6bdff, 0xf3e1b5ff, 0xefdcacff, 0xeedcacff, 0xe7c980ff, 0xdbad4bff, 
    0xdeb86bff, 0xdebb74ff, 0xd9b369ff, 0xd1a752ff, 0xd0a34aff, 0xcea048ff, 
    0xd1a753ff, 0xd2a752ff, 0xd1a653ff, 0xd2a858ff, 0xcfa44eff, 0xcba045ff, 
    0xca9f43ff, 0xc89a38ff, 0xc69834ff, 0xcea248ff, 0xdebe74ff, 0xdcba6dff, 
    0xdbb969ff, 0xd9b96cff, 0xd8ba6eff, 0xdbc389ff, 0xd8c187ff, 0xd1b166ff, 
    0xc7850eff, 0xd49b30ff, 0xeabc69ff, 0xf0c675ff, 0xf2c980ff, 0xf1cc89ff, 
    0xedc578ff, 0xedc780ff, 0xefcb86ff, 0xf2d396ff, 0xefcb80ff, 0xebc26cff, 
    0xe7b755ff, 0xe4b14cff, 0xebbd69ff, 0xeabf74ff, 0xe4b764ff, 0xe1b05bff, 
    0xdca953ff, 0xd4a046ff, 0xd39e42ff, 0xd59f47ff, 0xd09530ff, 0xce9129ff, 
    0xd29a35ff, 0xd59c34ff, 0xd9a64aff, 0xe1b563ff, 0xe3b460ff, 0xe0af52ff, 
    0xe0b158ff, 0xdfae55ff, 0xd3982eff, 0xc67e0aff, 0xc37a04ff, 0xc78409ff, 
    0xc8850aff, 0xc68208ff, 0xc37b06ff, 0xbd7204ff, 0xb77204ff, 0xb57205ff, 
    0xb17205ff, 0xb47505ff, 0xbf8412ff, 0xe4b552ff, 0xf6da91ff, 0xfbe8a9ff, 
    0xfbe9aaff, 0xfce8adff, 0xfae7b6ff, 0xf6dfa8ff, 0xf2d68eff, 0xf3d98cff, 
    0xf6de9aff, 0xf3d996ff, 0xeed28dff, 0xe2bf76ff, 0xcca550ff, 0x99742aff, 
    0x584d36ff, 0x737572ff, 0x8c8e89ff, 0x9b9d99ff, 0xd0d1cfff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xd0d1cfff, 0x717270ff, 0x453e2eff, 0x7e6b45ff, 0xa79065ff, 
    0xc3a878ff, 0xd1b172ff, 0xdcb76aff, 0xe3c27bff, 0xeaca88ff, 0xeccd8bff, 
    0xedcd87ff, 0xecca78ff, 0xf2d796ff, 0xebcc8aff, 0xdaad50ff, 0xedcd83ff, 
    0xefd593ff, 0xeed59bff, 0xedd8a8ff, 0xedd8a8ff, 0xefdbabff, 0xefdcadff, 
    0xead090ff, 0xe5be68ff, 0xe4c37eff, 0xdcb76eff, 0xd7af60ff, 0xd1a652ff, 
    0xd0a54eff, 0xcfa24bff, 0xcca049ff, 0xd2a856ff, 0xd2aa5cff, 0xd0a85aff, 
    0xcba047ff, 0xcda248ff, 0xcda246ff, 0xca9c3aff, 0xc3942aff, 0xcfa74fff, 
    0xddbe77ff, 0xdbb96bff, 0xdaba68ff, 0xd9bc70ff, 0xd9bc77ff, 0xd8be7dff, 
    0xd8be82ff, 0xd2b470ff, 0xcb881fff, 0xe7b559ff, 0xf3d091ff, 0xf3d492ff, 
    0xf1cc84ff, 0xeec67aff, 0xeec471ff, 0xeec679ff, 0xf3d491ff, 0xf3d694ff, 
    0xf0cd82ff, 0xeabd5fff, 0xe8b854ff, 0xe9b762ff, 0xe8b969ff, 0xe4b35cff, 
    0xdeac51ff, 0xdeac5bff, 0xddaa58ff, 0xdba84fff, 0xdda642ff, 0xdfa948ff, 
    0xe5b869ff, 0xe6b969ff, 0xe2b35dff, 0xe0ad4fff, 0xdca746ff, 0xd69f33ff, 
    0xd79e2eff, 0xd79e2dff, 0xd79c2eff, 0xd29119ff, 0xce8807ff, 0xcd8807ff, 
    0xcc8607ff, 0xcc8d12ff, 0xcc8c16ff, 0xc98910ff, 0xcb870aff, 0xc7840aff, 
    0xc57f06ff, 0xc47c05ff, 0xc17f09ff, 0xc0820cff, 0xb97c07ff, 0xce9c38ff, 
    0xebc671ff, 0xf5db91ff, 0xf8e19dff, 0xf7df9dff, 0xf7e1a5ff, 0xf5dd9fff, 
    0xf9e1a0ff, 0xf7df9cff, 0xf4da94ff, 0xedd18cff, 0xe2bf72ff, 0xcba248ff, 
    0x9b7934ff, 0x5a4f38ff, 0x777976ff, 0x8c8e89ff, 0x9ea09cff, 0xd3d4d1ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xd3d4d2ff, 0x787876ff, 0x453e30ff, 0x796642ff, 
    0x9f8653ff, 0xbea269ff, 0xd2b378ff, 0xdcbc7bff, 0xe0bf7aff, 0xe2c07aff, 
    0xe5c480ff, 0xe7c57dff, 0xe8c374ff, 0xe8c576ff, 0xebc982ff, 0xeccd8aff, 
    0xe5be6fff, 0xebc570ff, 0xf0d68eff, 0xefd795ff, 0xecd391ff, 0xedd497ff, 
    0xf1deabff, 0xf3e1b1ff, 0xeed7a0ff, 0xe5c274ff, 0xe1be75ff, 0xdcb76dff, 
    0xddb974ff, 0xdab36bff, 0xd2a754ff, 0xd3a754ff, 0xcea14bff, 0xd2a859ff, 
    0xcea452ff, 0xcba14dff, 0xc99d44ff, 0xc59636ff, 0xc99d3fff, 0xc69734ff, 
    0xc08d21ff, 0xd4ae5dff, 0xddbf7bff, 0xd9b86dff, 0xd8b460ff, 0xdabc74ff, 
    0xddc286ff, 0xdbc087ff, 0xdac08aff, 0xd4b679ff, 0xebb95fff, 0xf4cf8bff, 
    0xf3d293ff, 0xf3d28fff, 0xf5d899ff, 0xf3d291ff, 0xf0cb84ff, 0xf2d294ff, 
    0xf1d18bff, 0xf0ce82ff, 0xf0cc82ff, 0xefc877ff, 0xefc97eff, 0xe8bb69ff, 
    0xdeaa48ff, 0xdca745ff, 0xdfa94dff, 0xdfaf5bff, 0xe1b056ff, 0xe9bb68ff, 
    0xebbe6dff, 0xeec376ff, 0xeeca87ff, 0xefca89ff, 0xefca87ff, 0xedc57aff, 
    0xe8bb65ff, 0xe4b250ff, 0xdea739ff, 0xd99d27ff, 0xd59515ff, 0xd3900cff, 
    0xd49212ff, 0xd29214ff, 0xd29112ff, 0xd2951cff, 0xcf9017ff, 0xcf9016ff, 
    0xd09113ff, 0xce9013ff, 0xcc8b0cff, 0xcb8808ff, 0xcb880bff, 0xcb8c13ff, 
    0xc7880dff, 0xc98c17ff, 0xd29c31ff, 0xe7bf69ff, 0xf2d287ff, 0xf3d585ff, 
    0xf6de97ff, 0xf8e0a0ff, 0xf8df9fff, 0xf4d992ff, 0xeccc81ff, 0xdfb86aff, 
    0xcea655ff, 0x927130ff, 0x514938ff, 0x7a7c7aff, 0x8c8e89ff, 0xa1a39fff, 
    0xd5d6d4ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xd6d7d5ff, 0x7d7e7cff, 0x453e31ff, 
    0x776544ff, 0x9f885cff, 0xba9c63ff, 0xd0b379ff, 0xdab878ff, 0xddb970ff, 
    0xdebb71ff, 0xe2c181ff, 0xe3c280ff, 0xe3c177ff, 0xe3bc69ff, 0xe4bd6bff, 
    0xe6c06eff, 0xe7c274ff, 0xe8c274ff, 0xe7c068ff, 0xf0d387ff, 0xf3de9dff, 
    0xf4e1a1ff, 0xf5e3acff, 0xf5e4b1ff, 0xf2e2b2ff, 0xedd79fff, 0xe4c06fff, 
    0xe2be71ff, 0xdeba71ff, 0xdbb56dff, 0xd8b166ff, 0xd2a956ff, 0xcda14bff, 
    0xcc9f43ff, 0xcda24dff, 0xca9d48ff, 0xc69739ff, 0xc59537ff, 0xc49230ff, 
    0xc29029ff, 0xc08f28ff, 0xc19025ff, 0xd7b568ff, 0xdec285ff, 0xd8b872ff, 
    0xd4b05fff, 0xd5b364ff, 0xd8bb77ff, 0xdbc28aff, 0xd9c290ff, 0xd9bf8aff, 
    0xf4cf8dff, 0xf7d8a0ff, 0xf4d697ff, 0xf1d191ff, 0xf6dba7ff, 0xf6dca9ff, 
    0xf3d69aff, 0xf2d392ff, 0xeec670ff, 0xf1cd82ff, 0xf1cf8aff, 0xf3d492ff, 
    0xf1d191ff, 0xedc67aff, 0xe3b257ff, 0xe1ae52ff, 0xe3b057ff, 0xe8b864ff, 
    0xefc778ff, 0xf0c97aff, 0xf1cc82ff, 0xf1cd82ff, 0xf2d28fff, 0xf3d296ff, 
    0xf0cc8aff, 0xefc87cff, 0xedc26dff, 0xecbe63ff, 0xebbb60ff, 0xe3ac40ff, 
    0xe0a935ff, 0xdfa733ff, 0xe0a93dff, 0xdda739ff, 0xdca431ff, 0xd9a12dff, 
    0xd69c24ff, 0xd9a232ff, 0xd89f2fff, 0xd49a24ff, 0xd29923ff, 0xd39824ff, 
    0xd19721ff, 0xd19821ff, 0xd49b2cff, 0xd49b29ff, 0xce9117ff, 0xd39b2cff, 
    0xe6ba5cff, 0xecc368ff, 0xefc76cff, 0xefca78ff, 0xebc572ff, 0xe3ba63ff, 
    0xcc9e46ff, 0xb98d39ff, 0x8a6827ff, 0x524a3aff, 0x7d7f7cff, 0x8c8d89ff, 
    0xa4a6a2ff, 0xd7d8d6ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xd8d9d7ff, 0x838482ff, 
    0x403a2dff, 0x72603dff, 0x9c8358ff, 0xbca172ff, 0xcbab6dff, 0xd9bb7eff, 
    0xdfc082ff, 0xdfbe7fff, 0xdebc78ff, 0xdeba73ff, 0xddb567ff, 0xdeb667ff, 
    0xe1bc71ff, 0xe2bc6fff, 0xe2bb6aff, 0xe2b865ff, 0xe3bc6aff, 0xe1b65bff, 
    0xebc774ff, 0xf4dd99ff, 0xf4de9fff, 0xf4dfa1ff, 0xf5e4aeff, 0xf2e0acff, 
    0xedd79bff, 0xeacc80ff, 0xe8ca85ff, 0xe6c88aff, 0xdebe7cff, 0xd8b164ff, 
    0xd3aa58ff, 0xcca24bff, 0xc89a3cff, 0xca9f49ff, 0xc99c47ff, 0xc59438ff, 
    0xc49232ff, 0xc39130ff, 0xbd891eff, 0xb98517ff, 0xc1902aff, 0xd7b465ff, 
    0xdcc081ff, 0xd9ba7bff, 0xd5b46dff, 0xd3af5fff, 0xd3b266ff, 0xd9bc7cff, 
    0xdbc089ff, 0xd9be86ff, 0xf3d499ff, 0xf4d59fff, 0xf3d79aff, 0xf1d393ff, 
    0xf1d193ff, 0xf3d599ff, 0xf0cd82ff, 0xebbd5eff, 0xedc163ff, 0xf1ce88ff, 
    0xf0cd89ff, 0xf6d9a3ff, 0xf1d190ff, 0xecc376ff, 0xe7b962ff, 0xe6b55cff, 
    0xecc06aff, 0xf3d187ff, 0xf4d690ff, 0xf2d183ff, 0xf3cf7eff, 0xf6d78fff, 
    0xf6d891ff, 0xf3d289ff, 0xf2cd7cff, 0xf4cf7cff, 0xf5d27fff, 0xf5d17fff, 
    0xf5cf7cff, 0xf3c970ff, 0xf3cb72ff, 0xf2cc75ff, 0xf1c86bff, 0xeec365ff, 
    0xeabb55ff, 0xe8b953ff, 0xe6b64fff, 0xe7b653ff, 0xe4b04cff, 0xe2b04eff, 
    0xe1b04dff, 0xe1ae4aff, 0xdeab3eff, 0xdba739ff, 0xdca942ff, 0xdea942ff, 
    0xdda740ff, 0xd79f2fff, 0xd3971fff, 0xdaa333ff, 0xd9a438ff, 0xd5a23dff, 
    0xcc9931ff, 0xbd8822ff, 0xa67112ff, 0x7d5811ff, 0x534c3cff, 0x808280ff, 
    0x8c8d89ff, 0xa7a9a5ff, 0xd9dad8ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xdadbd9ff, 
    0x8a8b89ff, 0x403a2eff, 0x6d5a33ff, 0x997f4eff, 0xb69962ff, 0xccac71ff, 
    0xd5b26eff, 0xdbb770ff, 0xdcb86eff, 0xdeb970ff, 0xddb870ff, 0xdbb76eff, 
    0xd7ae5bff, 0xd8af58ff, 0xdfb96cff, 0xdfbb73ff, 0xe0bb73ff, 0xdeb96bff, 
    0xe0bb6eff, 0xdfb65cff, 0xe6be62ff, 0xefd48cff, 0xf2de9cff, 0xf0da94ff, 
    0xf4e1a4ff, 0xf7e7b7ff, 0xf1dda7ff, 0xecd08bff, 0xe8c882ff, 0xe5c481ff, 
    0xddbb74ff, 0xd7b163ff, 0xd6b060ff, 0xd3aa5aff, 0xc99940ff, 0xc49335ff, 
    0xc39133ff, 0xc18d29ff, 0xc29232ff, 0xc08e2cff, 0xba8518ff, 0xb77e0aff, 
    0xbc8a1fff, 0xd3ae5eff, 0xddbe83ff, 0xdcbe84ff, 0xd7b879ff, 0xd4b26bff, 
    0xd1ae60ff, 0xd4b268ff, 0xd9bc7dff, 0xd8ba7eff, 0xf3d49aff, 0xf4d9a4ff, 
    0xf2d69aff, 0xf4d69dff, 0xf1d092ff, 0xefcc83ff, 0xecc16bff, 0xefc777ff, 
    0xf1cc86ff, 0xf0ce8dff, 0xefce8dff, 0xf4d799ff, 0xf4d593ff, 0xf1cf8dff, 
    0xebc06aff, 0xeec062ff, 0xf4d282ff, 0xf6d78dff, 0xf8db97ff, 0xf7da92ff, 
    0xf5d580ff, 0xf8da8aff, 0xf9da89ff, 0xf9d784ff, 0xf9d780ff, 0xfad883ff, 
    0xfad886ff, 0xfad988ff, 0xfada8bff, 0xf9d888ff, 0xf9d67fff, 0xf9d883ff, 
    0xf7d581ff, 0xf4ce70ff, 0xf3cb6cff, 0xf1c868ff, 0xf1c667ff, 0xefc361ff, 
    0xeec060ff, 0xeec269ff, 0xedc267ff, 0xeabd5fff, 0xe4b44fff, 0xe2b14bff, 
    0xe4b14eff, 0xe2af48ff, 0xe1ad46ff, 0xdeaa42ff, 0xdda842ff, 0xdaa63cff, 
    0xc28615ff, 0xb27506ff, 0xaa6e06ff, 0x9d6402ff, 0x714d0aff, 0x575040ff, 
    0x838582ff, 0x8c8e89ff, 0xabada9ff, 0xdbdcdaff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xdcdddbff, 0x90918fff, 0x423c30ff, 0x6b5834ff, 0x967b46ff, 0xb49557ff, 
    0xc8a560ff, 0xd2ad64ff, 0xd9b36bff, 0xdbb66dff, 0xd9b469ff, 0xdab566ff, 
    0xdcb76dff, 0xddba73ff, 0xdab261ff, 0xd9b15bff, 0xd9ad54ff, 0xdcb465ff, 
    0xdeb974ff, 0xddb873ff, 0xdfbb73ff, 0xe0b966ff, 0xe2b95fff, 0xeac771ff, 
    0xf3dc9cff, 0xf4e1a3ff, 0xf3e1a5ff, 0xf7e8bbff, 0xf3e2b3ff, 0xedd498ff, 
    0xeace8cff, 0xe6c784ff, 0xdcb86eff, 0xd6b05eff, 0xd4ac58ff, 0xd0a650ff, 
    0xca9e47ff, 0xc5963bff, 0xc18e2dff, 0xbf8b25ff, 0xc1902eff, 0xbe8c27ff, 
    0xb98313ff, 0xb77d08ff, 0xb98312ff, 0xcfa74fff, 0xdbbd7eff, 0xdbbe82ff, 
    0xd7b877ff, 0xd4b575ff, 0xd4b375ff, 0xceac64ff, 0xcca859ff, 0xcda453ff, 
    0xf3d79fff, 0xf3d8a3ff, 0xf6daa6ff, 0xf3d59aff, 0xefc97eff, 0xebba61ff, 
    0xebbd71ff, 0xefcc8eff, 0xf3d398ff, 0xf4d59aff, 0xf5d89aff, 0xf7dc9eff, 
    0xf5d99cff, 0xf1d08dff, 0xf2cb79ff, 0xf6d685ff, 0xf8db92ff, 0xf9de97ff, 
    0xf9dd93ff, 0xfadc8cff, 0xfada81ff, 0xfadd89ff, 0xfbdd88ff, 0xfbdb7cff, 
    0xfbdb83ff, 0xfbdd86ff, 0xfbdd8aff, 0xfbdd88ff, 0xfbde8cff, 0xfbe08fff, 
    0xfbdf8dff, 0xfbdc8aff, 0xfbda8aff, 0xf9d67eff, 0xf7d071ff, 0xf4cd68ff, 
    0xf6cf72ff, 0xf5cf77ff, 0xf3cc6fff, 0xf4cf78ff, 0xf3ce7cff, 0xf1ca77ff, 
    0xedc570ff, 0xedc46aff, 0xecbf62ff, 0xe8b858ff, 0xe2b14cff, 0xe0ae49ff, 
    0xe4b45aff, 0xe3b963ff, 0xc7922eff, 0xaf7406ff, 0xa16904ff, 0x6a4609ff, 
    0x524d43ff, 0x858784ff, 0x8c8e89ff, 0xafb0acff, 0xdcdddbff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xdddedcff, 0x979896ff, 0x443e32ff, 0x665025ff, 0x92763dff, 
    0xb08f50ff, 0xc49f55ff, 0xcfa652ff, 0xd7af5eff, 0xdbb56eff, 0xddb871ff, 
    0xdcb770ff, 0xddba74ff, 0xddb971ff, 0xdfbe7cff, 0xdebc75ff, 0xdeba71ff, 
    0xd8ad53ff, 0xd7ad53ff, 0xddb76cff, 0xdbb467ff, 0xdab15cff, 0xdeb865ff, 
    0xe1b966ff, 0xe9c36dff, 0xeac47aff, 0xf5e0aaff, 0xf2dd9eff, 0xf4e2b0ff, 
    0xf2deb1ff, 0xecd497ff, 0xedd295ff, 0xe7c785ff, 0xddbb72ff, 0xd7b165ff, 
    0xd4ad5eff, 0xd2aa59ff, 0xcb9e47ff, 0xc08f2dff, 0xc08c29ff, 0xc18e2cff, 
    0xbe8d29ff, 0xbb8920ff, 0xbc891fff, 0xba8515ff, 0xb88310ff, 0xcda44bff, 
    0xdbbc7eff, 0xdcbf86ff, 0xd6b878ff, 0xd3b574ff, 0xd4b57aff, 0xcead69ff, 
    0xc9a14dff, 0xc6983aff, 0xf5ddaeff, 0xf6dbabff, 0xf6d9a1ff, 0xf1cf84ff, 
    0xebbb5dff, 0xeebf6eff, 0xefc880ff, 0xf0cf91ff, 0xf5d69fff, 0xf9e0adff, 
    0xf9dfa6ff, 0xf8dea6ff, 0xf5d698ff, 0xf6d48bff, 0xf9d67bff, 0xfbda85ff, 
    0xfbe39aff, 0xfce7a4ff, 0xfce49eff, 0xfbe191ff, 0xfce08aff, 0xfce08bff, 
    0xfcdf83ff, 0xfbde7dff, 0xfcdf88ff, 0xfce188ff, 0xfce18aff, 0xfce493ff, 
    0xfce596ff, 0xfce38dff, 0xfce28dff, 0xfcdf88ff, 0xfbdc81ff, 0xfbdc82ff, 
    0xfad87dff, 0xfad77aff, 0xfadb88ff, 0xf9da8aff, 0xf9d988ff, 0xf7d582ff, 
    0xf8d88dff, 0xf8d992ff, 0xf6d68aff, 0xf4d283ff, 0xf1cb74ff, 0xf0c76dff, 
    0xedc46eff, 0xebbf66ff, 0xeac16cff, 0xe7c479ff, 0xd3ad61ff, 0xa4700eff, 
    0x6a480aff, 0x555148ff, 0x878986ff, 0x8c8e89ff, 0xb2b4b0ff, 0xdededcff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e1ff, 0xdedfddff, 0x9c9d9cff, 0x433e34ff, 0x644f25ff, 
    0x8f7033ff, 0xaf8c4aff, 0xc59f58ff, 0xd0a95dff, 0xd3aa5bff, 0xd7ab5aff, 
    0xdab368ff, 0xddba75ff, 0xdcb66cff, 0xdbb56bff, 0xddb76fff, 0xdeba73ff, 
    0xdebc78ff, 0xdebb7aff, 0xd9b46eff, 0xd7af5eff, 0xd6ae5dff, 0xd2a74fff, 
    0xd8af57ff, 0xdbb159ff, 0xd9af55ff, 0xe1b963ff, 0xecca7cff, 0xf0d696ff, 
    0xf0d894ff, 0xf2e0a8ff, 0xf0dda9ff, 0xeed89aff, 0xeed69bff, 0xe6ca8aff, 
    0xdcba71ff, 0xd7b266ff, 0xd6af64ff, 0xcea450ff, 0xc8993fff, 0xbf8e2bff, 
    0xc08e30ff, 0xc39438ff, 0xc29233ff, 0xbd8b25ff, 0xbd8922ff, 0xbb871bff, 
    0xb8830fff, 0xcda54eff, 0xdbbd7eff, 0xdcc087ff, 0xd7b97aff, 0xd4b36dff, 
    0xd0ae67ff, 0xcca858ff, 0xcba54eff, 0xcfa852ff, 0xf5dca9ff, 0xf6dba6ff, 
    0xf4d693ff, 0xf1cc7eff, 0xf0c573ff, 0xf1c877ff, 0xf7d99bff, 0xf7dba5ff, 
    0xf4d49dff, 0xf4d49bff, 0xf6d99eff, 0xf5d59aff, 0xf4d18aff, 0xf8d476ff, 
    0xfad672ff, 0xfbdc82ff, 0xfce597ff, 0xfceaa1ff, 0xfceba1ff, 0xfceda0ff, 
    0xfcea9aff, 0xfce896ff, 0xfce58cff, 0xfce587ff, 0xfce689ff, 0xfce791ff, 
    0xfce895ff, 0xfce998ff, 0xfcea9dff, 0xfce895ff, 0xfce592ff, 0xfce38dff, 
    0xfbdf83ff, 0xfce289ff, 0xfce390ff, 0xfce18dff, 0xfce18fff, 0xfbe092ff, 
    0xfbdf90ff, 0xfbe192ff, 0xfbe297ff, 0xfbe39aff, 0xfae299ff, 0xfae096ff, 
    0xfae098ff, 0xf8dd94ff, 0xf4d487ff, 0xf3d48dff, 0xf0d696ff, 0xe6ce93ff, 
    0xd0b781ff, 0x735a29ff, 0x59564cff, 0x888a87ff, 0x8d8f8aff, 0xb6b7b4ff, 
    0xdfe0deff, 0xe2e2e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe0e0deff, 0xa2a4a2ff, 0x444037ff, 
    0x5f4b25ff, 0x8e6e31ff, 0xac863cff, 0xc19a4cff, 0xcca14bff, 0xd5ac5bff, 
    0xdab266ff, 0xdab269ff, 0xd7ae5eff, 0xd8af5eff, 0xd9b15dff, 0xdab160ff, 
    0xdbb569ff, 0xdcb465ff, 0xdab466ff, 0xdbb873ff, 0xdbb977ff, 0xdab872ff, 
    0xdab873ff, 0xd5ae61ff, 0xd8b05cff, 0xd7ad55ff, 0xd5a94cff, 0xd8ac56ff, 
    0xe3bc62ff, 0xefd48bff, 0xf2dc98ff, 0xf2dfa0ff, 0xf2dea5ff, 0xf0dba0ff, 
    0xefd9a0ff, 0xe6ca88ff, 0xddba6fff, 0xdab76bff, 0xd6b368ff, 0xcea550ff, 
    0xc79a40ff, 0xc39435ff, 0xbe8b2bff, 0xbe8b29ff, 0xbf8d2bff, 0xba871dff, 
    0xb88318ff, 0xb88518ff, 0xbb8718ff, 0xcca751ff, 0xdcbf82ff, 0xdbc28bff, 
    0xd8be85ff, 0xd4b570ff, 0xd1b06aff, 0xceac5dff, 0xceaa55ff, 0xd1ae5eff, 
    0xf5dba7ff, 0xf1cf83ff, 0xf3cf83ff, 0xf1cd89ff, 0xf3cf87ff, 0xf1ca7dff, 
    0xf3ce8bff, 0xf7dca7ff, 0xf4d59cff, 0xf5d699ff, 0xf4d798ff, 0xf2cf8aff, 
    0xf6d17cff, 0xf9d36dff, 0xf9d672ff, 0xfada7bff, 0xfce391ff, 0xfce89cff, 
    0xfdf0b0ff, 0xfdf1afff, 0xfcf2acff, 0xfcf0a6ff, 0xfcea9bff, 0xfce892ff, 
    0xfce78dff, 0xfce893ff, 0xfce891ff, 0xfce793ff, 0xfce997ff, 0xfce895ff, 
    0xfce693ff, 0xfce38eff, 0xfce28aff, 0xfce48eff, 0xfce595ff, 0xfce593ff, 
    0xfce190ff, 0xfce69cff, 0xfce8a2ff, 0xfce69bff, 0xfce69bff, 0xfce397ff, 
    0xfce69aff, 0xfce8a0ff, 0xfbe49cff, 0xfae29dff, 0xf8e19eff, 0xf5e5abff, 
    0xebdfafff, 0xd2c496ff, 0x82795dff, 0x62625cff, 0x898b87ff, 0x8e908bff, 
    0xb9bbb7ff, 0xdfe0deff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe0e1dfff, 0xa9aaa8ff, 
    0x47443cff, 0x5d4b28ff, 0x8f723bff, 0xaa863dff, 0xbe9440ff, 0xcda04bff, 
    0xd2a44bff, 0xd7ab56ff, 0xdab061ff, 0xdfbb78ff, 0xdeb974ff, 0xd9b167ff, 
    0xd7ac5bff, 0xd4a94fff, 0xd8ae58ff, 0xd7af58ff, 0xd6ac56ff, 0xd9b367ff, 
    0xdbb771ff, 0xdcb975ff, 0xdbb772ff, 0xd8b368ff, 0xd4ab56ff, 0xd3aa57ff, 
    0xd2a74dff, 0xd3a950ff, 0xd6ad56ff, 0xdfba64ff, 0xf0d78bff, 0xf4e2a4ff, 
    0xf4e3adff, 0xf0dca5ff, 0xedd69bff, 0xe5c883ff, 0xdcb86bff, 0xd8b466ff, 
    0xd2ad5bff, 0xd1aa59ff, 0xca9e47ff, 0xc5963aff, 0xc29131ff, 0xbe8a23ff, 
    0xbc8722ff, 0xb68318ff, 0xb57f12ff, 0xb78418ff, 0xbb881cff, 0xc89d3fff, 
    0xd9bb77ff, 0xdcbf87ff, 0xdbc089ff, 0xd9bd86ff, 0xd4b675ff, 0xd0af63ff, 
    0xd0ae5fff, 0xd1b168ff, 0xefce91ff, 0xefc87eff, 0xf4d18eff, 0xf1ce8aff, 
    0xf3d392ff, 0xf0cc87ff, 0xedbf71ff, 0xf6d69dff, 0xf7d8a0ff, 0xf5d595ff, 
    0xf2cf88ff, 0xf3cb79ff, 0xf5ce6bff, 0xf7d16cff, 0xf9d574ff, 0xf8d473ff, 
    0xf8d77eff, 0xfbe18aff, 0xfceeb1ff, 0xfcedaaff, 0xfceea6ff, 0xfce998ff, 
    0xfce992ff, 0xfcea99ff, 0xfce791ff, 0xfceb9dff, 0xfce893ff, 0xfce790ff, 
    0xfce795ff, 0xfce895ff, 0xfce592ff, 0xfce594ff, 0xfbe48eff, 0xfce48dff, 
    0xfce696ff, 0xfce89bff, 0xfce79dff, 0xfce69eff, 0xfce499ff, 0xfce69dff, 
    0xfce9a3ff, 0xfceaa0ff, 0xfceba5ff, 0xfceca9ff, 0xfce8a5ff, 0xf9e6a3ff, 
    0xf4e1a4ff, 0xe9daa2ff, 0xcec191ff, 0x766f54ff, 0x60615dff, 0x8a8c88ff, 
    0x8f918cff, 0xbdbebbff, 0xe0e1dfff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xaeb0aeff, 0x4b473fff, 0x594621ff, 0x8a6b2fff, 0xa9843dff, 0xbe9443ff, 
    0xca9c41ff, 0xd5a853ff, 0xdaaf60ff, 0xdaad5bff, 0xd7ab59ff, 0xd3a752ff, 
    0xd6aa57ff, 0xd9ad60ff, 0xd3a44eff, 0xd1a54bff, 0xd2a342ff, 0xd3a649ff, 
    0xd4a950ff, 0xd6ac53ff, 0xdbb361ff, 0xd9af60ff, 0xd5ab57ff, 0xd5ab55ff, 
    0xd3a958ff, 0xcea451ff, 0xcfa64fff, 0xcfa446ff, 0xd2a74dff, 0xd2a343ff, 
    0xe4c26eff, 0xf2dc9aff, 0xf4e2a8ff, 0xf1dda6ff, 0xead291ff, 0xe8cb88ff, 
    0xe0be72ff, 0xdab76bff, 0xd3ac5aff, 0xcfa652ff, 0xc99e45ff, 0xc79d46ff, 
    0xc89c45ff, 0xc39234ff, 0xb9851eff, 0xb68117ff, 0xb47f13ff, 0xb68216ff, 
    0xb68318ff, 0xc2922eff, 0xd5b46dff, 0xdabd7fff, 0xdbbd82ff, 0xd9bc83ff, 
    0xd5b877ff, 0xd1af61ff, 0xd2b26cff, 0xd1b36fff, 0xf1d298ff, 0xf1cf90ff, 
    0xf6d9a5ff, 0xf3d393ff, 0xf2cf89ff, 0xf0ca80ff, 0xedc170ff, 0xf0c87eff, 
    0xefc77eff, 0xf1cb84ff, 0xf4d294ff, 0xf0c87bff, 0xf1c662ff, 0xf0c762ff, 
    0xe9bc54ff, 0xe4b03fff, 0xe1ab37ff, 0xf6d372ff, 0xfce79cff, 0xfceaa1ff, 
    0xfceb9eff, 0xfce894ff, 0xfce997ff, 0xfce792ff, 0xfce895ff, 0xfce99aff, 
    0xfce794ff, 0xfce795ff, 0xfce693ff, 0xfce58aff, 0xfce68dff, 0xfce798ff, 
    0xfbe697ff, 0xfce48fff, 0xfce695ff, 0xfce595ff, 0xfce79cff, 0xfce69bff, 
    0xfce8a0ff, 0xfceba7ff, 0xfcedabff, 0xfceda9ff, 0xfceba9ff, 0xfbe9a3ff, 
    0xfae9abff, 0xf4e4a9ff, 0xe6d397ff, 0xcab781ff, 0x736c53ff, 0x646461ff, 
    0x8b8d88ff, 0x90928dff, 0xc0c2bfff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xb3b5b3ff, 0x4f4c43ff, 0x57421cff, 0x87672cff, 0xa47d32ff, 
    0xb98e38ff, 0xc89940ff, 0xcfa043ff, 0xd2a247ff, 0xd6a850ff, 0xd7ab57ff, 
    0xd7aa56ff, 0xd3a347ff, 0xd3a44bff, 0xd3a550ff, 0xd0a248ff, 0xd1a249ff, 
    0xcf9f43ff, 0xce9c3fff, 0xcf9e3dff, 0xd3a343ff, 0xd4a747ff, 0xd4a644ff, 
    0xd4a849ff, 0xd3a749ff, 0xd2a853ff, 0xd1a750ff, 0xcfa348ff, 0xcfa243ff, 
    0xcfa03fff, 0xcd9e3dff, 0xcc9e3fff, 0xd8b15bff, 0xeacd82ff, 0xf2dea5ff, 
    0xedd69bff, 0xedd497ff, 0xe3c47eff, 0xd8b466ff, 0xd5b064ff, 0xcea755ff, 
    0xcaa04bff, 0xc89f4cff, 0xc89e4cff, 0xc5983fff, 0xb68218ff, 0xb47e0dff, 
    0xb37d0dff, 0xb17d0eff, 0xb27d0fff, 0xb7851dff, 0xcfac5bff, 0xd9bc7cff, 
    0xd9bd83ff, 0xd9bc86ff, 0xd7ba7eff, 0xd1b169ff, 0xd5b679ff, 0xd2b372ff, 
    0xf4d7a1ff, 0xf0cc8eff, 0xf3d193ff, 0xf5d691ff, 0xf9da9aff, 0xf7d795ff, 
    0xf3d085ff, 0xf1c977ff, 0xf0c77bff, 0xf2ce8dff, 0xf2ce8aff, 0xefc57aff, 
    0xecbc67ff, 0xe8b554ff, 0xe1aa3bff, 0xdda128ff, 0xdb9c1bff, 0xe4af3aff, 
    0xf7da7fff, 0xfae697ff, 0xfce79aff, 0xfcea9fff, 0xfceba2ff, 0xfce693ff, 
    0xfce48cff, 0xfbe389ff, 0xfce590ff, 0xfce490ff, 0xfce48fff, 0xfce590ff, 
    0xfce695ff, 0xfce593ff, 0xfce698ff, 0xfbdf87ff, 0xfbe38fff, 0xfbe393ff, 
    0xfce597ff, 0xfce79aff, 0xfce9a4ff, 0xfceaa6ff, 0xfceaa8ff, 0xfbe79fff, 
    0xfae39bff, 0xf8e5a0ff, 0xf1db93ff, 0xe5cf8dff, 0xc8b77fff, 0x736b51ff, 
    0x686966ff, 0x8b8d89ff, 0x92948fff, 0xc4c5c2ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xb9bab8ff, 0x504d47ff, 0x503c14ff, 0x825f20ff, 
    0x9e7526ff, 0xb5862dff, 0xc8983eff, 0xd1a34dff, 0xd4a44fff, 0xd19f41ff, 
    0xd2a145ff, 0xd2a144ff, 0xd1a142ff, 0xd09f41ff, 0xcf9d3fff, 0xcc9a3aff, 
    0xca9838ff, 0xcb9a3dff, 0xc89532ff, 0xcb9737ff, 0xcb9938ff, 0xcc9936ff, 
    0xcf9e3fff, 0xcd9c3cff, 0xca9932ff, 0xc9972eff, 0xcd9c39ff, 0xcd9f3fff, 
    0xce9e3dff, 0xcd9f3cff, 0xcc9e3cff, 0xcc9e42ff, 0xc89736ff, 0xc7932cff, 
    0xd3a546ff, 0xeacc80ff, 0xeed595ff, 0xedd59bff, 0xe2c57fff, 0xd6b364ff, 
    0xd5b165ff, 0xcea553ff, 0xc99f4aff, 0xc59b45ff, 0xc59942ff, 0xc19234ff, 
    0xb7831cff, 0xb47e11ff, 0xb47e0fff, 0xb58118ff, 0xb0790bff, 0xb47e14ff, 
    0xc9a046ff, 0xd8ba77ff, 0xdbc08cff, 0xdac192ff, 0xd8be8bff, 0xd3b576ff, 
    0xd4b678ff, 0xcfad63ff, 0xf5dba6ff, 0xf0cf8eff, 0xf2cf8aff, 0xf8db9eff, 
    0xf8dba2ff, 0xf8da9cff, 0xf4d389ff, 0xf2cc7dff, 0xf4d18eff, 0xf3d090ff, 
    0xf2c97eff, 0xedc274ff, 0xedbc69ff, 0xe8b558ff, 0xe2ad42ff, 0xdda32dff, 
    0xda9a19ff, 0xda9918ff, 0xf0ca6aff, 0xfce89fff, 0xfce8a2ff, 0xfce89cff, 
    0xfbe391ff, 0xfbe28cff, 0xfbe18cff, 0xfbe38fff, 0xfce799ff, 0xfbe18dff, 
    0xfbe28eff, 0xfbe28cff, 0xfbe493ff, 0xfbe697ff, 0xfbe18cff, 0xfbde87ff, 
    0xfbe294ff, 0xfbe49aff, 0xfbe59bff, 0xfae298ff, 0xfae295ff, 0xf9e297ff, 
    0xfae6a2ff, 0xf8e296ff, 0xf5de95ff, 0xefd996ff, 0xe3ce8aff, 0xbdaa6fff, 
    0x696149ff, 0x6d6e6aff, 0x8b8d89ff, 0x949691ff, 0xc7c8c5ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xbebfbdff, 0x555552ff, 0x53452bff, 
    0x8a6f3dff, 0xa78546ff, 0xb88b38ff, 0xbc8722ff, 0xc68e2aff, 0xc99432ff, 
    0xcc9837ff, 0xd09c41ff, 0xcf9a3bff, 0xd0a042ff, 0xd3a54bff, 0xd1a347ff, 
    0xce9c3eff, 0xc99536ff, 0xc69130ff, 0xc89436ff, 0xc89536ff, 0xc79435ff, 
    0xc89637ff, 0xc89534ff, 0xca993aff, 0xc79531ff, 0xc48e23ff, 0xc28c1fff, 
    0xc18a20ff, 0xbd841cff, 0xbb821eff, 0xba7f22ff, 0xb0741dff, 0xad7d24ff, 
    0xb57f1dff, 0xaf781bff, 0xad771aff, 0xc99941ff, 0xeacc80ff, 0xebd293ff, 
    0xe3c782ff, 0xd9b76bff, 0xd3ad5eff, 0xcea652ff, 0xc9a24dff, 0xc49843ff, 
    0xc2953cff, 0xbd8b2eff, 0xb78423ff, 0xb4801dff, 0xb57f18ff, 0xb57f16ff, 
    0xb37d10ff, 0xb58013ff, 0xc59a40ff, 0xd5b267ff, 0xdbbe82ff, 0xdbbf8aff, 
    0xd8bb82ff, 0xd3b778ff, 0xd2b36fff, 0xcfac5fff, 0xf7dda9ff, 0xf5d99dff, 
    0xf5d99eff, 0xf5d89dff, 0xf4d697ff, 0xf4d389ff, 0xf1ca7bff, 0xf1ca7fff, 
    0xf2cd8bff, 0xf2ce8cff, 0xf1ca81ff, 0xeec370ff, 0xedbe67ff, 0xe7b353ff, 
    0xe2aa3cff, 0xdda32fff, 0xda9c1eff, 0xd99b20ff, 0xe9bc55ff, 0xfbe499ff, 
    0xfae49aff, 0xfae190ff, 0xf9dd84ff, 0xf9db81ff, 0xf9dd89ff, 0xfade88ff, 
    0xfadd87ff, 0xfad97bff, 0xfadb82ff, 0xfbdf8eff, 0xfbe08cff, 0xfbe08cff, 
    0xfbe192ff, 0xfbe297ff, 0xfadf91ff, 0xf9df91ff, 0xfadf8fff, 0xf9de91ff, 
    0xfae39bff, 0xf8e197ff, 0xf5da8aff, 0xf1d789ff, 0xead087ff, 0xdfc57dff, 
    0xb8a46cff, 0x615a45ff, 0x6d6e6cff, 0x8c8e89ff, 0x969893ff, 0xcacbc9ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xc2c3c1ff, 0x5b5b58ff, 
    0x554d3dff, 0x908160ff, 0xb39e73ff, 0xc9ad79ff, 0xd6b378ff, 0xce9f4bff, 
    0xc58b27ff, 0xc1841aff, 0xc2881cff, 0xc58d29ff, 0xc88f2eff, 0xce9a3dff, 
    0xd09f45ff, 0xcf9d40ff, 0xd1a046ff, 0xd0a14eff, 0xc9943aff, 0xc48b29ff, 
    0xc89234ff, 0xc79233ff, 0xc58f2dff, 0xc58d2cff, 0xc0871fff, 0xba8014ff, 
    0xba7f14ff, 0xb4720cff, 0xad6907ff, 0xa45d07ff, 0x944304ff, 0x7b1d05ff, 
    0x4e0b03ff, 0x281204ff, 0x351d02ff, 0x291302ff, 0x461403ff, 0x803308ff, 
    0xbd9249ff, 0xe8cc84ff, 0xe8ce8dff, 0xddbc72ff, 0xd5b165ff, 0xcea95bff, 
    0xcba558ff, 0xc69c49ff, 0xc0923bff, 0xb8882dff, 0xb5852cff, 0xb6872fff, 
    0xb58223ff, 0xb17b12ff, 0xb37d12ff, 0xb6821cff, 0xba8727ff, 0xcba44eff, 
    0xd5b772ff, 0xd7ba7eff, 0xd5b779ff, 0xd3b578ff, 0xcfb16dff, 0xcfae65ff, 
    0xf8dfa2ff, 0xf9e1a7ff, 0xf6db9dff, 0xf7dc9dff, 0xf6da9aff, 0xf5d38fff, 
    0xf2ca7fff, 0xf3ce81ff, 0xf3ce88ff, 0xf2cb84ff, 0xf0c77eff, 0xeec171ff, 
    0xeab95eff, 0xeab95fff, 0xe3af48ff, 0xe0a736ff, 0xdea531ff, 0xda9f26ff, 
    0xebc15fff, 0xf8df91ff, 0xf9de8dff, 0xf9df8fff, 0xf8dc85ff, 0xf7d87eff, 
    0xf8d97fff, 0xf8dc87ff, 0xf6d475ff, 0xf6d169ff, 0xf8d87cff, 0xfadb86ff, 
    0xf9da81ff, 0xf8d983ff, 0xf8dc8dff, 0xf8db8bff, 0xf8db8aff, 0xf5d783ff, 
    0xf8dd91ff, 0xf8de94ff, 0xf7dd93ff, 0xf5d78dff, 0xeecf7cff, 0xe6c87bff, 
    0xdbbd73ff, 0xb39d5fff, 0x5f5944ff, 0x717370ff, 0x8c8e89ff, 0x989a95ff, 
    0xcdceccff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xc6c7c5ff, 
    0x61615eff, 0x544d3eff, 0x8f8266ff, 0xb4a583ff, 0xcfbb91ff, 0xdbbd82ff, 
    0xddba76ff, 0xddb469ff, 0xd2a250ff, 0xc48726ff, 0xc28624ff, 0xc18622ff, 
    0xc28721ff, 0xc99333ff, 0xcc9a3dff, 0xcc9533ff, 0xcd9839ff, 0xcf9d47ff, 
    0xcb963eff, 0xc58b23ff, 0xc58d27ff, 0xc38d27ff, 0xc0881fff, 0xbd8218ff, 
    0xb6760bff, 0xb16b05ff, 0xae6606ff, 0xac6205ff, 0xaa6204ff, 0xa35904ff, 
    0x964504ff, 0x791e04ff, 0x5f1203ff, 0x3e0902ff, 0x240001ff, 0x430b02ff, 
    0x732b04ff, 0x833402ff, 0x995c24ff, 0xe3c57cff, 0xead293ff, 0xe1c280ff, 
    0xd6b56cff, 0xcda85aff, 0xc79e4cff, 0xc0933cff, 0xbb8f38ff, 0xb4852bff, 
    0xb6852dff, 0xb7872dff, 0xb27e1bff, 0xb07810ff, 0xb07810ff, 0xb07813ff, 
    0xac7511ff, 0xc19436ff, 0xd1b065ff, 0xd3b677ff, 0xd1b36fff, 0xcfb26dff, 
    0xcfb06aff, 0xcfaf68ff, 0xf8de9fff, 0xf9e0a7ff, 0xf8dfa5ff, 0xf8dda5ff, 
    0xf6d999ff, 0xf2d18eff, 0xf1cb82ff, 0xf1cc83ff, 0xf0c883ff, 0xf0c57dff, 
    0xefc477ff, 0xedc06dff, 0xecbd65ff, 0xe8b75eff, 0xe4b24fff, 0xe4b251ff, 
    0xdda635ff, 0xdda42dff, 0xefc86fff, 0xf6da89ff, 0xf6da85ff, 0xf6d987ff, 
    0xf6d985ff, 0xf5d67dff, 0xf4d373ff, 0xf4d16eff, 0xf2cb60ff, 0xf3cc62ff, 
    0xf5d173ff, 0xf5d171ff, 0xf5d377ff, 0xf5d47fff, 0xf6d47eff, 0xf5d684ff, 
    0xf5d786ff, 0xf4d583ff, 0xf4d78aff, 0xf6db90ff, 0xf3d586ff, 0xeed183ff, 
    0xe7cc85ff, 0xd9bf7cff, 0xb09b61ff, 0x615a46ff, 0x757774ff, 0x8c8e89ff, 
    0x9b9c98ff, 0xd0d1ceff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xcacbc9ff, 0x656563ff, 0x524b3dff, 0x8d7f62ff, 0xb2a27fff, 0xcdba91ff, 
    0xdfca9eff, 0xe7cd9cff, 0xe1bf7eff, 0xdeb567ff, 0xdeb569ff, 0xd2a152ff, 
    0xc68d38ff, 0xbf852bff, 0xbd801cff, 0xc38a26ff, 0xc89132ff, 0xc78f2cff, 
    0xc8902bff, 0xc99330ff, 0xc68d28ff, 0xbf8318ff, 0xbc8016ff, 0xbe851eff, 
    0xbc8014ff, 0xb77508ff, 0xb37004ff, 0xb16c05ff, 0xaf6905ff, 0xae6c05ff, 
    0xaf6e05ff, 0xac6a05ff, 0xa96205ff, 0x9e5704ff, 0x8f4204ff, 0x7d2b03ff, 
    0x7e2503ff, 0x8b3904ff, 0x904804ff, 0x8b4203ff, 0x874408ff, 0xc8a14eff, 
    0xe8ca82ff, 0xe6cb8fff, 0xd8b973ff, 0xcfad63ff, 0xc6a053ff, 0xbf933eff, 
    0xba8e38ff, 0xb5862cff, 0xb48127ff, 0xb27c1bff, 0xac730cff, 0xa66605ff, 
    0xa66905ff, 0xa86f08ff, 0xa8700aff, 0xb48323ff, 0xc6a050ff, 0xc8a760ff, 
    0xc8a458ff, 0xc8a65aff, 0xccac65ff, 0xccab66ff, 0xf6db9cff, 0xf9dda1ff, 
    0xf7dc9eff, 0xf3d495ff, 0xf2d08aff, 0xf0c97dff, 0xf0cb88ff, 0xefca8aff, 
    0xefc682ff, 0xeec275ff, 0xebbb67ff, 0xebbb65ff, 0xebbc69ff, 0xe8b55aff, 
    0xe3ae49ff, 0xe1ab40ff, 0xe0aa3bff, 0xe0ab3bff, 0xefc869ff, 0xf5d783ff, 
    0xf5d883ff, 0xf5d985ff, 0xf3d47aff, 0xf3d06fff, 0xf3d070ff, 0xf2ce72ff, 
    0xf0c860ff, 0xf0c85fff, 0xf1c968ff, 0xf0c65fff, 0xf1cb6cff, 0xf1cb72ff, 
    0xf2cc71ff, 0xf3d17bff, 0xf3d381ff, 0xf2d380ff, 0xf2d383ff, 0xf2d68cff, 
    0xefd38eff, 0xe4c67bff, 0xd5b771ff, 0xa28c57ff, 0x5a5444ff, 0x797b78ff, 
    0x8b8d89ff, 0x9e9f9bff, 0xd2d4d1ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xcecfcdff, 0x6a6a68ff, 0x4b4538ff, 0x897c5fff, 0xaf9d77ff, 
    0xcbb689ff, 0xddc594ff, 0xe5cc9aff, 0xeace9bff, 0xe6c68cff, 0xdeb971ff, 
    0xdfbc78ff, 0xd9b06aff, 0xcd9b47ff, 0xcc9a4cff, 0xc18528ff, 0xbe8119ff, 
    0xc78f35ff, 0xc79033ff, 0xc48b23ff, 0xc28a22ff, 0xc1871dff, 0xbe821bff, 
    0xbe801bff, 0xb8790dff, 0xb37107ff, 0xb16c06ff, 0xb97b15ff, 0xc18a26ff, 
    0xba7e13ff, 0xb77c12ff, 0xb67b16ff, 0xb1730cff, 0xb0710cff, 0xa96a08ff, 
    0xa26004ff, 0x9d5a04ff, 0x985104ff, 0x954e04ff, 0x954f04ff, 0x965104ff, 
    0x934e03ff, 0xa26716ff, 0xdbb769ff, 0xe5c989ff, 0xdabb77ff, 0xd2b16cff, 
    0xcba964ff, 0xc39c4fff, 0xbb8e39ff, 0xb38023ff, 0xb07c1dff, 0xae761cff, 
    0xa76a0cff, 0xa26606ff, 0xa76e10ff, 0xa8700eff, 0xa97210ff, 0xaf7c20ff, 
    0xb5852aff, 0xba8c33ff, 0xb98930ff, 0xc29848ff, 0xcdac6bff, 0xceac6cff, 
    0xf7dda7ff, 0xf9dfa9ff, 0xf9e1aaff, 0xf8dba1ff, 0xf3d18dff, 0xf0c87cff, 
    0xeec780ff, 0xecc279ff, 0xedc47eff, 0xeabb6bff, 0xe9b660ff, 0xe9b964ff, 
    0xe8ba68ff, 0xe6b259ff, 0xe3ae47ff, 0xe0aa3eff, 0xdfaa3dff, 0xe1ab3fff, 
    0xf0ca6dff, 0xf4d989ff, 0xf5d98aff, 0xf4d681ff, 0xf2d177ff, 0xf2d076ff, 
    0xf2ce6fff, 0xf0c968ff, 0xeec560ff, 0xedc45bff, 0xeec562ff, 0xefc564ff, 
    0xeec566ff, 0xeec365ff, 0xedc363ff, 0xefc869ff, 0xf0cf7bff, 0xf1d07dff, 
    0xefcd7aff, 0xeac670ff, 0xe2bf6eff, 0xd3b162ff, 0x9d834aff, 0x565042ff, 
    0x7b7d7aff, 0x8b8d88ff, 0xa1a29eff, 0xd5d6d4ff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xd1d2d0ff, 0x70706eff, 0x484131ff, 0x847553ff, 
    0xad996fff, 0xc9b081ff, 0xdabd88ff, 0xe4c994ff, 0xe8cc94ff, 0xe7c88fff, 
    0xe2bf80ff, 0xdeb870ff, 0xdfbb78ff, 0xdcb470ff, 0xd4a856ff, 0xd7ad65ff, 
    0xd0a357ff, 0xc48d33ff, 0xc28829ff, 0xc48b2bff, 0xc58d2dff, 0xbf8520ff, 
    0xbc7e14ff, 0xb97b0fff, 0xb97b0fff, 0xb47106ff, 0xb4710cff, 0xc0882bff, 
    0xcea14dff, 0xd1a453ff, 0xc69237ff, 0xbf8a28ff, 0xbd8625ff, 0xba8123ff, 
    0xb67f20ff, 0xb27818ff, 0xad7110ff, 0xab6f10ff, 0xa56607ff, 0xa36306ff, 
    0xa26205ff, 0xa05d04ff, 0x9d5904ff, 0xa05e06ff, 0xcca14aff, 0xe2c177ff, 
    0xdcbd7cff, 0xd3b26fff, 0xc7a256ff, 0xc6a15aff, 0xbe9345ff, 0xb48127ff, 
    0xad7617ff, 0x9c5d0aff, 0x945004ff, 0x9a5b06ff, 0xa66e14ff, 0xa97315ff, 
    0xac781bff, 0xae7c22ff, 0xaf7e23ff, 0xb18028ff, 0xb07d23ff, 0xb6842aff, 
    0xc29646ff, 0xc6a056ff, 0xf8e1adff, 0xf8dfacff, 0xf6d9a1ff, 0xf7d99eff, 
    0xf2ce88ff, 0xf2cd89ff, 0xf1cf8eff, 0xeabc6cff, 0xebbd6fff, 0xe8b660ff, 
    0xe8b865ff, 0xe8b966ff, 0xe6b55fff, 0xe6b45eff, 0xe4af52ff, 0xe0aa42ff, 
    0xdea93dff, 0xe0ab3dff, 0xf0ca6dff, 0xf4d686ff, 0xf4d783ff, 0xf3d57eff, 
    0xf2d176ff, 0xf2cf74ff, 0xf0cc6aff, 0xf0c966ff, 0xf0ca71ff, 0xefc667ff, 
    0xeec463ff, 0xeec567ff, 0xe9bd55ff, 0xe8bb51ff, 0xeabd59ff, 0xecc15fff, 
    0xefc972ff, 0xedc872ff, 0xe9c473ff, 0xe1bd6cff, 0xd1b063ff, 0x99824fff, 
    0x585346ff, 0x7e807eff, 0x8b8d88ff, 0xa4a5a1ff, 0xd7d8d6ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xd4d5d3ff, 0x777775ff, 0x484335ff, 
    0x827554ff, 0xa9966aff, 0xc5ad79ff, 0xd5b878ff, 0xdfbe7bff, 0xe4c589ff, 
    0xe5c78bff, 0xe3c483ff, 0xdfbb77ff, 0xe1bb77ff, 0xdebb77ff, 0xdcb76fff, 
    0xdcb56fff, 0xdfbb80ff, 0xddb87bff, 0xd4a964ff, 0xc1872bff, 0xbe8019ff, 
    0xbf841bff, 0xbd8018ff, 0xb8780aff, 0xb7780aff, 0xb87b0fff, 0xb47206ff, 
    0xb06c08ff, 0xcc9b43ff, 0xd8b060ff, 0xd5ad5fff, 0xd3a961ff, 0xca9c48ff, 
    0xc38e36ff, 0xbf8b31ff, 0xbf8c36ff, 0xbc8932ff, 0xb37a15ff, 0xb27917ff, 
    0xab6f09ff, 0xaa6d08ff, 0xa96e09ff, 0xa96a07ff, 0xa76806ff, 0xa96a05ff, 
    0xc6973cff, 0xe1c178ff, 0xdabb79ff, 0xd0ae66ff, 0xc8a35bff, 0xc5a159ff, 
    0xbe9245ff, 0xb37e28ff, 0xa3640aff, 0x975203ff, 0x874003ff, 0x925005ff, 
    0xa7690dff, 0xab7415ff, 0xac791bff, 0xac7c22ff, 0xb0812dff, 0xaf7c26ff, 
    0xb1802aff, 0xb07b1eff, 0xaf7615ff, 0xba852aff, 0xf7e1afff, 0xf8dfabff, 
    0xf8dba4ff, 0xf8dda3ff, 0xf6d79bff, 0xf6d69dff, 0xf2ce8fff, 0xeec47cff, 
    0xecc075ff, 0xe9b961ff, 0xe8b865ff, 0xe8b967ff, 0xe4b25cff, 0xe4b058ff, 
    0xe4b055ff, 0xe1ac4bff, 0xdea839ff, 0xe5b348ff, 0xf2ce70ff, 0xf3d685ff, 
    0xf4da8fff, 0xf3d789ff, 0xf1d382ff, 0xf1d07cff, 0xefcb74ff, 0xefca76ff, 
    0xefc76bff, 0xecc15cff, 0xebc05bff, 0xe8bd56ff, 0xe6ba4fff, 0xe7ba51ff, 
    0xe9bf58ff, 0xeac061ff, 0xeabf61ff, 0xe5bb5eff, 0xdeba70ff, 0xcfb067ff, 
    0x977f4bff, 0x5b574cff, 0x818380ff, 0x8b8d88ff, 0xa7a9a5ff, 0xd9dad8ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xd6d7d5ff, 0x7d7e7bff, 
    0x484235ff, 0x7c6b4bff, 0xa8956bff, 0xc7b182ff, 0xd8bd86ff, 0xdec17fff, 
    0xdfc07bff, 0xe0bf7bff, 0xe1c07eff, 0xe2c281ff, 0xe1be7aff, 0xe1be7bff, 
    0xe1c17eff, 0xddba72ff, 0xddb976ff, 0xe0bd83ff, 0xdeb97cff, 0xd9b067ff, 
    0xcb9945ff, 0xbd7f16ff, 0xb9780dff, 0xb8770cff, 0xb77609ff, 0xb57206ff, 
    0xb67509ff, 0xb37005ff, 0xb16e07ff, 0xce9c41ff, 0xd8b05fff, 0xd7b060ff, 
    0xd5ad5eff, 0xd2a95aff, 0xcb9e4cff, 0xc99d4cff, 0xc29033ff, 0xc08e33ff, 
    0xba8526ff, 0xb47d17ff, 0xb17914ff, 0xaf7410ff, 0xaf7514ff, 0xad730bff, 
    0xac7107ff, 0xac6f07ff, 0xc59635ff, 0xe3c175ff, 0xdabb77ff, 0xd0ae68ff, 
    0xcfad6eff, 0xc8a25fff, 0xb88530ff, 0xa86809ff, 0xa8680bff, 0xae7217ff, 
    0xa5670aff, 0xac7012ff, 0xb47e27ff, 0xb88634ff, 0xb3822dff, 0xac7b25ff, 
    0xae7e2aff, 0xaf802cff, 0xb2832fff, 0xb0802cff, 0xb07f23ff, 0xb48126ff, 
    0xf7dfaaff, 0xf8dda4ff, 0xf7dda3ff, 0xf6d89dff, 0xf3d295ff, 0xf2cf93ff, 
    0xeec785ff, 0xecc17aff, 0xebbe73ff, 0xe9bb69ff, 0xe8b662ff, 0xe8b864ff, 
    0xe6b35cff, 0xe3ad52ff, 0xe2ac4eff, 0xe1ab4cff, 0xdea941ff, 0xe6b54aff, 
    0xf1cb6fff, 0xf4da90ff, 0xf4db92ff, 0xf3da90ff, 0xf2d58dff, 0xf0d28bff, 
    0xefcf85ff, 0xeecb79ff, 0xedc466ff, 0xeabf5eff, 0xe7bc59ff, 0xe7ba58ff, 
    0xe6ba54ff, 0xe8bc56ff, 0xebc166ff, 0xe9bf62ff, 0xe3bb5cff, 0xdbb661ff, 
    0xcdad69ff, 0x887548ff, 0x59564cff, 0x848682ff, 0x8b8d88ff, 0xaaaca8ff, 
    0xdbdcd9ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xd9dad8ff, 
    0x828382ff, 0x453f31ff, 0x7b6c4dff, 0x9e8551ff, 0xbca165ff, 0xd6bd83ff, 
    0xddbf7eff, 0xdebd75ff, 0xe0c179ff, 0xe1c27dff, 0xe0c07aff, 0xe4c687ff, 
    0xe3c282ff, 0xdcb66cff, 0xdeb971ff, 0xdab264ff, 0xdbb46aff, 0xddbb7cff, 
    0xdbb675ff, 0xd8ae66ff, 0xce9d48ff, 0xc89030ff, 0xc0851fff, 0xb8770bff, 
    0xb67204ff, 0xb36f04ff, 0xb57307ff, 0xb26e04ff, 0xbd811dff, 0xd9b063ff, 
    0xdebb7cff, 0xdab671ff, 0xdab46eff, 0xd6af66ff, 0xd4ad69ff, 0xcda251ff, 
    0xc5963dff, 0xc39035ff, 0xc2913bff, 0xbc872bff, 0xb9852aff, 0xb57f20ff, 
    0xb27d1aff, 0xb07912ff, 0xae7409ff, 0xac6d04ff, 0xbe8b26ff, 0xdfbc6cff, 
    0xd6b46dff, 0xd3b26fff, 0xd2b072ff, 0xc3994dff, 0xb2791bff, 0xad6f09ff, 
    0xb37918ff, 0xb67e20ff, 0xb88125ff, 0xbc882fff, 0xbd892fff, 0xc39648ff, 
    0xc39a50ff, 0xbd9346ff, 0xb38633ff, 0xb0802bff, 0xaf7c25ff, 0xb2822eff, 
    0xb28530ff, 0xb0822eff, 0xf4dda7ff, 0xf4da9dff, 0xf2d695ff, 0xf3d597ff, 
    0xf0cc89ff, 0xe9bc6dff, 0xe8ba6aff, 0xeabb6aff, 0xe9bb6aff, 0xeabd6cff, 
    0xe6b55dff, 0xe6b65cff, 0xe6b55aff, 0xe3ae52ff, 0xe0aa49ff, 0xe0a947ff, 
    0xe0a945ff, 0xe5b448ff, 0xefcc72ff, 0xf3d891ff, 0xf2d98fff, 0xf1d487ff, 
    0xf1d387ff, 0xf1d184ff, 0xefce7bff, 0xefcc7bff, 0xedca7cff, 0xebc97eff, 
    0xebc67bff, 0xeac274ff, 0xeac062ff, 0xe8bc57ff, 0xe6ba5aff, 0xe3ba60ff, 
    0xdab35bff, 0xc9a85cff, 0x836e3eff, 0x57544bff, 0x858784ff, 0x8c8e89ff, 
    0xaeb0acff, 0xdcdddbff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xdbdcdaff, 0x898a88ff, 0x423e33ff, 0x786a4cff, 0xa49067ff, 0xbfa56fff, 
    0xd2b477ff, 0xdcc082ff, 0xdebd76ff, 0xdcb766ff, 0xe0c075ff, 0xe3c787ff, 
    0xe2c481ff, 0xe1bf7aff, 0xdcb66aff, 0xdab161ff, 0xdcb971ff, 0xddba76ff, 
    0xdcb56cff, 0xdcb771ff, 0xdcb877ff, 0xd9b16cff, 0xcf9e4aff, 0xcc973eff, 
    0xc89231ff, 0xc48c2bff, 0xbe8020ff, 0xbd8124ff, 0xb47007ff, 0xb16c03ff, 
    0xc8943aff, 0xe0bc77ff, 0xddbb7fff, 0xdebd7fff, 0xe3c28dff, 0xdab676ff, 
    0xd5ad65ff, 0xd1a85cff, 0xcda254ff, 0xc79b49ff, 0xc5963fff, 0xc49642ff, 
    0xbf8e39ff, 0xba872cff, 0xb57f1cff, 0xb7801dff, 0xb67f1aff, 0xb27a10ff, 
    0xbe8922ff, 0xd4a745ff, 0xd8b46aff, 0xd7b776ff, 0xd0ab5fff, 0xc2923aff, 
    0xb88227ff, 0xb47c17ff, 0xb8821fff, 0xbd882aff, 0xc18e35ff, 0xc2923cff, 
    0xc3943fff, 0xc79c4fff, 0xc8a25cff, 0xc8a35fff, 0xc2984eff, 0xb4822dff, 
    0xb38532ff, 0xb88f41ff, 0xb98f44ff, 0xb88e48ff, 0xf4deadff, 0xf5dca6ff, 
    0xf2d593ff, 0xf1d38eff, 0xebc373ff, 0xe5b459ff, 0xe4af50ff, 0xe4b255ff, 
    0xebc176ff, 0xe8bc71ff, 0xe5b159ff, 0xe5b153ff, 0xe3af51ff, 0xe2ac53ff, 
    0xe1ac4fff, 0xe1ac4eff, 0xdea741ff, 0xe5b653ff, 0xf0ce79ff, 0xf2d791ff, 
    0xf3da98ff, 0xf2d794ff, 0xf0d288ff, 0xf0d083ff, 0xefd188ff, 0xf0d491ff, 
    0xeed08aff, 0xedce86ff, 0xedc97cff, 0xecc570ff, 0xecc46cff, 0xe7bc5fff, 
    0xe1b65aff, 0xd8b25dff, 0xc7a559ff, 0x816d40ff, 0x5b5952ff, 0x878986ff, 
    0x8c8e89ff, 0xb2b3b0ff, 0xdddedcff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xdcdddbff, 0x90918fff, 0x433e31ff, 0x71613bff, 0xa08b5dff, 
    0xbaa066ff, 0xd0b473ff, 0xdec080ff, 0xe1c17dff, 0xe4c583ff, 0xe3c47cff, 
    0xe2c27bff, 0xdebb71ff, 0xe0bd75ff, 0xdfbc73ff, 0xdbb460ff, 0xdcb362ff, 
    0xdebb73ff, 0xdfbc79ff, 0xe0bd77ff, 0xdcb56dff, 0xdcb672ff, 0xd9b06aff, 
    0xd1a251ff, 0xcf9f48ff, 0xce9d43ff, 0xce9d48ff, 0xc99440ff, 0xbd8229ff, 
    0xb16a07ff, 0xac6703ff, 0xc59034ff, 0xe0bd73ff, 0xe1c184ff, 0xe2c48aff, 
    0xe3c28bff, 0xddba7cff, 0xd7af64ff, 0xd4ae68ff, 0xcfa75cff, 0xcb9f4cff, 
    0xca9e4cff, 0xc89c4aff, 0xc1913bff, 0xbf8e36ff, 0xba8829ff, 0xba8628ff, 
    0xb98320ff, 0xb8821fff, 0xba8523ff, 0xc28d24ff, 0xcfa54eff, 0xcea65dff, 
    0xcaa155ff, 0xc79948ff, 0xc79748ff, 0xbe872bff, 0xbe8728ff, 0xc4943cff, 
    0xc79946ff, 0xc99a4aff, 0xc99c4cff, 0xc79744ff, 0xc79b4dff, 0xc49648ff, 
    0xc2964fff, 0xb9893dff, 0xb7893bff, 0xbe9753ff, 0xbd9956ff, 0xbc9955ff, 
    0xf3d9a2ff, 0xf5d99dff, 0xf3d798ff, 0xecc67aff, 0xe9bd65ff, 0xebc06aff, 
    0xecc06dff, 0xebbe68ff, 0xeec273ff, 0xebc074ff, 0xe7b763ff, 0xe7b763ff, 
    0xe4b35fff, 0xe4b262ff, 0xe4b363ff, 0xe2af5aff, 0xdea946ff, 0xe0ab41ff, 
    0xeec76bff, 0xf2d795ff, 0xf4d89aff, 0xf4d896ff, 0xefd388ff, 0xefd186ff, 
    0xf0d38dff, 0xf0d493ff, 0xefd392ff, 0xedcf89ff, 0xedce88ff, 0xebca7fff, 
    0xeac679ff, 0xe5c378ff, 0xdab76cff, 0xc5a45fff, 0x7e6b41ff, 0x5f5e56ff, 
    0x898b87ff, 0x8d8f8aff, 0xb5b7b3ff, 0xdedfddff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xdedfddff, 0x969796ff, 0x454034ff, 0x6b5a34ff, 
    0x987f4aff, 0xb39553ff, 0xc6a356ff, 0xd6b567ff, 0xdbb86aff, 0xdcb765ff, 
    0xddba6bff, 0xe0c077ff, 0xe2c178ff, 0xdeb968ff, 0xdcb869ff, 0xdebb6fff, 
    0xddb967ff, 0xdbb35eff, 0xd9b15eff, 0xdeb86fff, 0xddb66dff, 0xddb56bff, 
    0xdab166ff, 0xd5a957ff, 0xd1a14bff, 0xd3a551ff, 0xd4a756ff, 0xd2a552ff, 
    0xc99339ff, 0xb37008ff, 0xad6404ff, 0xac6403ff, 0xc89539ff, 0xe3c176ff, 
    0xe7c889ff, 0xe6c88bff, 0xe5c68dff, 0xdebd80ff, 0xd9b873ff, 0xd4ae66ff, 
    0xd3ac64ff, 0xcfa65bff, 0xcb9f4eff, 0xc69843ff, 0xc2913aff, 0xc2943eff, 
    0xc09238ff, 0xc1943fff, 0xbf8f39ff, 0xbd892aff, 0xb98520ff, 0xba8623ff, 
    0xbc8823ff, 0xc08d32ff, 0xc49744ff, 0xcb9e4eff, 0xd0a65aff, 0xcda255ff, 
    0xc99845ff, 0xcda252ff, 0xcd9f4cff, 0xc79335ff, 0xc8963cff, 0xc7953dff, 
    0xc79945ff, 0xc18f3eff, 0xbe8f46ff, 0xc29857ff, 0xc29a59ff, 0xbd9857ff, 
    0xbb9755ff, 0xbb9855ff, 0xf4dca8ff, 0xf4dba0ff, 0xf3d798ff, 0xecc77dff, 
    0xedc673ff, 0xefc878ff, 0xf1cc81ff, 0xf1cd85ff, 0xf0cb83ff, 0xedc475ff, 
    0xeab962ff, 0xe8b762ff, 0xe7b767ff, 0xe6b667ff, 0xe3b05aff, 0xdfa94aff, 
    0xdca53eff, 0xdca337ff, 0xeabd5eff, 0xf1d28bff, 0xf1d38eff, 0xf0d28aff, 
    0xeecf87ff, 0xefd18eff, 0xf1d294ff, 0xf0d394ff, 0xf0d395ff, 0xeed092ff, 
    0xeccd8aff, 0xeac980ff, 0xe4c277ff, 0xdcbe7eff, 0xc4ab75ff, 0x716346ff, 
    0x5f5f5aff, 0x8a8c88ff, 0x8d8f8bff, 0xb9bab7ff, 0xdfe0deff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xdfe0deff, 0x9c9d9bff, 0x474237ff, 
    0x6a5a35ff, 0x967d48ff, 0xb39454ff, 0xc7a355ff, 0xd3af5cff, 0xdbb765ff, 
    0xdfbb6cff, 0xe1bf74ff, 0xdeb96dff, 0xe5c683ff, 0xe3c27aff, 0xdfbb6eff, 
    0xe1c077ff, 0xe1be73ff, 0xddb767ff, 0xdbb25cff, 0xd9af59ff, 0xd7ac58ff, 
    0xd6aa56ff, 0xd8ac59ff, 0xdbb263ff, 0xd5a850ff, 0xd0a044ff, 0xd6a754ff, 
    0xd9af66ff, 0xd2a553ff, 0xc99234ff, 0xbb7c13ff, 0xad6503ff, 0xb77817ff, 
    0xdcb666ff, 0xe2c077ff, 0xe9cb8aff, 0xebcf99ff, 0xe6c991ff, 0xe1c283ff, 
    0xdcbc7aff, 0xd6af64ff, 0xd5ae65ff, 0xd4ad6aff, 0xcea455ff, 0xcca04eff, 
    0xc99f4dff, 0xc59a44ff, 0xc49743ff, 0xc69a47ff, 0xc69b49ff, 0xc3943fff, 
    0xc19138ff, 0xc19037ff, 0xc09035ff, 0xc69a47ff, 0xcb9f4fff, 0xcfa453ff, 
    0xd3ac61ff, 0xd8b476ff, 0xdab77aff, 0xdbb97cff, 0xdcbb79ff, 0xd7b36cff, 
    0xd2ab60ff, 0xcda352ff, 0xc99b49ff, 0xcba158ff, 0xc2944eff, 0xbf934cff, 
    0xc9a568ff, 0xbe9955ff, 0xbe9c5bff, 0xbe9e5cff, 0xf0d7a6ff, 0xf3dca7ff, 
    0xf3d79cff, 0xf1d08eff, 0xf1d290ff, 0xf0cc82ff, 0xf0ce85ff, 0xefcb82ff, 
    0xeec779ff, 0xebc06aff, 0xe8b758ff, 0xe6b254ff, 0xe3ae52ff, 0xe1ac50ff, 
    0xe2ac4dff, 0xdfa944ff, 0xdca439ff, 0xdea53bff, 0xe8b95cff, 0xefcc7dff, 
    0xefd18aff, 0xefd388ff, 0xefd68aff, 0xf1d691ff, 0xf2d699ff, 0xf2d49aff, 
    0xf0d294ff, 0xedcd8aff, 0xe8c57cff, 0xe3c279ff, 0xd8b976ff, 0xbfa268ff, 
    0x6e6246ff, 0x62625fff, 0x8b8d88ff, 0x8f908cff, 0xbcbebbff, 0xe0e1dfff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe0e1dfff, 0xa2a3a2ff, 
    0x44423aff, 0x655532ff, 0x937a43ff, 0xb49757ff, 0xc9a761ff, 0xd7b467ff, 
    0xdbb866ff, 0xdfba68ff, 0xdfbb6eff, 0xe0be74ff, 0xe0bf78ff, 0xdebb6fff, 
    0xdcb769ff, 0xe3c380ff, 0xe4c582ff, 0xe0bd72ff, 0xe0bd72ff, 0xdcb464ff, 
    0xdbb15eff, 0xdaad59ff, 0xd6aa53ff, 0xd9ad58ff, 0xdbb164ff, 0xd9ad5cff, 
    0xd6a752ff, 0xd7aa59ff, 0xd6aa5aff, 0xd1a14fff, 0xcb9538ff, 0xc0851cff, 
    0xb6750cff, 0xcc9b41ff, 0xe4c47eff, 0xe2bf71ff, 0xe5c377ff, 0xe8cc8fff, 
    0xe6c78cff, 0xe3c587ff, 0xdfbe7eff, 0xdab772ff, 0xd7b26eff, 0xd4ad69ff, 
    0xd0a654ff, 0xcea450ff, 0xcfaa61ff, 0xcca458ff, 0xcaa157ff, 0xc99f4eff, 
    0xca9f4eff, 0xc69944ff, 0xc69d4bff, 0xcca55cff, 0xcca459ff, 0xcca65cff, 
    0xc0944fff, 0xd2ad67ff, 0xd9b775ff, 0xdebd82ff, 0xe1c287ff, 0xe0c083ff, 
    0xe1c386ff, 0xe4c792ff, 0xe1c48fff, 0xd8b575ff, 0xd0a85eff, 0xcdaa66ff, 
    0xc2974bff, 0xbd9141ff, 0xbf994fff, 0xbc964cff, 0xc0a15fff, 0xbf9c5bff, 
    0xf0daacff, 0xf4e0afff, 0xf4ddaaff, 0xf4d89eff, 0xf4d79bff, 0xf2d594ff, 
    0xefce86ff, 0xeeca7dff, 0xecc574ff, 0xe9be5fff, 0xe8ba59ff, 0xe7b558ff, 
    0xe4af4fff, 0xe6b055ff, 0xe3af4fff, 0xe3af4fff, 0xe5b257ff, 0xe5b155ff, 
    0xe4b055ff, 0xeabd65ff, 0xeecc7fff, 0xefd084ff, 0xefd185ff, 0xf0d38eff, 
    0xf0d391ff, 0xefcf8aff, 0xecca82ff, 0xe9c885ff, 0xe3c27eff, 0xd9b874ff, 
    0xbea46fff, 0x6d6045ff, 0x666663ff, 0x8b8d89ff, 0x90928dff, 0xc0c1beff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xa8aaa8ff, 0x47453fff, 0x625436ff, 0x957e4cff, 0xb0914dff, 0xc9a95fff, 
    0xd7b56aff, 0xddbb6dff, 0xddbb6bff, 0xdeb966ff, 0xdeba6aff, 0xe0bd73ff, 
    0xe1c17aff, 0xe0c079ff, 0xdcb86bff, 0xddb96dff, 0xdfbb6dff, 0xddb765ff, 
    0xdeb766ff, 0xdbb05dff, 0xdbb05bff, 0xdcb25eff, 0xdcb360ff, 0xd7a950ff, 
    0xd6a74eff, 0xd6a750ff, 0xd4a44aff, 0xd2a044ff, 0xd1a047ff, 0xcf9c42ff, 
    0xc9932fff, 0xc18617ff, 0xc48c24ff, 0xddb460ff, 0xe5c279ff, 0xe5c275ff, 
    0xe3c072ff, 0xe5c680ff, 0xe5c787ff, 0xe4c688ff, 0xdebd78ff, 0xdcb974ff, 
    0xd9b371ff, 0xd6b06bff, 0xd5af64ff, 0xd0a551ff, 0xcda554ff, 0xcfa65aff, 
    0xcea961ff, 0xcaa052ff, 0xcba04eff, 0xcca354ff, 0xcda761ff, 0xd6b377ff, 
    0xd4af6dff, 0xd3ad66ff, 0xd2ae6eff, 0xdab97dff, 0xdfc287ff, 0xdebf7eff, 
    0xe2c487ff, 0xe6c78eff, 0xe5c88fff, 0xe5cc99ff, 0xe1c591ff, 0xdfc28eff, 
    0xddc18cff, 0xd7b77eff, 0xd1b173ff, 0xc7a35eff, 0xbd9549ff, 0xbc964bff, 
    0xc3a263ff, 0xc09f5dff, 0xf2ddb1ff, 0xf2ddadff, 0xf2d89eff, 0xf2d699ff, 
    0xf0d493ff, 0xf0d495ff, 0xefcd88ff, 0xeec674ff, 0xedc26dff, 0xecc26aff, 
    0xecc167ff, 0xe9bb62ff, 0xe9b75dff, 0xebbc6bff, 0xe8ba66ff, 0xe9b865ff, 
    0xe7b665ff, 0xe5b25bff, 0xe3ac4fff, 0xe2ab4bff, 0xebc371ff, 0xeecb7eff, 
    0xefcf88ff, 0xf0d391ff, 0xefd28eff, 0xedcd86ff, 0xe9cb86ff, 0xe2c07dff, 
    0xd6b167ff, 0xb99d64ff, 0x6d634eff, 0x6b6b68ff, 0x8c8d89ff, 0x91938fff, 
    0xc3c5c2ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xaeafaeff, 0x4b4943ff, 0x615437ff, 0x927c4cff, 0xb19659ff, 
    0xc7a75fff, 0xd8b568ff, 0xdcb967ff, 0xe0be6eff, 0xdfbc6cff, 0xddb765ff, 
    0xdfba69ff, 0xe1be74ff, 0xe0c177ff, 0xe2c37cff, 0xe0be72ff, 0xdbb763ff, 
    0xdeb768ff, 0xddb768ff, 0xdfba6fff, 0xdeb669ff, 0xdeb668ff, 0xdcb15eff, 
    0xdaad57ff, 0xd4a445ff, 0xd4a445ff, 0xd5a448ff, 0xd1a041ff, 0xcd9a36ff, 
    0xcb952fff, 0xc8932bff, 0xc78f26ff, 0xc2881cff, 0xd2a241ff, 0xe2ba66ff, 
    0xe6c47aff, 0xe5c57fff, 0xe6c57cff, 0xe6c783ff, 0xe4c483ff, 0xe1c17cff, 
    0xe3c482ff, 0xe1c184ff, 0xdcb974ff, 0xd9b36dff, 0xd7b16aff, 0xd6af68ff, 
    0xd3ab5fff, 0xd1a95cff, 0xd2ac65ff, 0xd1ac67ff, 0xd0a960ff, 0xd1aa65ff, 
    0xd6b275ff, 0xdbba81ff, 0xd5b06eff, 0xd5ae66ff, 0xdebb80ff, 0xe2c58fff, 
    0xe4c78fff, 0xe0c182ff, 0xe5c78eff, 0xe5c890ff, 0xe4c790ff, 0xe0c189ff, 
    0xe2c792ff, 0xe1ca9aff, 0xdfc799ff, 0xdcc293ff, 0xdac091ff, 0xd4b784ff, 
    0xc4a25eff, 0xbe9c57ff, 0xc1a05cff, 0xc3a461ff, 0xf1dcadff, 0xf2dcaaff, 
    0xf1d998ff, 0xf1d693ff, 0xf0d495ff, 0xf0d192ff, 0xeece89ff, 0xefce86ff, 
    0xeec87cff, 0xefcd86ff, 0xedc77aff, 0xecc373ff, 0xedc16eff, 0xeec37cff, 
    0xebbf74ff, 0xe8b764ff, 0xe7b461ff, 0xe6b463ff, 0xe5b15eff, 0xe1a949ff, 
    0xe2b154ff, 0xedca80ff, 0xf0d293ff, 0xeecf8cff, 0xedce8aff, 0xebcd8cff, 
    0xe3c481ff, 0xd5b46cff, 0xb39860ff, 0x5f5642ff, 0x6d6e6cff, 0x8c8e89ff, 
    0x939591ff, 0xc7c8c5ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xb4b5b3ff, 0x4f4e47ff, 0x5d4f32ff, 0x917b4aff, 
    0xae9253ff, 0xc5a55cff, 0xd2b060ff, 0xddb96cff, 0xddbb6eff, 0xdcb765ff, 
    0xdeb967ff, 0xddb866ff, 0xd8af57ff, 0xd9b059ff, 0xdbb45fff, 0xdab35cff, 
    0xddb765ff, 0xdeb868ff, 0xdfb96aff, 0xdeb666ff, 0xe1be76ff, 0xdfb86bff, 
    0xdbb25eff, 0xd9ad54ff, 0xd5a647ff, 0xd4a242ff, 0xd19f3eff, 0xd09d38ff, 
    0xce9a35ff, 0xcc9832ff, 0xc58c1cff, 0xc48b18ff, 0xc28618ff, 0xcb9530ff, 
    0xe2ba65ff, 0xe4c179ff, 0xe4c378ff, 0xe6c781ff, 0xe6c784ff, 0xe8ca8aff, 
    0xe7c98bff, 0xe5c585ff, 0xe7c88dff, 0xe4c488ff, 0xdfbe7dff, 0xddbb79ff, 
    0xd9b670ff, 0xd9b46cff, 0xd8b370ff, 0xd4af64ff, 0xd9b778ff, 0xd8b370ff, 
    0xd6af67ff, 0xd4af67ff, 0xd9b576ff, 0xddbc81ff, 0xd9b674ff, 0xd9b66fff, 
    0xdfbd7fff, 0xe1c184ff, 0xe6c892ff, 0xe4c589ff, 0xe2c283ff, 0xe3c48bff, 
    0xe1c48dff, 0xe1c895ff, 0xe6d1a3ff, 0xe4cea0ff, 0xddc390ff, 0xddc494ff, 
    0xddc498ff, 0xd9c090ff, 0xcaaa6aff, 0xc4a466ff, 0xc7a96dff, 0xc7a968ff, 
    0xefd8a6ff, 0xf1d9a1ff, 0xefd48fff, 0xf0d591ff, 0xefd290ff, 0xedcc89ff, 
    0xedcd89ff, 0xf0d191ff, 0xeecc89ff, 0xe5be73ff, 0xe1b460ff, 0xe4b662ff, 
    0xedc57cff, 0xebc582ff, 0xe8bd72ff, 0xe7b45dff, 0xe8b663ff, 0xe9bb70ff, 
    0xe5b15fff, 0xe2ad54ff, 0xdda743ff, 0xe8bf73ff, 0xf0d398ff, 0xf0d29aff, 
    0xedcf94ff, 0xe3c386ff, 0xd5b16cff, 0xae9157ff, 0x5e5644ff, 0x70716fff, 
    0x8c8e89ff, 0x959793ff, 0xcacbc8ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xb9bab8ff, 0x52504bff, 0x5b5034ff, 
    0x907c4eff, 0xae9559ff, 0xc6aa69ff, 0xd4b36aff, 0xd7b160ff, 0xdab665ff, 
    0xddb96dff, 0xdab360ff, 0xddb86aff, 0xd9b15dff, 0xcea043ff, 0xcfa245ff, 
    0xd6ac59ff, 0xd9b160ff, 0xdcb665ff, 0xddb665ff, 0xddb565ff, 0xdfba6bff, 
    0xe4c37dff, 0xddb767ff, 0xd8ac51ff, 0xd8a94bff, 0xd7a74aff, 0xd9a950ff, 
    0xd4a448ff, 0xd19f3aff, 0xd09e37ff, 0xcc972aff, 0xc58c1bff, 0xc48b1dff, 
    0xc48b1fff, 0xd8aa4aff, 0xe5bf6bff, 0xe3c074ff, 0xe3bf70ff, 0xe8c986ff, 
    0xeace90ff, 0xeacc92ff, 0xe9ca90ff, 0xe5c786ff, 0xe4c585ff, 0xe4c281ff, 
    0xe1c07fff, 0xe0c07eff, 0xdcb871ff, 0xdcb771ff, 0xdbb871ff, 0xdab468ff, 
    0xdbb56cff, 0xdbb66fff, 0xdab46eff, 0xdbb97bff, 0xdcbc7dff, 0xe1c286ff, 
    0xe2c28bff, 0xe0c081ff, 0xe2c386ff, 0xe3c48bff, 0xe4c58dff, 0xe5c792ff, 
    0xe1c186ff, 0xddbd7fff, 0xe2c692ff, 0xe2cb9bff, 0xe3cc9fff, 0xe3cfa5ff, 
    0xdfc89aff, 0xddc292ff, 0xdec59aff, 0xdcc497ff, 0xd3b57bff, 0xc9a668ff, 
    0xccac74ff, 0xccac6fff, 0xf0daa9ff, 0xf1dba4ff, 0xefd395ff, 0xf0d497ff, 
    0xf0d293ff, 0xeecf8cff, 0xefd292ff, 0xe5c37eff, 0xd3a457ff, 0xd3a253ff, 
    0xdbb062ff, 0xe0b66bff, 0xe6be77ff, 0xebc680ff, 0xeac074ff, 0xe5b663ff, 
    0xd89f4aff, 0xdfa855ff, 0xe9ba6aff, 0xe5b45fff, 0xdfa843ff, 0xdda747ff, 
    0xe5bc72ff, 0xe8c689ff, 0xe4c38dff, 0xd4b277ff, 0xaa8b51ff, 0x5d533eff, 
    0x747573ff, 0x8c8e89ff, 0x989a95ff, 0xcdcecbff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xbebfbdff, 0x54534fff, 
    0x54492fff, 0x8d7a4dff, 0xad945aff, 0xc5a765ff, 0xd5b672ff, 0xdbb86dff, 
    0xdbb669ff, 0xdab567ff, 0xdcb565ff, 0xd4aa4eff, 0xd4a951ff, 0xd0a34aff, 
    0xcd9f46ff, 0xd2a751ff, 0xd9b05fff, 0xdbb466ff, 0xdab466ff, 0xdcb666ff, 
    0xdbb25fff, 0xdeb767ff, 0xe0b96cff, 0xddb361ff, 0xdcb15fff, 0xdaad56ff, 
    0xdbad5aff, 0xdcae59ff, 0xd9a94cff, 0xd8a547ff, 0xd2a13bff, 0xce9a2dff, 
    0xc99223ff, 0xcc9833ff, 0xe1b864ff, 0xe5be6aff, 0xe3bb64ff, 0xe1b75aff, 
    0xe4c16eff, 0xeacd8bff, 0xeacc8dff, 0xecce96ff, 0xe9cc91ff, 0xe5c585ff, 
    0xe8ca8bff, 0xe4c383ff, 0xe1bf78ff, 0xe1c17dff, 0xe0bd77ff, 0xdebb77ff, 
    0xe1bd7aff, 0xe4c284ff, 0xe0bf7dff, 0xddb870ff, 0xddba72ff, 0xe1c080ff, 
    0xe2c284ff, 0xe4c587ff, 0xe3c484ff, 0xe1c17dff, 0xe4c588ff, 0xe0bf7dff, 
    0xe0c07fff, 0xe1c181ff, 0xe0c082ff, 0xe1c58bff, 0xe5cc9dff, 0xe5d0a4ff, 
    0xe2cb9eff, 0xe2cca4ff, 0xdfc797ff, 0xdec390ff, 0xdcc390ff, 0xdcc291ff, 
    0xdabe8aff, 0xccad70ff, 0xc9a96cff, 0xcfb27aff, 0xf3e0b3ff, 0xf3deadff, 
    0xf1daa9ff, 0xeccf96ff, 0xdfb672ff, 0xd09f5aff, 0xcf9f62ff, 0xb57330ff, 
    0x9f4f0fff, 0xbe833eff, 0xddb272ff, 0xe8c28aff, 0xe9c689ff, 0xe7c283ff, 
    0xe2b870ff, 0xd7a85eff, 0xaf6411ff, 0xbc7622ff, 0xcb8d3cff, 0xd29438ff, 
    0xd0922dff, 0xce8f24ff, 0xcd9332ff, 0xce9b45ff, 0xc9a05cff, 0x9f8047ff, 
    0x5d5442ff, 0x787977ff, 0x8b8d89ff, 0x9a9c98ff, 0xcfd1ceff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xc3c4c2ff, 
    0x5a5956ff, 0x534932ff, 0x8d7b4fff, 0xb09b68ff, 0xc5a968ff, 0xd1af62ff, 
    0xd9b565ff, 0xd9b365ff, 0xdcb86cff, 0xd7ae5aff, 0xd8b05bff, 0xd6ac56ff, 
    0xd5ab56ff, 0xd7ae60ff, 0xd4aa5bff, 0xd6ad5dff, 0xd7ae5cff, 0xd7ad5aff, 
    0xdbb260ff, 0xdbb15dff, 0xd8ad55ff, 0xdbb05cff, 0xdbad5aff, 0xdbb05cff, 
    0xddb05eff, 0xdfb260ff, 0xdfb462ff, 0xddb05cff, 0xdcac56ff, 0xd8a64aff, 
    0xd7a644ff, 0xd19f35ff, 0xcc9528ff, 0xd6a748ff, 0xe9c375ff, 0xe8bf6aff, 
    0xe3bd66ff, 0xe5bf68ff, 0xeac97dff, 0xeed395ff, 0xebcf91ff, 0xeccf91ff, 
    0xecd39fff, 0xe7c88cff, 0xe8c98bff, 0xe7c98dff, 0xe9c98bff, 0xe4c381ff, 
    0xe5c684ff, 0xe4c382ff, 0xe4c281ff, 0xe6c686ff, 0xe6c583ff, 0xe4c27dff, 
    0xe4c27cff, 0xe7c789ff, 0xe7c78aff, 0xe8c78aff, 0xe9cb8fff, 0xe7c885ff, 
    0xe6c786ff, 0xe3c481ff, 0xe0c17dff, 0xdfbf78ff, 0xe0c07dff, 0xe4ca91ff, 
    0xe6cd9eff, 0xe3cc9dff, 0xe1c691ff, 0xdec58fff, 0xdec38bff, 0xdabd80ff, 
    0xd7ba77ff, 0xdcc089ff, 0xdbbe8bff, 0xcdac6fff, 0xc9a969ff, 0xccad71ff, 
    0xe0bd7fff, 0xeacc96ff, 0xe3c084ff, 0xc8933bff, 0xb16a0eff, 0xaa5c05ff, 
    0xa95b08ff, 0xa45303ff, 0x9a4003ff, 0xaa5f1cff, 0xd6a66aff, 0xebca9aff, 
    0xe6c185ff, 0xe0b670ff, 0xd2a560ff, 0xb36f25ff, 0xa65704ff, 0xa95b04ff, 
    0xae6003ff, 0xb16404ff, 0xb16604ff, 0xb36a05ff, 0xb06907ff, 0xa96607ff, 
    0x81510dff, 0x514837ff, 0x7a7c79ff, 0x8b8d88ff, 0x9d9f9bff, 0xd2d3d1ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xc7c8c6ff, 0x605f5cff, 0x514731ff, 0x8a784dff, 0xaf9964ff, 0xc8af74ff, 
    0xd4b56eff, 0xd7b363ff, 0xdab462ff, 0xdab463ff, 0xd9b361ff, 0xd7ac58ff, 
    0xd6ab59ff, 0xd6ad5dff, 0xd5ab53ff, 0xdab468ff, 0xe0be7bff, 0xddb86fff, 
    0xd8af5fff, 0xd8af5eff, 0xdbb361ff, 0xd8ac56ff, 0xd5a648ff, 0xd7a84eff, 
    0xd6a84fff, 0xd5a649ff, 0xd8aa52ff, 0xd9aa51ff, 0xd6a74cff, 0xd6a74fff, 
    0xd7a94fff, 0xdcac57ff, 0xdbac50ff, 0xd4a139ff, 0xcf9727ff, 0xdcac48ff, 
    0xe8be65ff, 0xe8c16bff, 0xe8c477ff, 0xeacb84ff, 0xeacb81ff, 0xeacc85ff, 
    0xebce8dff, 0xeacc8cff, 0xeed29fff, 0xedd19cff, 0xedd096ff, 0xedd097ff, 
    0xebca8fff, 0xe6c581ff, 0xe7c885ff, 0xe6c681ff, 0xe7c880ff, 0xeacc86ff, 
    0xeacc8cff, 0xe9ca88ff, 0xe7c780ff, 0xe9cc8aff, 0xe9cc8cff, 0xe9cb8cff, 
    0xeace92ff, 0xebce90ff, 0xe9cc8dff, 0xe5c482ff, 0xe1bf77ff, 0xdfbc71ff, 
    0xe0c17cff, 0xe2c382ff, 0xe6cd9bff, 0xe5cf9dff, 0xe0c488ff, 0xd9bd7bff, 
    0xe1c996ff, 0xdec590ff, 0xdcc489ff, 0xdcc38dff, 0xdabe8cff, 0xcaa762ff, 
    0xc4a156ff, 0xc49e50ff, 0xcb973cff, 0xce9b47ff, 0xce9c46ff, 0xca9338ff, 
    0xb67110ff, 0xae6104ff, 0xad6204ff, 0xad6405ff, 0xa45303ff, 0xb56d21ff, 
    0xe4be88ff, 0xe5c18fff, 0xe0b87cff, 0xd3a159ff, 0xb36f26ff, 0xa35203ff, 
    0xa85904ff, 0xac5d04ff, 0xb06404ff, 0xb36a05ff, 0xb06905ff, 0xad6704ff, 
    0xa56203ff, 0x7c4d05ff, 0x4f4533ff, 0x7d7f7dff, 0x8b8d88ff, 0xa0a29eff, 
    0xd4d6d3ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xcbcccaff, 0x656663ff, 0x534b39ff, 0x8c7953ff, 0xae9866ff, 
    0xc7ae73ff, 0xd5b772ff, 0xdab96cff, 0xdab765ff, 0xd9b360ff, 0xdab566ff, 
    0xdbb564ff, 0xd7ad56ff, 0xd3a850ff, 0xd4a951ff, 0xd4aa54ff, 0xd9b165ff, 
    0xdbb366ff, 0xd6aa52ff, 0xd3a64bff, 0xd6ac52ff, 0xd8ad56ff, 0xd7ac55ff, 
    0xd3a445ff, 0xd4a64aff, 0xd2a346ff, 0xd2a242ff, 0xd3a343ff, 0xd19f3dff, 
    0xd1a03eff, 0xd2a240ff, 0xd5a446ff, 0xd7a648ff, 0xd6a544ff, 0xd39f35ff, 
    0xd5a136ff, 0xe3b85eff, 0xebc778ff, 0xebc87cff, 0xe7c274ff, 0xe9c982ff, 
    0xeccd87ff, 0xf0d79bff, 0xefd49aff, 0xeccd8dff, 0xeaca8bff, 0xe9c989ff, 
    0xeed097ff, 0xefd39fff, 0xeccd8fff, 0xe8c77eff, 0xeac986ff, 0xe7c579ff, 
    0xe9c87cff, 0xedd08dff, 0xedd08cff, 0xebce88ff, 0xe8c779ff, 0xe9c87bff, 
    0xebcc83ff, 0xedd090ff, 0xedd297ff, 0xeed195ff, 0xeacc8dff, 0xe5c682ff, 
    0xe2bf76ff, 0xe2c27bff, 0xe2c17fff, 0xdfbe79ff, 0xe3c68eff, 0xe8d3a5ff, 
    0xe3cc98ff, 0xe3cb98ff, 0xe2ca9aff, 0xe1cc9fff, 0xdfc898ff, 0xddc38eff, 
    0xd8bc86ff, 0xc09644ff, 0xb17f1fff, 0xb18023ff, 0xcc993fff, 0xcc983dff, 
    0xca973bff, 0xcb963dff, 0xc08524ff, 0xaf6a04ff, 0xac6304ff, 0xab6205ff, 
    0xa95c04ff, 0xb06615ff, 0xc28741ff, 0xc99352ff, 0xb37132ff, 0xa3530eff, 
    0x9e4a05ff, 0xa35104ff, 0xa85a05ff, 0xac5e04ff, 0xb06805ff, 0xae6605ff, 
    0xa96505ff, 0xa15e04ff, 0x794806ff, 0x534939ff, 0x818380ff, 0x8b8d88ff, 
    0xa3a5a1ff, 0xd7d8d5ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xcecfceff, 0x686967ff, 0x4d4533ff, 0x88774dff, 
    0xac965fff, 0xc8b274ff, 0xd8be7fff, 0xddbd72ff, 0xe0be6fff, 0xdfbe72ff, 
    0xdbb767ff, 0xdbb565ff, 0xddb869ff, 0xdbb462ff, 0xd6ac53ff, 0xd4a84cff, 
    0xd4a950ff, 0xd4aa53ff, 0xd3a74bff, 0xd6ac56ff, 0xd6ab57ff, 0xd5aa55ff, 
    0xd6aa56ff, 0xd7ab58ff, 0xd2a449ff, 0xcf9d39ff, 0xd2a13eff, 0xd7a64bff, 
    0xd1a03aff, 0xcd992cff, 0xcc992bff, 0xce9b32ff, 0xce9a34ff, 0xd19d35ff, 
    0xd4a03cff, 0xd29f37ff, 0xd7a53eff, 0xe3b95dff, 0xe6bd63ff, 0xe9c77aff, 
    0xe4bd66ff, 0xe4bc68ff, 0xecce89ff, 0xf0d498ff, 0xf0d59bff, 0xf0d59bff, 
    0xefd397ff, 0xeed195ff, 0xeacb8bff, 0xe9c987ff, 0xedce8eff, 0xeccd87ff, 
    0xebcc83ff, 0xeccd84ff, 0xeacc82ff, 0xeccf88ff, 0xedcf87ff, 0xebcc81ff, 
    0xecca7aff, 0xebc56dff, 0xebc776ff, 0xeccd83ff, 0xedcf89ff, 0xeacb83ff, 
    0xe8c77bff, 0xe6c57bff, 0xe5c377ff, 0xe7c984ff, 0xe7cb8dff, 0xe6c88bff, 
    0xe2c484ff, 0xe3c78eff, 0xe3c994ff, 0xe4cc9bff, 0xe4cc9bff, 0xe1ca9aff, 
    0xe1c797ff, 0xe0c591ff, 0xdbc28eff, 0xb98b36ff, 0xa77004ff, 0xa87207ff, 
    0xcb993cff, 0xd0a251ff, 0xcd9c49ff, 0xc9963dff, 0xc48e2dff, 0xb2730bff, 
    0xa95f05ff, 0xa85a05ff, 0xab6005ff, 0xad6304ff, 0xac6103ff, 0xa95e04ff, 
    0x9b4903ff, 0x9d4803ff, 0xa04d04ff, 0xa65505ff, 0xa85a05ff, 0xab6004ff, 
    0xad6504ff, 0xa66004ff, 0x9d5b03ff, 0x704307ff, 0x564c3eff, 0x838582ff, 
    0x8b8d88ff, 0xa7a8a4ff, 0xd9dad7ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xd2d3d1ff, 0x6f6f6cff, 0x463d2aff, 
    0x84724aff, 0xab935bff, 0xc8b072ff, 0xd9be79ff, 0xe2c47dff, 0xe6c984ff, 
    0xe4c279ff, 0xe1be74ff, 0xddb768ff, 0xdcb665ff, 0xdcb766ff, 0xd9b058ff, 
    0xd5ab4dff, 0xd6ac52ff, 0xd5aa51ff, 0xd4a94fff, 0xd5aa53ff, 0xd6aa57ff, 
    0xd7ac57ff, 0xd5a951ff, 0xd3a64cff, 0xd1a345ff, 0xd09f3fff, 0xcc9831ff, 
    0xca972aff, 0xce9c34ff, 0xcb972cff, 0xc89221ff, 0xca9523ff, 0xca9527ff, 
    0xc99528ff, 0xcb9628ff, 0xcd9829ff, 0xd29e31ff, 0xdaa945ff, 0xe6bc65ff, 
    0xe6bf69ff, 0xe9c579ff, 0xeac87bff, 0xebcb7fff, 0xefd291ff, 0xf2d99fff, 
    0xf0d49cff, 0xf0d59bff, 0xf1daa2ff, 0xefd497ff, 0xeccd88ff, 0xeaca82ff, 
    0xeed28eff, 0xeccd85ff, 0xedd088ff, 0xeed288ff, 0xf1d691ff, 0xf1d794ff, 
    0xf1d593ff, 0xeecb7eff, 0xedc978ff, 0xedc771ff, 0xebc469ff, 0xecc872ff, 
    0xedc875ff, 0xebc879ff, 0xeac878ff, 0xedcf84ff, 0xeed493ff, 0xecd091ff, 
    0xe8cb86ff, 0xe7c985ff, 0xe6c88aff, 0xe4c787ff, 0xe5ca8fff, 0xe5cd98ff, 
    0xe4cc97ff, 0xe3cb9aff, 0xe4d0a2ff, 0xe2cb98ff, 0xdbbe86ff, 0xbc9040ff, 
    0xae7b1aff, 0xb17f1eff, 0xcb973dff, 0xce9f4fff, 0xcc9c4aff, 0xcb963cff, 
    0xc89337ff, 0xb87b1aff, 0xa75b04ff, 0xa65605ff, 0xaf6806ff, 0xb57409ff, 
    0xb5730dff, 0xa76109ff, 0x933e04ff, 0x9a4204ff, 0x9f4a04ff, 0xa65404ff, 
    0xa75704ff, 0xa95e05ff, 0xa55f05ff, 0x9a5903ff, 0x694008ff, 0x524d43ff, 
    0x858784ff, 0x8b8d88ff, 0xaaaca8ff, 0xdadbd9ff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xd5d6d4ff, 0x757573ff, 
    0x423720ff, 0x7c6637ff, 0xa8925eff, 0xc5aa69ff, 0xd8bc74ff, 0xe2c172ff, 
    0xe7ca81ff, 0xe4c275ff, 0xdeb766ff, 0xe2c077ff, 0xdfbb6fff, 0xdbb361ff, 
    0xdbb25eff, 0xd9af59ff, 0xd7ae56ff, 0xd5aa50ff, 0xd4aa4fff, 0xd7ac55ff, 
    0xd7ab54ff, 0xd5a850ff, 0xd3a546ff, 0xcf9e3cff, 0xcb9a34ff, 0xcb982eff, 
    0xcd992dff, 0xce9c38ff, 0xcf9d39ff, 0xd4a440ff, 0xd3a342ff, 0xd0a13dff, 
    0xcb982dff, 0xcb9729ff, 0xd09d34ff, 0xcf9a2eff, 0xc9911eff, 0xcb9420ff, 
    0xd29c2cff, 0xe4b95dff, 0xe9c677ff, 0xe8c475ff, 0xecca7fff, 0xe9c87cff, 
    0xeccf88ff, 0xf0d696ff, 0xf1d99dff, 0xf0d79bff, 0xf0d599ff, 0xefd493ff, 
    0xefd492ff, 0xf2d897ff, 0xf2d998ff, 0xf1d797ff, 0xf0d58fff, 0xf2d792ff, 
    0xf2d790ff, 0xf2d88fff, 0xf2d78eff, 0xf0d286ff, 0xeec873ff, 0xeec770ff, 
    0xeec569ff, 0xefca71ff, 0xefcb72ff, 0xefcd79ff, 0xefd286ff, 0xf2d892ff, 
    0xf2da9fff, 0xeed495ff, 0xebd08aff, 0xecd193ff, 0xeace91ff, 0xead19aff, 
    0xead4a5ff, 0xe9d4a5ff, 0xead5a4ff, 0xe6cf9dff, 0xe3cd9fff, 0xe3cca1ff, 
    0xdbbe84ff, 0xc2994cff, 0xb98d38ff, 0xbc9240ff, 0xc89332ff, 0xc79330ff, 
    0xcb9b44ff, 0xca9a45ff, 0xca9a44ff, 0xc0892aff, 0xac6206ff, 0xaa5e04ff, 
    0xb27005ff, 0xba7d11ff, 0xba7d15ff, 0xaf6a07ff, 0x9e4a04ff, 0x983d04ff, 
    0x9a4004ff, 0x9f4804ff, 0xa25105ff, 0xa35a05ff, 0x985803ff, 0x653f0aff, 
    0x555149ff, 0x878986ff, 0x8b8d88ff, 0xaeafabff, 0xdcdddbff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xd7d8d6ff, 
    0x7c7d7aff, 0x42361fff, 0x745925ff, 0xa18345ff, 0xc4a86aff, 0xd9bc79ff, 
    0xe3c67fff, 0xe7c97fff, 0xe5c77bff, 0xe2c16fff, 0xddb660ff, 0xe1bf76ff, 
    0xdfb96cff, 0xd9b056ff, 0xd7ab51ff, 0xd7ac51ff, 0xd6aa50ff, 0xd2a648ff, 
    0xd4a648ff, 0xd6aa51ff, 0xd3a449ff, 0xd1a241ff, 0xd6a646ff, 0xcf9c37ff, 
    0xc9922aff, 0xc89024ff, 0xc88e24ff, 0xd7a03bff, 0xe5b856ff, 0xedcb7dff, 
    0xf0d290ff, 0xf1d499ff, 0xeecf8eff, 0xeccc84ff, 0xebc87eff, 0xe9c578ff, 
    0xe2b965ff, 0xdcae50ff, 0xd49f33ff, 0xe4bb63ff, 0xf0d190ff, 0xf0d394ff, 
    0xeed18bff, 0xeaca7dff, 0xedd087ff, 0xf0d58eff, 0xf3dc9dff, 0xf0d694ff, 
    0xebcc86ff, 0xf0d391ff, 0xf2d899ff, 0xf2d895ff, 0xf2d795ff, 0xf3d899ff, 
    0xf2d68fff, 0xf3d890ff, 0xf5dc97ff, 0xf2d88eff, 0xf0d281ff, 0xf0cf7eff, 
    0xefcb75ff, 0xeec86bff, 0xefc86cff, 0xf1cc73ff, 0xf2d17dff, 0xf2d483ff, 
    0xf2d78cff, 0xf4db97ff, 0xf5dda1ff, 0xf1d898ff, 0xeed491ff, 0xeed395ff, 
    0xecd090ff, 0xecd39cff, 0xecd6a7ff, 0xead5a8ff, 0xebd7abff, 0xe9d6a8ff, 
    0xe5cd9bff, 0xe0c590ff, 0xdabc80ff, 0xc9a359ff, 0xc29b4bff, 0xc49f50ff, 
    0xc89733ff, 0xc49028ff, 0xc89639ff, 0xca9840ff, 0xcea150ff, 0xca9a46ff, 
    0xb57312ff, 0xae6304ff, 0xb37107ff, 0xb97a12ff, 0xbf8625ff, 0xb67813ff, 
    0xaf6b07ff, 0x9c4a05ff, 0x963b04ff, 0x9a4304ff, 0x9b4b04ff, 0x935103ff, 
    0x643e0cff, 0x5a564eff, 0x898b87ff, 0x8c8e89ff, 0xb1b3afff, 0xdddedcff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 
    0xd9dad8ff, 0x828381ff, 0x423723ff, 0x6d501bff, 0x96722cff, 0xbb974cff, 
    0xd7b974ff, 0xe3c781ff, 0xe8cc87ff, 0xe9ce87ff, 0xe8c87dff, 0xe4c375ff, 
    0xe1be6cff, 0xe1bc6dff, 0xdeb661ff, 0xd9ae55ff, 0xd7ab4fff, 0xd3a542ff, 
    0xd3a446ff, 0xd0a142ff, 0xd3a341ff, 0xdbb159ff, 0xe2bc6aff, 0xe9c77aff, 
    0xf1d291ff, 0xe8c075ff, 0xd9a13dff, 0xd8a03fff, 0xd8a043ff, 0xe3b551ff, 
    0xebc66dff, 0xeccd83ff, 0xeed08cff, 0xefd291ff, 0xf1d497ff, 0xf2d79eff, 
    0xf2d79fff, 0xf1d79eff, 0xf0d197ff, 0xeac782ff, 0xe3b763ff, 0xe8c171ff, 
    0xf0d492ff, 0xefd28eff, 0xeecf89ff, 0xf1d592ff, 0xf2d995ff, 0xf7e1a6ff, 
    0xf6e1abff, 0xf5dea3ff, 0xf3dca0ff, 0xf6dfa8ff, 0xf6dda1ff, 0xf4db9aff, 
    0xf3da94ff, 0xf6e0a4ff, 0xf5de9cff, 0xf7e09dff, 0xf6df99ff, 0xf5dc95ff, 
    0xf4d788ff, 0xf4d88cff, 0xf1d07dff, 0xf0cb6cff, 0xf3d27aff, 0xf5d887ff, 
    0xf6dd96ff, 0xf6df99ff, 0xf6e09eff, 0xf5df9bff, 0xf5dd99ff, 0xf5de9fff, 
    0xf3dea3ff, 0xf3dda7ff, 0xf1d89fff, 0xeed6a3ff, 0xebd3a4ff, 0xead4a6ff, 
    0xe9d5a6ff, 0xe8d5a9ff, 0xe7d1a0ff, 0xe0c184ff, 0xd5b26aff, 0xcca85eff, 
    0xcba85cff, 0xcca95eff, 0xcb9a3dff, 0xca993fff, 0xcc9d47ff, 0xc99741ff, 
    0xcd9f4bff, 0xcfa355ff, 0xba8021ff, 0xae6504ff, 0xb26f07ff, 0xb77812ff, 
    0xc18b32ff, 0xbf882cff, 0xb87b18ff, 0xa35908ff, 0x964104ff, 0x934004ff, 
    0x8e4303ff, 0x61360bff, 0x5b5750ff, 0x8a8c88ff, 0x8c8e89ff, 0xb5b6b3ff, 
    0xdedfddff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe2e3e1ff, 0xdbdcdaff, 0x888988ff, 0x3d3422ff, 0x684c1aff, 0x8e681fff, 
    0xaf8533ff, 0xcfa858ff, 0xe3c580ff, 0xe8ca83ff, 0xeace89ff, 0xeacc85ff, 
    0xe7c879ff, 0xe5c273ff, 0xe2be6dff, 0xdbaf54ff, 0xdaae51ff, 0xd8ab4eff, 
    0xd9a94aff, 0xcf9d32ff, 0xd09d36ff, 0xd8a94bff, 0xe8c475ff, 0xeed08bff, 
    0xf2d899ff, 0xf4daa0ff, 0xf2d69bff, 0xf0d398ff, 0xeccd88ff, 0xe9c67aff, 
    0xeac678ff, 0xe9c777ff, 0xeaca80ff, 0xebc981ff, 0xeaca80ff, 0xecce8aff, 
    0xeed195ff, 0xf0d39bff, 0xefd39cff, 0xefd39bff, 0xeece93ff, 0xe8c47fff, 
    0xe5bb6cff, 0xe6bd69ff, 0xeecd82ff, 0xeecf87ff, 0xf0d58fff, 0xf0d58eff, 
    0xf1d68fff, 0xf7e2a8ff, 0xfae6b0ff, 0xf8e4aaff, 0xf7e2a7ff, 0xf7e1a2ff, 
    0xf7e2a0ff, 0xfae4a6ff, 0xf9e3a3ff, 0xfbe9b6ff, 0xf9e4a8ff, 0xfae5a9ff, 
    0xf9e5a5ff, 0xf8e298ff, 0xf8e094ff, 0xf9e29bff, 0xf6da8aff, 0xf4d279ff, 
    0xf6d986ff, 0xfae399ff, 0xfae59fff, 0xfae6aaff, 0xfbe7b5ff, 0xf9e5a9ff, 
    0xf8e4aaff, 0xf6e1a7ff, 0xf4dfa6ff, 0xf4e1aeff, 0xf2deabff, 0xeed79fff, 
    0xebd5a2ff, 0xead3a4ff, 0xe9d4a6ff, 0xe9d5a8ff, 0xe9d3a7ff, 0xe3c68eff, 
    0xd3ad60ff, 0xcda959ff, 0xcfab5bff, 0xd1ae66ff, 0xcb9b45ff, 0xcda251ff, 
    0xcea250ff, 0xca9b47ff, 0xcfa454ff, 0xd2a85dff, 0xc79540ff, 0xb26e0aff, 
    0xb16c04ff, 0xb3730aff, 0xc0872cff, 0xc38d33ff, 0xb87c19ff, 0xa55f0bff, 
    0x8c3604ff, 0x893804ff, 0x6b3205ff, 0x524a40ff, 0x888b87ff, 0x8d8f8aff, 
    0xb8bab6ff, 0xdfe0deff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xdddedcff, 0x8f908fff, 0x3e3526ff, 0x5f420dff, 
    0x865e10ff, 0xa3751fff, 0xc29440ff, 0xd8af5cff, 0xe9cb88ff, 0xedd18eff, 
    0xeacc81ff, 0xe7c478ff, 0xe9ca84ff, 0xe7c67eff, 0xe3bd69ff, 0xd9ae4eff, 
    0xd4a640ff, 0xd6a743ff, 0xd3a137ff, 0xcf9a2bff, 0xd19d31ff, 0xe8c678ff, 
    0xf3daa0ff, 0xf3d99eff, 0xf2daa2ff, 0xf1d7a0ff, 0xefd397ff, 0xeed094ff, 
    0xecce8dff, 0xeacc88ff, 0xeacb85ff, 0xeacd88ff, 0xebcc88ff, 0xeccb89ff, 
    0xeacb88ff, 0xeacb89ff, 0xeacb8aff, 0xeccd8fff, 0xedcd93ff, 0xedcd94ff, 
    0xeac98cff, 0xe1b76cff, 0xe1b35eff, 0xe4b960ff, 0xf1d490ff, 0xf3d997ff, 
    0xf5de9aff, 0xf1d58bff, 0xefce7fff, 0xf5de9aff, 0xfce9b3ff, 0xfbe7aeff, 
    0xf7e19fff, 0xf8e3a5ff, 0xfae5a7ff, 0xfae7aaff, 0xfce9aeff, 0xfce8afff, 
    0xfce8b1ff, 0xfce9aeff, 0xfce9aaff, 0xfbe6a2ff, 0xfae49aff, 0xfae39aff, 
    0xfae294ff, 0xf9df91ff, 0xf9e192ff, 0xfbe49dff, 0xfbe7a9ff, 0xfce9adff, 
    0xfbe9b0ff, 0xfae5a7ff, 0xf8e4a6ff, 0xf6e3abff, 0xf6e1a9ff, 0xf6e3aeff, 
    0xf4e1afff, 0xf1dca4ff, 0xedd69eff, 0xead49fff, 0xe9d3a5ff, 0xe9d4a8ff, 
    0xe8d2a4ff, 0xe6ce9eff, 0xd8b670ff, 0xd0aa57ff, 0xd2af63ff, 0xd5b268ff, 
    0xd0a357ff, 0xd1a659ff, 0xd2a85cff, 0xd1a65bff, 0xcfa456ff, 0xd1a659ff, 
    0xc99b49ff, 0xb3720fff, 0xb16c04ff, 0xb3740bff, 0xbc821fff, 0xc7943dff, 
    0xc18b30ff, 0xac6c14ff, 0x863304ff, 0x762b02ff, 0x492f20ff, 0x7c7f7cff, 
    0x8d8f8aff, 0xb7b8b5ff, 0xe0e1dfff, 0xe2e3e1ff, 0xe4e5e4ff, 0xe4e5e3ff, 
    0xe3e4e3ff, 0xe4e5e4ff, 0xe3e4e2ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xdedfddff, 0x969795ff, 0x3f3829ff, 
    0x604415ff, 0x855d14ff, 0x9a6b0fff, 0xb17d1eff, 0xcd9e47ff, 0xdfba6fff, 
    0xe8c57eff, 0xf0d494ff, 0xe7c574ff, 0xe3bb66ff, 0xe5c173ff, 0xe5c06fff, 
    0xe4bc69ff, 0xddb354ff, 0xd6a941ff, 0xd6a63fff, 0xd19e30ff, 0xce992aff, 
    0xe2b65cff, 0xf0d390ff, 0xf0d499ff, 0xf0d396ff, 0xf0d59eff, 0xf0d39bff, 
    0xeed294ff, 0xedd094ff, 0xebcc8bff, 0xeacb89ff, 0xe8c883ff, 0xe7c57dff, 
    0xe7c47cff, 0xe8c67fff, 0xe7c781ff, 0xe7c884ff, 0xe7c781ff, 0xe7c682ff, 
    0xe7c583ff, 0xe7c27fff, 0xe4bf77ff, 0xe1b869ff, 0xdfb55eff, 0xe0b356ff, 
    0xf0ce84ff, 0xf6e09fff, 0xf7e19eff, 0xf5dd97ff, 0xf3d88dff, 0xf8e3a0ff, 
    0xfbe7acff, 0xfbe7adff, 0xf8e3a1ff, 0xf8e3a1ff, 0xfbe6a7ff, 0xfceaaeff, 
    0xfdeab4ff, 0xfceab4ff, 0xfdeab9ff, 0xfdeab3ff, 0xfce9aaff, 0xfce9a8ff, 
    0xfce8a5ff, 0xfce6a0ff, 0xfbe69cff, 0xfce8a5ff, 0xfceaaeff, 0xfce9b0ff, 
    0xfceab5ff, 0xfcebb8ff, 0xfbe9b3ff, 0xfbe9b1ff, 0xf9e6a9ff, 0xf7e5abff, 
    0xf6e3abff, 0xf6e2acff, 0xf6e6b7ff, 0xf4e2b3ff, 0xf0d9a4ff, 0xebd094ff, 
    0xe8ce93ff, 0xe7cc97ff, 0xe6cc94ff, 0xe5cb95ff, 0xdcbc7bff, 0xd4ae5dff, 
    0xd4b162ff, 0xd9b76dff, 0xd3aa64ff, 0xd2a961ff, 0xd3a962ff, 0xd4aa64ff, 
    0xd3aa63ff, 0xd1a85aff, 0xcc9f50ff, 0xb57413ff, 0xb26d05ff, 0xbd8321ff, 
    0xc5923aff, 0xc99842ff, 0xc49038ff, 0xad721fff, 0x7c2d05ff, 0x5a1f05ff, 
    0x5a5551ff, 0x8c8e8aff, 0xa5a7a3ff, 0xdddedcff, 0xe3e4e2ff, 0xdadbdaff, 
    0xaeaeadff, 0x8e8c88ff, 0x8a8680ff, 0x9e9c98ff, 0xc9c9c9ff, 0xe3e4e3ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe0e0dfff, 0x9c9e9cff, 
    0x423a2aff, 0x5d4112ff, 0x855d18ff, 0x9e6f1cff, 0xad7715ff, 0xb9811aff, 
    0xd1a04aff, 0xdcb163ff, 0xe3ba6bff, 0xe8c477ff, 0xe3be69ff, 0xdfb656ff, 
    0xdeb24fff, 0xe0b755ff, 0xe1b85cff, 0xdaae4bff, 0xd5a53bff, 0xd29e2eff, 
    0xcf9a26ff, 0xd7a53dff, 0xe9c375ff, 0xedcc8bff, 0xeccb8bff, 0xebcb89ff, 
    0xeccd8fff, 0xeccd90ff, 0xeccc8fff, 0xebcc8fff, 0xe9ca87ff, 0xe8c886ff, 
    0xe7c57bff, 0xe4bf6eff, 0xe3bd6cff, 0xe3bb69ff, 0xe4be70ff, 0xe5c47fff, 
    0xe5c480ff, 0xe4c178ff, 0xe3bd6fff, 0xe2bb6eff, 0xe2ba6bff, 0xe1bb6cff, 
    0xe1ba6dff, 0xe2b866ff, 0xf1d187ff, 0xf8e3a3ff, 0xf7e09dff, 0xf6df9bff, 
    0xf7e19fff, 0xf9e4a4ff, 0xf7df97ff, 0xfae6abff, 0xfae5a8ff, 0xfae5a6ff, 
    0xfceab1ff, 0xfdebb4ff, 0xfceaadff, 0xfceab2ff, 0xfcecbbff, 0xfcebb4ff, 
    0xfce9aaff, 0xfdeaadff, 0xfdeaafff, 0xfdeab0ff, 0xfce9abff, 0xfbe8aaff, 
    0xfbe7b0ff, 0xfcecbeff, 0xfcedc1ff, 0xfbecbcff, 0xfbeabaff, 0xfbe9b6ff, 
    0xfae9b4ff, 0xf9e8b2ff, 0xf7e3adff, 0xf5e0a7ff, 0xf4e1aaff, 0xf2e1b0ff, 
    0xf1dda6ff, 0xefd89cff, 0xe8cc84ff, 0xe5ca88ff, 0xe2c581ff, 0xe1c285ff, 
    0xd9b670ff, 0xd7b365ff, 0xd9b76dff, 0xdcb970ff, 0xd3aa62ff, 0xd1a65dff, 
    0xd3a75eff, 0xd4aa60ff, 0xd6ae69ff, 0xd4ad6aff, 0xd0a65dff, 0xb7781cff, 
    0xb46f08ff, 0xc1872dff, 0xcd9c50ff, 0xcc9e4eff, 0xc79745ff, 0xad772bff, 
    0x732804ff, 0x472112ff, 0x757774ff, 0x90928dff, 0xcaccc9ff, 0xe3e4e2ff, 
    0xd3d4d2ff, 0x706f6bff, 0x423822ff, 0x4e3c1aff, 0x493108ff, 0x442d08ff, 
    0x4c3e2eff, 0xa1a09fff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe0e1dfff, 
    0xa2a4a2ff, 0x413a2dff, 0x563806ff, 0x7d520aff, 0x95630fff, 0xab7319ff, 
    0xb57b17ff, 0xb87b12ff, 0xc58b29ff, 0xd4a243ff, 0xddb158ff, 0xe6c075ff, 
    0xe2bb66ff, 0xdcaf4bff, 0xdbb04cff, 0xddb251ff, 0xdcaf4fff, 0xd4a63bff, 
    0xcf9c27ff, 0xcc9318ff, 0xce961fff, 0xdfb24fff, 0xe6bb5eff, 0xe8bf6cff, 
    0xe8c172ff, 0xe7c274ff, 0xe8c47dff, 0xe7c47cff, 0xe6c175ff, 0xe7c378ff, 
    0xe7c682ff, 0xe6c47fff, 0xe3be6eff, 0xe1ba68ff, 0xe0b561ff, 0xdfb45aff, 
    0xdfb45bff, 0xe3bd72ff, 0xe3c07bff, 0xe2be75ff, 0xe2bc71ff, 0xe0b96dff, 
    0xdfb668ff, 0xdeb665ff, 0xdfba6fff, 0xe5c077ff, 0xf4d792ff, 0xfae5aaff, 
    0xfbe8b2ff, 0xfbe8adff, 0xfae4a1ff, 0xf5db93ff, 0xf6db90ff, 0xfae5a1ff, 
    0xfce9aaff, 0xfde9acff, 0xfceaaaff, 0xfcebabff, 0xfdeaabff, 0xfceab1ff, 
    0xfcecb5ff, 0xfdedb3ff, 0xfdebb2ff, 0xfdeab2ff, 0xfcebb7ff, 0xfdeab7ff, 
    0xfdeab4ff, 0xfbe9b0ff, 0xf9e5a9ff, 0xf9e6b0ff, 0xf6e1a8ff, 0xf4dd9dff, 
    0xf5dfa0ff, 0xf6e3aaff, 0xf8e8b4ff, 0xf8e9b8ff, 0xf7e7b4ff, 0xf6e4b3ff, 
    0xf6e6b6ff, 0xf2e0acff, 0xf1dca2ff, 0xf0db9eff, 0xeed798ff, 0xecd59dff, 
    0xe5c787ff, 0xdcb97aff, 0xd9b56dff, 0xdcb971ff, 0xe0c079ff, 0xdfbc71ff, 
    0xcfa45bff, 0xcfa257ff, 0xcc9d4cff, 0xcd9f4dff, 0xd4ab65ff, 0xd6b174ff, 
    0xd1a860ff, 0xba7d20ff, 0xb26a05ff, 0xb77715ff, 0xbd7f26ff, 0xc48d36ff, 
    0xc89a4aff, 0xad7c39ff, 0x692504ff, 0x432c23ff, 0x828582ff, 0x9ea09bff, 
    0xdddedcff, 0xdddedcff, 0x8e8f8dff, 0x3f3724ff, 0x6b572aff, 0x714d15ff, 
    0x714703ff, 0x6a4405ff, 0x513402ff, 0x423827ff, 0xbbbcbbff, 0xe3e4e2ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xa9aaa9ff, 0x413a31ff, 0x4c2b03ff, 0x734403ff, 0x8e5705ff, 
    0xa16507ff, 0xae700dff, 0xb4740dff, 0xb37208ff, 0xb8780dff, 0xd09c37ff, 
    0xddb059ff, 0xe2bb6bff, 0xe0b761ff, 0xdbae4cff, 0xdbae49ff, 0xd8ab43ff, 
    0xd3a337ff, 0xd19e2dff, 0xcb9216ff, 0xcb9016ff, 0xdcad48ff, 0xe9c269ff, 
    0xe7bd63ff, 0xe3b95cff, 0xe2b85cff, 0xe3bb63ff, 0xe5bf70ff, 0xe5bc69ff, 
    0xe4bb64ff, 0xe5bc67ff, 0xe5bd70ff, 0xe4bc6dff, 0xe0b55dff, 0xdcae4fff, 
    0xdaaa4bff, 0xd8a945ff, 0xdaab4aff, 0xdfb35eff, 0xe2ba70ff, 0xe1ba70ff, 
    0xe0b96cff, 0xdfb669ff, 0xddb262ff, 0xdbb361ff, 0xddb567ff, 0xe0b76dff, 
    0xeecc83ff, 0xfae3a5ff, 0xfbe5a8ff, 0xfceaaeff, 0xfbe8a2ff, 0xf8df94ff, 
    0xf8e197ff, 0xfceaadff, 0xfdecb3ff, 0xfceaaeff, 0xfceaa8ff, 0xfcecb1ff, 
    0xfdefbaff, 0xfdefb8ff, 0xfdedb6ff, 0xfdecb7ff, 0xfdebb6ff, 0xfce9b1ff, 
    0xf9e4a7ff, 0xf9e5a9ff, 0xf9e5b0ff, 0xefd396ff, 0xe3c071ff, 0xe1bc67ff, 
    0xdab257ff, 0xd7ad52ff, 0xd7ae50ff, 0xd7ad52ff, 0xe0bb69ff, 0xe9ce8cff, 
    0xefda9eff, 0xf1db9fff, 0xf3e1a8ff, 0xf4e2b0ff, 0xf2dfaaff, 0xf0da9cff, 
    0xeed898ff, 0xefdca8ff, 0xecd7a2ff, 0xeacf93ff, 0xe2c17bff, 0xdfbe75ff, 
    0xe3c380ff, 0xe2c27bff, 0xcea157ff, 0xcfa154ff, 0xcb9c4aff, 0xc58f33ff, 
    0xd2a458ff, 0xd9b47bff, 0xd2a964ff, 0xbb8023ff, 0xb26f08ff, 0xb87816ff, 
    0xbc7d1fff, 0xc18628ff, 0xc69749ff, 0xa06823ff, 0x611d01ff, 0x43322cff, 
    0x868985ff, 0xacada9ff, 0xe2e3e1ff, 0xc4c6c3ff, 0x626360ff, 0x594b2cff, 
    0x886b38ff, 0x85500aff, 0x925c06ff, 0x8b5c11ff, 0x755013ff, 0x42300fff, 
    0x868785ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xafb0aeff, 0x443e37ff, 0x462603ff, 0x6c3a03ff, 
    0x854903ff, 0x9b5b04ff, 0xa86404ff, 0xad6704ff, 0xb06c05ff, 0xaf6b05ff, 
    0xb06b05ff, 0xc88d2aff, 0xdbac56ff, 0xe0b764ff, 0xdfb65fff, 0xdaac48ff, 
    0xdaaa43ff, 0xd6a439ff, 0xd29f2cff, 0xcf9922ff, 0xcd941bff, 0xe0b351ff, 
    0xeac165ff, 0xeeca74ff, 0xf1d081ff, 0xebc976ff, 0xe4bc60ff, 0xe1b75aff, 
    0xe2b860ff, 0xe1b45aff, 0xdfb458ff, 0xe0b45cff, 0xdfb35cff, 0xddb056ff, 
    0xd8a743ff, 0xd29f31ff, 0xd1a032ff, 0xd5a53eff, 0xd7a844ff, 0xd9ab4cff, 
    0xdcb057ff, 0xdbb058ff, 0xddb361ff, 0xdbb05bff, 0xd9ae56ff, 0xdaaf5aff, 
    0xdaae58ff, 0xdaae59ff, 0xe3b964ff, 0xf6db91ff, 0xfbe7a7ff, 0xfdeaafff, 
    0xfce9a8ff, 0xfbe49fff, 0xfae296ff, 0xfceaabff, 0xfdeeb8ff, 0xfcecb0ff, 
    0xfdecb2ff, 0xfdeeb9ff, 0xfceeb3ff, 0xfcebabff, 0xfdeab7ff, 0xfceabaff, 
    0xfce8b0ff, 0xfbe6a8ff, 0xf6dc97ff, 0xebcd86ff, 0xe0bc73ff, 0xd7ad55ff, 
    0xd3a743ff, 0xd2a53bff, 0xd3a63fff, 0xcfa33cff, 0xcca034ff, 0xc99827ff, 
    0xc89420ff, 0xcc9d30ff, 0xd3a947ff, 0xe0bc66ff, 0xebd38bff, 0xf0dba1ff, 
    0xefdba1ff, 0xf0db9fff, 0xefd99aff, 0xefdba3ff, 0xeddba6ff, 0xedd79eff, 
    0xe8cb88ff, 0xdfbd74ff, 0xe1c076ff, 0xe3c276ff, 0xd0a457ff, 0xd0a355ff, 
    0xc99740ff, 0xc58d30ff, 0xd3a55dff, 0xd6ae71ff, 0xcfa35dff, 0xba7d1fff, 
    0xb3720fff, 0xb67613ff, 0xb97b1aff, 0xbe8526ff, 0xbd8633ff, 0x9b6013ff, 
    0x642401ff, 0x412f28ff, 0x858784ff, 0xb0b2aeff, 0xe1e2e0ff, 0xb0b1aeff, 
    0x61615eff, 0x645430ff, 0x885d26ff, 0x935203ff, 0xa66908ff, 0x9d6914ff, 
    0x845913ff, 0x4b3511ff, 0x797a77ff, 0xdbdcd9ff, 0xe2e3e1ff, 0xe1e2e0ff, 
    0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xb4b6b4ff, 0x49433dff, 0x422204ff, 
    0x663502ff, 0x804404ff, 0x945004ff, 0xa05804ff, 0xa75d04ff, 0xa96104ff, 
    0xab6204ff, 0xab6004ff, 0xab6103ff, 0xb77614ff, 0xd4a145ff, 0xddb15dff, 
    0xe2bc6fff, 0xe0b561ff, 0xd9ac4dff, 0xd7a646ff, 0xd39f33ff, 0xcc9319ff, 
    0xd9a435ff, 0xecc268ff, 0xedc365ff, 0xf0cd74ff, 0xf3d78eff, 0xf6da97ff, 
    0xf0cf7fff, 0xe9c168ff, 0xe1b557ff, 0xddad4fff, 0xd9a742ff, 0xd5a13aff, 
    0xd39f38ff, 0xd09927ff, 0xce9922ff, 0xd09d2aff, 0xd1a036ff, 0xd3a339ff, 
    0xd4a53dff, 0xd9ac53ff, 0xd5a643ff, 0xd6a544ff, 0xd7a74fff, 0xd3a44bff, 
    0xd1a245ff, 0xd1a349ff, 0xd0a247ff, 0xd2a54dff, 0xd6aa52ff, 0xeecd7fff, 
    0xf9e09aff, 0xfce7a7ff, 0xfce9a5ff, 0xfdeba8ff, 0xfbe499ff, 0xfce8a1ff, 
    0xfdecabff, 0xfdedb3ff, 0xfdecafff, 0xfceaa8ff, 0xfbe79fff, 0xf8df97ff, 
    0xf9e3a9ff, 0xf8e2a2ff, 0xf6de9aff, 0xf2d78fff, 0xe6c37aff, 0xdab467ff, 
    0xd8af5aff, 0xd4aa47ff, 0xd5ac49ff, 0xd6ab46ff, 0xd3a740ff, 0xcfa53eff, 
    0xcea237ff, 0xcc9c2aff, 0xca9922ff, 0xcb9c28ff, 0xcb9c2eff, 0xcd9e36ff, 
    0xd3ab4bff, 0xe4c575ff, 0xead087ff, 0xecd594ff, 0xefdca4ff, 0xefdba6ff, 
    0xeedaa2ff, 0xefdba4ff, 0xedd59eff, 0xe0bc71ff, 0xe1bd6bff, 0xe4c172ff, 
    0xd2a354ff, 0xd2a65bff, 0xca9843ff, 0xc28825ff, 0xcb9542ff, 0xd0a45bff, 
    0xca994cff, 0xb57516ff, 0xb06d07ff, 0xb67713ff, 0xbc812aff, 0xbf872eff, 
    0xb67c1aff, 0xa0650aff, 0x753b02ff, 0x3e271bff, 0x7c7f7cff, 0xabada9ff, 
    0xe0e1dfff, 0xabaca8ff, 0x767775ff, 0x50442aff, 0x723f0dff, 0x8c4903ff, 
    0xa56707ff, 0x9d6712ff, 0x7e5514ff, 0x483c27ff, 0x8b8c8aff, 0xd7d8d6ff, 
    0xe2e3e1ff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe2e3e1ff, 0xbabbb9ff, 0x4c4742ff, 
    0x3f1f05ff, 0x623002ff, 0x7b3f03ff, 0x904b04ff, 0x9a5005ff, 0xa15404ff, 
    0xa35404ff, 0xa25305ff, 0xa35504ff, 0xa45604ff, 0xa55904ff, 0xa85d06ff, 
    0xc1862bff, 0xdbaf5aff, 0xe2bb70ff, 0xe2ba6eff, 0xdcb05eff, 0xd5a143ff, 
    0xd6a138ff, 0xdba73dff, 0xe8bc5aff, 0xebc25fff, 0xeec86eff, 0xedc769ff, 
    0xefcd7dff, 0xf0d087ff, 0xf0cd78ff, 0xf1cc78ff, 0xedc772ff, 0xe8c068ff, 
    0xd7a945ff, 0xca9424ff, 0xce992cff, 0xce982bff, 0xcf9a27ff, 0xd4a238ff, 
    0xd5a440ff, 0xd3a23aff, 0xd5a340ff, 0xd6a649ff, 0xd3a445ff, 0xd2a347ff, 
    0xd1a24cff, 0xd1a44fff, 0xd1a755ff, 0xcfa250ff, 0xcb9c48ff, 0xcea04dff, 
    0xcfa251ff, 0xdfb664ff, 0xf2d381ff, 0xfae4a1ff, 0xfbe7a3ff, 0xfce9a2ff, 
    0xfbe79fff, 0xfbe69bff, 0xfceaa5ff, 0xfdebafff, 0xfdeab1ff, 0xfce9afff, 
    0xf9e4a6ff, 0xf5d98eff, 0xf4d78aff, 0xf2d58aff, 0xe8c67cff, 0xdbb361ff, 
    0xd5ab52ff, 0xd8ae5aff, 0xd7ae56ff, 0xd1a742ff, 0xd4aa46ff, 0xd9b054ff, 
    0xd4aa49ff, 0xd0a43fff, 0xcfa33cff, 0xcfa336ff, 0xcea132ff, 0xcd9f2fff, 
    0xc99923ff, 0xc89722ff, 0xca9c2eff, 0xd2a946ff, 0xe3c371ff, 0xe9ce85ff, 
    0xeed99fff, 0xefdca8ff, 0xf0dba4ff, 0xf1dea9ff, 0xecd39cff, 0xe2bf75ff, 
    0xe3be6fff, 0xe3bf6eff, 0xd19d45ff, 0xd1a04eff, 0xca9538ff, 0xbf821cff, 
    0xc28325ff, 0xce9c4fff, 0xca984cff, 0xb37110ff, 0xb16f07ff, 0xb97b1fff, 
    0xbd8329ff, 0xc28b32ff, 0xbe8726ff, 0xa87013ff, 0x854e06ff, 0x49280aff, 
    0x666664ff, 0x9fa19dff, 0xdcdddbff, 0xb7b8b5ff, 0x8b8d89ff, 0x5c5b55ff, 
    0x47270eff, 0x6e3402ff, 0x8b5404ff, 0x7e520bff, 0x503c1dff, 0x646561ff, 
    0xa5a7a3ff, 0xdddedcff, 0xe1e2e0ff, 0xe1e2e0ff, 0xe3e4e2ff, 0xbfc0beff, 
    0x4f4c49ff, 0x381905ff, 0x5c2702ff, 0x773803ff, 0x8b4304ff, 0x984a04ff, 
    0x9e4d05ff, 0x9f5004ff, 0xa05004ff, 0xa15105ff, 0xa15105ff, 0xa25504ff, 
    0xa15405ff, 0xa25404ff, 0xa75d0aff, 0xd3a14fff, 0xe1b76dff, 0xddae5dff, 
    0xd6a549ff, 0xd6a23eff, 0xe9bf66ff, 0xeec569ff, 0xeac05aff, 0xe7bc51ff, 
    0xedc975ff, 0xecc877ff, 0xecc56eff, 0xebc670ff, 0xebc469ff, 0xebc364ff, 
    0xeecb77ff, 0xf2d287ff, 0xeecd81ff, 0xd6a545ff, 0xd5a143ff, 0xd6a54aff, 
    0xd4a345ff, 0xd19f3dff, 0xd4a13eff, 0xd2a13aff, 0xd1a13fff, 0xd09f42ff, 
    0xd0a048ff, 0xd0a24fff, 0xcfa252ff, 0xcda050ff, 0xcc9d4dff, 0xc99846ff, 
    0xc89644ff, 0xc89744ff, 0xcb9d4eff, 0xd3a658ff, 0xe9c471ff, 0xf5d98fff, 
    0xfbe4a0ff, 0xfce49cff, 0xfbe399ff, 0xfce79dff, 0xfce9a4ff, 0xfbe6a6ff, 
    0xfae5acff, 0xfbe4acff, 0xf4d898ff, 0xeaca7dff, 0xe4bf6eff, 0xdfb665ff, 
    0xd7af5dff, 0xd4aa51ff, 0xd6ad53ff, 0xd5ab4fff, 0xd3a94aff, 0xd2a741ff, 
    0xd4aa44ff, 0xd5ac4eff, 0xd5ab4bff, 0xd3a944ff, 0xd3a844ff, 0xd1a43cff, 
    0xcda138ff, 0xcd9f36ff, 0xcd9f31ff, 0xca9d2eff, 0xcc9e32ff, 0xcc9f34ff, 
    0xdab55bff, 0xe8cd84ff, 0xeacf8dff, 0xedd89eff, 0xeedba2ff, 0xf0dca5ff, 
    0xe8c988ff, 0xe1b969ff, 0xe3bc6aff, 0xe4bf6eff, 0xce983bff, 0xcf9d45ff, 
    0xcd9a43ff, 0xc18425ff, 0xc3872bff, 0xcb994eff, 0xc48e3dff, 0xb16d0bff, 
    0xb3710aff, 0xb87a1bff, 0xbe8429ff, 0xc48e38ff, 0xc59342ff, 0xaf7a27ff, 
    0x8a4e05ff, 0x582902ff, 0x443b36ff, 0x8d8f8cff, 0xc8cac7ff, 0xd4d5d3ff, 
    0x969793ff, 0x898b87ff, 0x686864ff, 0x4e4439ff, 0x4e402cff, 0x534c3eff, 
    0x6e6f6cff, 0x949692ff, 0xc8c9c6ff, 0xe2e3e1ff, 0xe1e2e0ff, 0xe3e4e2ff, 
    0xc3c4c3ff, 0x54514fff, 0x351807ff, 0x582502ff, 0x6e2e03ff, 0x813804ff, 
    0x8f4104ff, 0x954604ff, 0x9f5104ff, 0xa15204ff, 0xa05104ff, 0x9e5004ff, 
    0x9e5005ff, 0xa05304ff, 0xa25604ff, 0xa35505ff, 0xa15304ff, 0xc1842cff, 
    0xd8a954ff, 0xd6a245ff, 0xd39d35ff, 0xe3b456ff, 0xeec774ff, 0xebc262ff, 
    0xeabf59ff, 0xecc464ff, 0xefd089ff, 0xedcd83ff, 0xe9c470ff, 0xe8c166ff, 
    0xebc874ff, 0xecc774ff, 0xeecd84ff, 0xf2d38fff, 0xf3d795ff, 0xe8bf67ff, 
    0xdead4fff, 0xdaa74aff, 0xd9aa4fff, 0xd6a447ff, 0xd19f3fff, 0xcf9f40ff, 
    0xd0a146ff, 0xcfa04bff, 0xcd9e4cff, 0xca9a48ff, 0xc99846ff, 0xc89743ff, 
    0xc28d35ff, 0xc5903aff, 0xc7933eff, 0xc7933cff, 0xca9b46ff, 0xcfa250ff, 
    0xd7ab56ff, 0xe9c26aff, 0xf3d688ff, 0xf7dc8eff, 0xf9e091ff, 0xfae297ff, 
    0xf7df97ff, 0xf6dc99ff, 0xf1d590ff, 0xeed292ff, 0xd8ae6dff, 0xb07826ff, 
    0xb07a1eff, 0xcea145ff, 0xd6ae5aff, 0xd5ab53ff, 0xd7ae56ff, 0xd4ab4dff, 
    0xd4aa4aff, 0xd5ab49ff, 0xd5aa4aff, 0xd4aa4fff, 0xd4aa4eff, 0xd4a948ff, 
    0xd3a847ff, 0xd0a540ff, 0xcea23aff, 0xcda139ff, 0xcda137ff, 0xc89929ff, 
    0xc99b2fff, 0xc99b31ff, 0xcfa443ff, 0xe3c67bff, 0xe5c87eff, 0xe9d193ff, 
    0xeddaa3ff, 0xf1dda8ff, 0xe6c580ff, 0xdbb25dff, 0xdeb65eff, 0xe4be6bff, 
    0xc78d25ff, 0xc88d2fff, 0xcd9743ff, 0xc1852aff, 0xbd7e1dff, 0xbf8127ff, 
    0xbb7b1cff, 0xaf6806ff, 0xb36d07ff, 0xb97a18ff, 0xbf8325ff, 0xc79142ff, 
    0xc59343ff, 0xad7119ff, 0x864404ff, 0x5b1d02ff, 0x37170cff, 0x636360ff, 
    0xa7a8a5ff, 0xdbdcdaff, 0xc7c9c6ff, 0x999b97ff, 0x8d8f8aff, 0x898b87ff, 
    0x868885ff, 0x8c8e8aff, 0x9ea09bff, 0xc7c8c5ff, 0xe1e2e0ff, 0xe1e2e0ff, 
    0xe3e4e2ff, 0xc7c8c7ff, 0x5d5c58ff, 0x39210cff, 0x552101ff, 0x6b2d03ff, 
    0x7c3503ff, 0x8b3e04ff, 0x944504ff, 0x964304ff, 0x9c4a04ff, 0x9f5004ff, 
    0x9f5204ff, 0x9d5004ff, 0x9d5104ff, 0xa05404ff, 0xa15504ff, 0xa05304ff, 
    0xa15404ff, 0xac650dff, 0xcf9739ff, 0xdca84aff, 0xdba338ff, 0xe6b85eff, 
    0xe8bd63ff, 0xeac26bff, 0xedc46aff, 0xf0d083ff, 0xf3d89cff, 0xf3d99dff, 
    0xf0d595ff, 0xeccd82ff, 0xeecf89ff, 0xebca81ff, 0xeccb84ff, 0xeece8dff, 
    0xefcf8bff, 0xecc774ff, 0xe9bf66ff, 0xeabe5fff, 0xeec979ff, 0xeabe63ff, 
    0xd8aa4fff, 0xcd9d44ff, 0xce9d47ff, 0xce9c49ff, 0xca9743ff, 0xc79139ff, 
    0xc58f35ff, 0xc58d34ff, 0xc28b32ff, 0xc28c36ff, 0xc18a30ff, 0xc58f37ff, 
    0xc99742ff, 0xce9f4eff, 0xcb9c4aff, 0xcc9a40ff, 0xd9af5fff, 0xe1bb70ff, 
    0xe3bc6cff, 0xddb566ff, 0xd3a34fff, 0xd3a454ff, 0xc7913eff, 0xbc822bff, 
    0xb06b18ff, 0x9d5d05ff, 0xa46d0bff, 0xc99b3cff, 0xd5ac56ff, 0xd5ac56ff, 
    0xd7b05dff, 0xd8af59ff, 0xd4ab4eff, 0xd5ad52ff, 0xd7ae57ff, 0xd4ab4fff, 
    0xd4ab51ff, 0xd4aa4cff, 0xd4a94aff, 0xd2a848ff, 0xd0a542ff, 0xd0a543ff, 
    0xcea33cff, 0xca9d2fff, 0xc99d32ff, 0xc79a2fff, 0xc99d35ff, 0xd4ac52ff, 
    0xe2c57bff, 0xe9d197ff, 0xebd39eff, 0xeed79fff, 0xe5c580ff, 0xd8af56ff, 
    0xdeb760ff, 0xe4be6fff, 0xc98d29ff, 0xc88b2aff, 0xc78c2eff, 0xbb7b19ff, 
    0xb6710aff, 0xb46e09ff, 0xaf6603ff, 0xad6303ff, 0xb06804ff, 0xb87513ff, 
    0xc08328ff, 0xc89345ff, 0xc48d40ff, 0xb1711bff, 0x7f3502ff, 0x5c1702ff, 
    0x481101ff, 0x36211bff, 0x737471ff, 0xb3b5b2ff, 0xdedfddff, 0xd8d9d7ff, 
    0xbfc0bdff, 0xafb1adff, 0xb0b1aeff, 0xc0c1beff, 0xd8d9d7ff, 0xe2e3e1ff, 
    0xe1e2e0ff, 0xe2e3e1ff, 0xc8c9c7ff, 0x63635fff, 0x4a3b23ff, 0x73501bff, 
    0x713709ff, 0x793203ff, 0x873b04ff, 0x8d3e04ff, 0x944504ff, 0x984a04ff, 
    0x994a04ff, 0x9b4e04ff, 0xa15405ff, 0xa05304ff, 0x9b5004ff, 0x9f5404ff, 
    0x9f5404ff, 0xa25504ff, 0xa05504ff, 0xa65c05ff, 0xce9130ff, 0xd99f3aff, 
    0xd79b28ff, 0xe3b359ff, 0xe5b963ff, 0xe9c26fff, 0xecc97aff, 0xf0d28eff, 
    0xf3d99eff, 0xf5dea7ff, 0xf3dba2ff, 0xf2d799ff, 0xf0d598ff, 0xeccc8aff, 
    0xeccd89ff, 0xedce8dff, 0xebca7eff, 0xeac573ff, 0xecc777ff, 0xedc770ff, 
    0xefcd81ff, 0xefcc7bff, 0xebc776ff, 0xd5a54dff, 0xc9943dff, 0xc8933bff, 
    0xc69136ff, 0xc58d2fff, 0xc68e33ff, 0xbf8529ff, 0xbd8326ff, 0xbe852aff, 
    0xbe852bff, 0xc18c35ff, 0xc69341ff, 0xc5913bff, 0xc18a2bff, 0xbb8227ff, 
    0xbc8740ff, 0xac681eff, 0xa7610bff, 0xa76309ff, 0xac6808ff, 0xad6a0bff, 
    0xaf6d0dff, 0xac6b0fff, 0x9d5803ff, 0x985c04ff, 0xa46c0aff, 0xc59437ff, 
    0xd2aa57ff, 0xd4ac59ff, 0xd4ad59ff, 0xd4ac55ff, 0xd4ac54ff, 0xd3ab51ff, 
    0xd5ae54ff, 0xd5ad52ff, 0xd5ac50ff, 0xd4ab4bff, 0xd2a846ff, 0xd1a745ff, 
    0xcfa541ff, 0xd1a644ff, 0xd0a642ff, 0xcea33dff, 0xcda23bff, 0xcca13cff, 
    0xcda23fff, 0xcb9f3aff, 0xd1a94cff, 0xdebd72ff, 0xe2c27fff, 0xe6c785ff, 
    0xdeb86aff, 0xdab45fff, 0xdeb863ff, 0xdfb761ff, 0xc98d2aff, 0xc48622ff, 
    0xbe7b18ff, 0xb26908ff, 0xad5e05ff, 0xb77119ff, 0xc18431ff, 0xc89343ff, 
    0xc79445ff, 0xcfa157ff, 0xd4a75eff, 0xd9af66ff, 0xd1a560ff, 0xbd8639ff, 
    0x9c5f26ff, 0x702c0bff, 0x571c05ff, 0x3f0f01ff, 0x35231dff, 0x6e6f6cff, 
    0xb1b2b0ff, 0xdbdcdaff, 0xe3e4e2ff, 0xe2e3e1ff, 0xe2e3e1ff, 0xe3e4e2ff, 
    0xe2e3e1ff, 0xe2e3e1ff, 0xdedfddff, 0xbabbb9ff, 0x61605bff, 0x49391eff, 
    0x7c5e32ff, 0x997331ff, 0xa16e25ff, 0x8a4109ff, 0x8b3c02ff, 0x8f4104ff, 
    0x924404ff, 0x984b04ff, 0x984b04ff, 0x984c04ff, 0x9c5204ff, 0x9a5004ff, 
    0x9d5204ff, 0x9d5404ff, 0x9f5604ff, 0xa15704ff, 0xa15604ff, 0xa45905ff, 
    0xcd8f2dff, 0xd79c30ff, 0xd59a28ff, 0xd9a138ff, 0xdfae4fff, 0xe7bc63ff, 
    0xebca7dff, 0xf4daa1ff, 0xf5dca5ff, 0xf4dda8ff, 0xf3dba2ff, 0xf6e0abff, 
    0xf1d697ff, 0xf0d395ff, 0xedce8fff, 0xebcc89ff, 0xebca84ff, 0xe9c77eff, 
    0xeac87cff, 0xebc97fff, 0xecca7cff, 0xecca7bff, 0xf1d490ff, 0xebc880ff, 
    0xd9ac54ff, 0xc99235ff, 0xbf8523ff, 0xc18726ff, 0xc0862aff, 0xb77b19ff, 
    0xb67a17ff, 0xb87d1cff, 0xb97e1eff, 0xbd8225ff, 0xc0882dff, 0xbd8424ff, 
    0xb87f1dff, 0x9b580fff, 0x92480bff, 0x934503ff, 0x964c03ff, 0x954d04ff, 
    0x985304ff, 0x9b5504ff, 0x985504ff, 0x955805ff, 0x955b04ff, 0x955e05ff, 
    0x9f6606ff, 0xc18f30ff, 0xd4ac5cff, 0xd5ae5dff, 0xd3ac58ff, 0xd4ad59ff, 
    0xd4ae59ff, 0xd4ad53ff, 0xd5ae56ff, 0xd6ae59ff, 0xd5ae58ff, 0xd3ab4cff, 
    0xd3a947ff, 0xd3aa49ff, 0xd3aa49ff, 0xd1a845ff, 0xd1a847ff, 0xcfa441ff, 
    0xd0a443ff, 0xd1a84bff, 0xcfa749ff, 0xcea443ff, 0xcfa548ff, 0xd2ab50ff, 
    0xdbb766ff, 0xdcb668ff, 0xdab563ff, 0xdcb96bff, 0xdeb966ff, 0xdeb861ff, 
    0xc2821cff, 0xbd7712ff, 0xb46806ff, 0xb36913ff, 0xc08134ff, 0xe3bb7eff, 
    0xe9c995ff, 0xebd0a0ff, 0xe9cd9dff, 0xecd1a6ff, 0xeacfa1ff, 0xe5c692ff, 
    0xe2c18cff, 0xdfbc84ff, 0xdab87eff, 0xc19a62ff, 0xa17942ff, 0x602c0bff, 
    0x451a06ff, 0x453825ff, 0x565651ff, 0x8f918fff, 0xbebfbdff, 0xd4d5d3ff, 
    0xdbdcdaff, 0xdbdcdaff, 0xd5d6d3ff, 0xc0c1bfff, 0x8d8f8dff, 0x4a463fff, 
    0x47310fff, 0x745016ff, 0x986f32ff, 0xb48946ff, 0xc69949ff, 0xbf8c42ff, 
    0x9d5815ff, 0x934502ff, 0x954804ff, 0x974a04ff, 0x974b04ff, 0x974d04ff, 
    0x994f04ff, 0x9a4f04ff, 0x9d5404ff, 0x9f5604ff, 0x9e5405ff, 0xa35805ff, 
    0xa35805ff, 0xa15504ff, 0xc58722ff, 0xd19625ff, 0xd79f32ff, 0xdda846ff, 
    0xe4b65fff, 0xebc77eff, 0xeccb86ff, 0xefd190ff, 0xf3daa5ff, 0xf3ddafff, 
    0xf2dba7ff, 0xf5e0b0ff, 0xf2d99dff, 0xf3dba5ff, 0xf1d79fff, 0xeed091ff, 
    0xeaca88ff, 0xe8c77fff, 0xebcb84ff, 0xebcc87ff, 0xebc880ff, 0xe9c577ff, 
    0xefd192ff, 0xf0d299ff, 0xeeca7eff, 0xe5ba63ff, 0xc68e2cff, 0xb87a16ff, 
    0xb67917ff, 0xb2750fff, 0xb3740fff, 0xb47512ff, 0xb2720fff, 0xad6b0aff, 
    0xac690cff, 0xaf6e0eff, 0xa46209ff, 0x874503ff, 0x914d04ff, 0x9e5404ff, 
    0x985004ff, 0x914b04ff, 0x904a04ff, 0x8d4d04ff, 0x8c5204ff, 0x8d5604ff, 
    0x905904ff, 0x925c04ff, 0x986005ff, 0xba8626ff, 0xd3aa55ff, 0xd4b05fff, 
    0xd4b062ff, 0xd5b064ff, 0xd5af60ff, 0xd5b05dff, 0xd6b160ff, 0xd6b15eff, 
    0xd7af5dff, 0xd5ad53ff, 0xd5ad52ff, 0xd5ad57ff, 0xd6b05dff, 0xd5ae57ff, 
    0xd1a84bff, 0xd0a645ff, 0xd0a74aff, 0xcfa64bff, 0xcfa74eff, 0xd0a750ff, 
    0xcfa84eff, 0xd1aa52ff, 0xd4ac54ff, 0xd4ad53ff, 0xd8b360ff, 0xdbb86dff, 
    0xddbc74ff, 0xdfbb6eff, 0xb86f09ff, 0xb86e0bff, 0xcc903aff, 0xe4bd79ff, 
    0xebcc96ff, 0xedd1a2ff, 0xedd3a8ff, 0xeed6b0ff, 0xeed3a7ff, 0xedd3a6ff, 
    0xebcf9fff, 0xeacb99ff, 0xecd4abff, 0xe8cd9dff, 0xe3c38eff, 0xdec08dff, 
    0xccab73ff, 0x9b7238ff, 0x6c4118ff, 0x735425ff, 0x564220ff, 0x3f3523ff, 
    0x4c4941ff, 0x656563ff, 0x767775ff, 0x767775ff, 0x656562ff, 0x49443eff, 
    0x392a16ff, 0x4c2b03ff, 0x6f4204ff, 0x8e5d14ff, 0xa87525ff, 0xc08d3eff, 
    0xce9c4bff, 0xd8ab5fff, 0xcb9a48ff, 0xaa6a1cff, 0x954a03ff, 0x994d03ff, 
    0x974d04ff, 0x974d04ff, 0x9b5104ff, 0x9d5304ff, 0x9f5304ff, 0x9e5304ff, 
    0x9c5104ff, 0xa15505ff, 0xa05503ff, 0xa65e0bff, 0xc68c2dff, 0xd39e37ff, 
    0xdaa53bff, 0xe1b154ff, 0xe8c174ff, 0xecca8eff, 0xefce98ff, 0xedcd90ff, 
    0xefd49bff, 0xf2daaaff, 0xf3dca9ff, 0xf4deaeff, 0xf0d496ff, 0xf4ddabff, 
    0xf4dbaaff, 0xf0d39aff, 0xe9c886ff, 0xe7c47aff, 0xe8c577ff, 0xeac985ff, 
    0xeccc8bff, 0xe8c57cff, 0xeac67eff, 0xefd198ff, 0xecca85ff, 0xedc671ff, 
    0xe5bb64ff, 0xc99237ff, 0xb2720eff, 0xad6a06ff, 0xad6a08ff, 0xaf6e09ff, 
    0xad6804ff, 0xab6504ff, 0xa45c04ff, 0xa15a04ff, 0x8d4b04ff, 0x955305ff, 
    0xab6705ff, 0xaa6405ff, 0xa25904ff, 0x9c5204ff, 0x9a5004ff, 0x964e04ff, 
    0x8e4e04ff, 0x8b5304ff, 0x8d5604ff, 0x905904ff, 0x925b03ff, 0xaa7110ff, 
    0xcb9e47ff, 0xd6b56eff, 0xd6b36cff, 0xd6b36dff, 0xd5b168ff, 0xd5b167ff, 
    0xd5af67ff, 0xd4ad5fff, 0xd7b260ff, 0xd7b260ff, 0xd6b360ff, 0xd6b262ff, 
    0xd6b05fff, 0xd5af5cff, 0xd3ad55ff, 0xd1a94eff, 0xd0a749ff, 0xd1a954ff, 
    0xcfa853ff, 0xcea64bff, 0xcfa74bff, 0xd1ab56ff, 0xd3ad56ff, 0xd5af58ff, 
    0xd9b667ff, 0xd9b870ff, 0xd8b66fff, 0xdcb972ff, 0xce9339ff, 0xdaa859ff, 
    0xe9c382ff, 0xefd09aff, 0xeccb99ff, 0xeac796ff, 0xe9c897ff, 0xeccca0ff, 
    0xebcd9fff, 0xeacf9fff, 0xebd0a1ff, 0xedd3abff, 0xecd8b5ff, 0xead3aaff, 
    0xe7ca98ff, 0xe4c48fff, 0xddbf8aff, 0xcdac74ff, 0xb49053ff, 0x926a30ff, 
    0x80581aff, 0x6b480dff, 0x523508ff, 0x361905ff, 0x2f1309ff, 0x321709ff, 
    0x391905ff, 0x4a2401ff, 0x5f3202ff, 0x733d03ff, 0x874b03ff, 0xa0630fff, 
    0xb47922ff, 0xbc8223ff, 0xc58a2aff, 0xd3a251ff, 0xd6a759ff, 0xd4a557ff, 
    0xaf7129ff, 0x9d540aff, 0x994e03ff, 0x9a4f04ff, 0x9d5304ff, 0x9e5404ff, 
    0x9f5304ff, 0x9d5104ff, 0x9d5104ff, 0x9e5304ff, 0xa76210ff, 0xcf9c50ff, 
    0xe0b66cff, 0xdbad54ff, 0xdcac4cff, 0xe4b963ff, 0xe9c379ff, 0xe9c389ff, 
    0xedcc9aff, 0xf1d39fff, 0xedce94ff, 0xeed196ff, 0xf1d6a0ff, 0xf2d8a2ff, 
    0xf0d396ff, 0xf2d9a7ff, 0xf2dbaaff, 0xf2d9a5ff, 0xeccc8dff, 0xe7c47aff, 
    0xe9c67aff, 0xeac884ff, 0xebc987ff, 0xe9c47bff, 0xe7be69ff, 0xe8c06cff, 
    0xe8be66ff, 0xe7bc57ff, 0xedc875ff, 0xe8c47cff, 0xb97d1fff, 0xab6403ff, 
    0xaa6205ff, 0xab6305ff, 0xaa6104ff, 0xab6304ff, 0xa45c04ff, 0x9c5204ff, 
    0x965304ff, 0xb0710cff, 0xb57811ff, 0xac6805ff, 0xa75e05ff, 0xa15805ff, 
    0xa05704ff, 0x9d5404ff, 0x984e04ff, 0x8f5104ff, 0x8c5404ff, 0x8f5704ff, 
    0x905903ff, 0x9d6207ff, 0xc28e36ff, 0xd3ae67ff, 0xd6b573ff, 0xd7b571ff, 
    0xd5b26aff, 0xd5b36eff, 0xd4af6aff, 0xd4ae64ff, 0xd7b367ff, 0xd7b264ff, 
    0xd5af5eff, 0xd5b05dff, 0xd4af5eff, 0xd5ae5dff, 0xd3ae5dff, 0xd1a954ff, 
    0xcea54aff, 0xcfa954ff, 0xcea753ff, 0xcea74eff, 0xd1aa55ff, 0xd0aa57ff, 
    0xd1aa55ff, 0xd3ad59ff, 0xd8b56cff, 0xd8b671ff, 0xd7b46dff, 0xd9b56cff, 
    0xeecc91ff, 0xefd096ff, 0xeece95ff, 0xeccc95ff, 0xedcc99ff, 0xf0d1a5ff, 
    0xeecf9fff, 0xeed2a6ff, 0xeed3a7ff, 0xedd0a2ff, 0xebcd9dff, 0xecd0a2ff, 
    0xecd2a9ff, 0xeacfa0ff, 0xebcf9fff, 0xe9cb99ff, 0xe6c896ff, 0xdaba83ff, 
    0xcaa668ff, 0xa5783fff, 0x986520ff, 0x8e5a0cff, 0x82540dff, 0x6b3a04ff, 
    0x511802ff, 0x551a02ff, 0x602502ff, 0x6e3303ff, 0x7c3f03ff, 0x894604ff, 
    0x985304ff, 0xaa690cff, 0xb57517ff, 0xbc7b1aff, 0xbd7e17ff, 0xc99035ff, 
    0xdbae63ff, 0xe1b874ff, 0xd2a252ff, 0xbb7f28ff, 0xaa6813ff, 0xa45b0aff, 
    0xa15707ff, 0x9a4f04ff, 0x9e5304ff, 0xa35805ff, 0xa25704ff, 0xa96414ff, 
    0xd5a75fff, 0xe8c68eff, 0xe5be7aff, 0xdbab4eff, 0xe1b663ff, 0xe8c486ff, 
    0xe9c684ff, 0xe9c78dff, 0xefcf9eff, 0xf0d3a3ff, 0xeecf98ff, 0xeccb90ff, 
    0xf0d398ff, 0xf1d39aff, 0xf2d8a4ff, 0xf2ddb0ff, 0xf1dcadff, 0xf1d9a8ff, 
    0xf1d6a4ff, 0xedcf95ff, 0xebc988ff, 0xedcd90ff, 0xeccd8fff, 0xeed195ff, 
    0xedcb89ff, 0xecc881ff, 0xeac577ff, 0xe7be65ff, 0xebc471ff, 0xeece88ff, 
    0xd8ac58ff, 0xab630eff, 0xa55603ff, 0xa75b05ff, 0xa65905ff, 0xa65a04ff, 
    0xa55c05ff, 0xa66107ff, 0xb16f08ff, 0xba7c13ff, 0xb57610ff, 0xaa6305ff, 
    0xa85e04ff, 0xa45b04ff, 0x9f5504ff, 0x9d5304ff, 0x9f5404ff, 0x975004ff, 
    0x8b5204ff, 0x8e5604ff, 0x905804ff, 0x935a03ff, 0xad751cff, 0xc69b48ff, 
    0xd3b06cff, 0xd8b87bff, 0xd6b474ff, 0xd7b473ff, 0xd5b271ff, 0xd4b16cff, 
    0xd6b165ff, 0xd6b060ff, 0xd4ae5bff, 0xd3af5aff, 0xd4af5fff, 0xd3af60ff, 
    0xd2ac5bff, 0xd0a851ff, 0xcea64dff, 0xcfa751ff, 0xcca446ff, 0xcfa64cff, 
    0xd2ad5bff, 0xd1aa58ff, 0xcba249ff, 0xcda54bff, 0xd2ad5aff, 0xd5b36dff, 
    0xd5b46eff, 0xd8b56dff, 0xeac788ff, 0xefd097ff, 0xefd29fff, 0xf1d5a4ff, 
    0xf1d8adff, 0xf1d8b0ff, 0xf0d6a9ff, 0xf0d4abff, 0xf0d6acff, 0xefd3a9ff, 
    0xedd1a6ff, 0xecce9dff, 0xefd3a5ff, 0xecd0a0ff, 0xedd3a4ff, 0xedd4a9ff, 
    0xeacf9dff, 0xe4c58cff, 0xe2c48fff, 0xdcbf90ff, 0xd2b284ff, 0xb68b49ff, 
    0x9b6411ff, 0x8e5106ff, 0x732f03ff, 0x691e03ff, 0x752a03ff, 0x823704ff, 
    0x8d4504ff, 0x974c05ff, 0xa45904ff, 0xaf6b0bff, 0xb77517ff, 0xb97712ff, 
    0xbc7b16ff, 0xc07e1dff, 0xca8e33ff, 0xd29946ff, 0xcf9433ff, 0xcc922bff, 
    0xcd9637ff, 0xc78c30ff, 0xbd8231ff, 0xac660aff, 0xa85e06ff, 0xa55a04ff, 
    0xa55d06ff, 0xc79345ff, 0xe7c38aff, 0xedcc9bff, 0xe8c489ff, 0xdcac50ff, 
    0xe3ba6eff, 0xeac791ff, 0xecca96ff, 0xecc995ff, 0xeecd9cff, 0xedcb99ff, 
    0xeac792ff, 0xebc994ff, 0xefd098ff, 0xeece96ff, 0xefd29eff, 0xf0d7a8ff, 
    0xf2d9abff, 0xf0d19bff, 0xf1d4a3ff, 0xeed19aff, 0xeaca8aff, 0xebcb8fff, 
    0xebcb8dff, 0xeecf95ff, 0xefd19cff, 0xeecf97ff, 0xedcd94ff, 0xeac780ff, 
    0xe8c068ff, 0xeac46eff, 0xebc879ff, 0xd6a95cff, 0xae6a1fff, 0x9e4e03ff, 
    0x9e4e04ff, 0xa25506ff, 0xb37112ff, 0xbf841cff, 0xbe821bff, 0xb97b16ff, 
    0xae6806ff, 0xa96204ff, 0xa65d05ff, 0xa35804ff, 0x9e5004ff, 0x9a4d04ff, 
    0x9d5104ff, 0x9e5204ff, 0x914b04ff, 0x8c5304ff, 0x8f5704ff, 0x915903ff, 
    0xa16711ff, 0xc19545ff, 0xceab61ff, 0xd6b478ff, 0xd6b377ff, 0xd4b270ff, 
    0xd4b16fff, 0xd5b069ff, 0xd3ae5fff, 0xd4af60ff, 0xd5b063ff, 0xd4b061ff, 
    0xd4b167ff, 0xd3af62ff, 0xd1a959ff, 0xd0a753ff, 0xd0a956ff, 0xd1ac5bff, 
    0xd2ab58ff, 0xd3ac58ff, 0xd2ac5aff, 0xd1ab59ff, 0xcda44cff, 0xcea651ff, 
    0xd0aa59ff, 0xd3af62ff, 0xd5b167ff, 0xd8b468ff, 0xebc98bff, 0xf0d19eff, 
    0xefd4a8ff, 0xf1d4a7ff, 0xf2d8afff, 0xf1d9b0ff, 0xf0d4a6ff, 0xf1d6a6ff, 
    0xf2daafff, 0xefd3a3ff, 0xf0d3a5ff, 0xf0d2a0ff, 0xefd19fff, 0xefd5a8ff, 
    0xefd8b1ff, 0xeed4adff, 0xeed3a6ff, 0xedd3a3ff, 0xecd3a5ff, 0xe8d0a4ff, 
    0xe6cfa6ff, 0xdec396ff, 0xc9a15fff, 0xa26829ff, 0x772903ff, 0x772603ff, 
    0x853204ff, 0x8b3a04ff, 0x944304ff, 0x9c4b04ff, 0xa45305ff, 0xab6105ff, 
    0xb16b0bff, 0xb6700aff, 0xbc7a16ff, 0xbb750cff, 0xbc740aff, 0xb76509ff, 
    0xbb6a13ff, 0xc98a27ff, 0xcd9431ff, 0xca902fff, 0xc38424ff, 0xbc7810ff, 
    0xb6710bff, 0xa75e03ff, 0xb97926ff, 0xe2bc7cff, 0xedce9fff, 0xeecfa0ff, 
    0xe9c791ff, 0xdcac55ff, 0xdeae59ff, 0xe7c082ff, 0xe9c58eff, 0xeac892ff, 
    0xeccc99ff, 0xeccc9dff, 0xedcd9dff, 0xeccc9bff, 0xeecf9eff, 0xeccd9cff, 
    0xebce9bff, 0xedd09eff, 0xf0d4a5ff, 0xefd098ff, 0xf0d29eff, 0xefd4a2ff, 
    0xeccb91ff, 0xeccc91ff, 0xeac98aff, 0xeac98aff, 0xedcd96ff, 0xeccc90ff, 
    0xe9c88dff, 0xeac88eff, 0xe8c784ff, 0xeac887ff, 0xedce8eff, 0xf1d497ff, 
    0xe6c685ff, 0xb2722eff, 0xa25609ff, 0xbf8224ff, 0xcb9633ff, 0xc89031ff, 
    0xbb7c16ff, 0xb06906ff, 0xac6404ff, 0xa86204ff, 0xa35904ff, 0xa05204ff, 
    0xa05504ff, 0x9c4f04ff, 0x9c4e04ff, 0x9d4f04ff, 0x994d04ff, 0x904f04ff, 
    0x8e5504ff, 0x915803ff, 0xac7420ff, 0xcca55eff, 0xcea95aff, 0xcea959ff, 
    0xd1ad65ff, 0xd1ad65ff, 0xd0a958ff, 0xd1aa5bff, 0xd3af63ff, 0xd4b26aff, 
    0xd4b064ff, 0xd2ad5bff, 0xd3af63ff, 0xd3af62ff, 0xd1ab56ff, 0xd3ab56ff, 
    0xd0a955ff, 0xd1a955ff, 0xd3ac5bff, 0xd2ad5dff, 0xd4af61ff, 0xd3af64ff, 
    0xd1ab58ff, 0xd1ab5bff, 0xd3ad62ff, 0xd4b065ff, 0xd5b167ff, 0xd6b168ff, 
    0xefd097ff, 0xeed0a1ff, 0xecc896ff, 0xeac48bff, 0xefcfa0ff, 0xf2d8abff, 
    0xf2d5a3ff, 0xf4ddb5ff, 0xf3dab2ff, 0xf2daaeff, 0xf4ddb0ff, 0xf3daa9ff, 
    0xf1d8acff, 0xf0d5a5ff, 0xf0d6a8ff, 0xefd4a6ff, 0xefd5a7ff, 0xf1dbb2ff, 
    0xefd6abff, 0xeed5abff, 0xedd8b0ff, 0xebd4a9ff, 0xe5c692ff, 0xd3ac6bff, 
    0xb68242ff, 0x8e4514ff, 0x863103ff, 0x8d3804ff, 0x964504ff, 0x9b4604ff, 
    0xa04e05ff, 0xa85a04ff, 0xae6304ff, 0xb36a05ff, 0xb9730eff, 0xba720bff, 
    0xba7006ff, 0xba6c06ff, 0xb35a04ff, 0xaa4c06ff, 0xbe6f12ff, 0xc6831dff, 
    0xc27f16ff, 0xbd7b11ff, 0xb67105ff, 0xab6308ff, 0xd19d50ff, 0xebca9bff, 
    0xefd0a8ff, 0xedcd9bff, 0xe3bd7aff, 0xd8a74cff, 0xdbad5aff, 0xe0b66bff, 
    0xe6c180ff, 0xeac791ff, 0xeac897ff, 0xeac997ff, 0xeac997ff, 0xebcb9aff, 
    0xedce9eff, 0xeccc9aff, 0xeccb97ff, 0xebcb96ff, 0xedcc99ff, 0xedce97ff, 
    0xefd099ff, 0xeecf9bff, 0xebc98eff, 0xe8c582ff, 0xe4c07fff, 0xe3c080ff, 
    0xe7c684ff, 0xe8c688ff, 0xebcb93ff, 0xebca94ff, 0xe6c381ff, 0xe9c789ff, 
    0xedcd93ff, 0xedcc8eff, 0xf2d69eff, 0xe5c185ff, 0xcd9741ff, 0xd39c3eff, 
    0xce993fff, 0xbe7f21ff, 0xb16906ff, 0xad6204ff, 0xac6204ff, 0xa85e05ff, 
    0xa25504ff, 0x9e5204ff, 0xa15604ff, 0xa15705ff, 0x9d5004ff, 0x9b4d04ff, 
    0x9e5504ff, 0xa36005ff, 0xa26204ff, 0xa56504ff, 0xba8428ff, 0xcfa95fff, 
    0xcda65bff, 0xd3af68ff, 0xd1ad5dff, 0xd2ac60ff, 0xd3ae65ff, 0xd3ac5eff, 
    0xd5b064ff, 0xd5b26bff, 0xd4b065ff, 0xd2ac58ff, 0xd1a952ff, 0xd2ab53ff, 
    0xd3ab50ff, 0xd5ad56ff, 0xd4ad5aff, 0xd3ac58ff, 0xd5ae5eff, 0xd3ae5fff, 
    0xd4af62ff, 0xd3ad61ff, 0xd1ad60ff, 0xd2ac5cff, 0xd4ae64ff, 0xd5b16bff, 
    0xd5b26cff, 0xd6b067ff, 0xf0d29eff, 0xf1d4a7ff, 0xedcd9bff, 0xeecb97ff, 
    0xf2d8aeff, 0xf1d7a9ff, 0xf2d8a7ff, 0xf3dcb3ff, 0xf1d59fff, 0xf0d28eff, 
    0xf3d89aff, 0xf3d9a2ff, 0xf0d49cff, 0xf3dbaaff, 0xefd299ff, 0xefd196ff, 
    0xf1daaeff, 0xf1dab4ff, 0xf0dab0ff, 0xf0d9adff, 0xf2deb9ff, 0xeed9afff, 
    0xebcf9dff, 0xe4c388ff, 0xe5c283ff, 0xbe8e52ff, 0x853406ff, 0x8b3402ff, 
    0x923b04ff, 0x984404ff, 0xa14f04ff, 0xaa5e05ff, 0xad6105ff, 0xb26705ff, 
    0xb86f07ff, 0xbb740bff, 0xba7006ff, 0xbc7207ff, 0xbb6b06ff, 0xb76304ff, 
    0xba6b04ff, 0xbf7507ff, 0xc17e11ff, 0xc08217ff, 0xba7608ff, 0xc3852bff, 
    0xe0b677ff, 0xeccca2ff, 0xeccea2ff, 0xe8c68dff, 0xe0b86dff, 0xd19e42ff, 
    0xd3a049ff, 0xd7a751ff, 0xdcae5aff, 0xe5c07fff, 0xe8c48bff, 0xe7c58dff, 
    0xe7c38eff, 0xe8c48fff, 0xe9c592ff, 0xe9c690ff, 0xeac892ff, 0xe8c58fff, 
    0xe9c790ff, 0xeac792ff, 0xe9c78aff, 0xeac88eff, 0xe6c17cff, 0xe5bb6dff, 
    0xe4bd76ff, 0xe5c181ff, 0xe6c27fff, 0xe7c485ff, 0xe9c68dff, 0xe7c489ff, 
    0xe6c481ff, 0xe9c684ff, 0xe7c380ff, 0xe7c178ff, 0xe9c377ff, 0xe7c282ff, 
    0xd6a24fff, 0xc98a2bff, 0xbe7a1cff, 0xb16707ff, 0xae6104ff, 0xab5c05ff, 
    0xa95905ff, 0xa55404ff, 0xa25504ff, 0xa05104ff, 0xa75804ff, 0xab6105ff, 
    0xad6805ff, 0xae6907ff, 0xb26f0aff, 0xb67a10ff, 0xb87a0fff, 0xba7c0fff, 
    0xbe8722ff, 0xd1a95fff, 0xd3af6fff, 0xd6b377ff, 0xd6b472ff, 0xd3ae63ff, 
    0xd5b26cff, 0xd7b470ff, 0xd7b56eff, 0xd7b46eff, 0xd7b46eff, 0xd6b164ff, 
    0xd4ad55ff, 0xd6ae57ff, 0xd5af5aff, 0xd6b05cff, 0xd9b56bff, 0xd8b36bff, 
    0xd9b369ff, 0xd6b265ff, 0xd6b268ff, 0xd6b370ff, 0xd4b16aff, 0xd2ad60ff, 
    0xd3ac5eff, 0xd3ad63ff, 0xd4b36dff, 0xd5b167ff, 0xf0d39fff, 0xf0d3a6ff, 
    0xf1d3a7ff, 0xf2d4a6ff, 0xf2d6a9ff, 0xf2d8a8ff, 0xf2d5a0ff, 0xf2d6a3ff, 
    0xf2d59aff, 0xf0d38fff, 0xf2d394ff, 0xf3d69eff, 0xf2d497ff, 0xf3daa2ff, 
    0xebcc8aff, 0xeed093ff, 0xf0d7a7ff, 0xf0d5a3ff, 0xf0d7a2ff, 0xf0daabff, 
    0xeed8b0ff, 0xefdab2ff, 0xedd4a3ff, 0xe9ca91ff, 0xeac991ff, 0xe7c78fff, 
    0xba884bff, 0x94480dff, 0x913c03ff, 0x9a4604ff, 0x9f4f05ff, 0xa55805ff, 
    0xab5c05ff, 0xb16505ff, 0xb8720aff, 0xbb7911ff, 0xba730aff, 0xba6e05ff, 
    0xbe7507ff, 0xbf7406ff, 0xc07c0dff, 0xbf7c0cff, 0xc07e10ff, 0xc2831aff, 
    0xbe7c10ff, 0xce973fff, 0xe3bd82ff, 0xe6c38eff, 0xe5c085ff, 0xe7c389ff, 
    0xe5c083ff, 0xd5a651ff, 0xd19f44ff, 0xd8a85aff, 0xdeb26cff, 0xe0b66aff, 
    0xe8c388ff, 0xeac794ff, 0xe8c593ff, 0xe5c088ff, 0xe5c189ff, 0xe8c48fff, 
    0xe5c287ff, 0xe6c387ff, 0xe7c48eff, 0xe8c38dff, 0xe6c180ff, 0xe7c07bff, 
    0xe3b767ff, 0xdeb059ff, 0xdfb25dff, 0xe4bc74ff, 0xe4be7aff, 0xe6c383ff, 
    0xe7c589ff, 0xe5c07eff, 0xe9c789ff, 0xe8c786ff, 0xe5c17fff, 0xe4bb6dff, 
    0xe4bb69ff, 0xe4bc78ff, 0xd9aa5bff, 0xbc771aff, 0xb06103ff, 0xae5b04ff, 
    0xac5a04ff, 0xaa5604ff, 0xa65004ff, 0xaa5904ff, 0xa65404ff, 0xa34f06ff, 
    0xba7717ff, 0xbe8220ff, 0xbf841cff, 0xc38927ff, 0xc48a28ff, 0xc1851fff, 
    0xbe8014ff, 0xbc7e12ff, 0xbd8218ff, 0xcea355ff, 0xd5b276ff, 0xd7b47bff, 
    0xdab882ff, 0xd6b26eff, 0xd4af62ff, 0xd8b775ff, 0xd8b774ff, 0xd9b66fff, 
    0xd9b772ff, 0xd9b76fff, 0xd9b569ff, 0xd7b364ff, 0xd8b369ff, 0xd8b66fff, 
    0xd9b774ff, 0xd9b773ff, 0xd9b773ff, 0xdbba78ff, 0xdab874ff, 0xd6b672ff, 
    0xd6b470ff, 0xd5b065ff, 0xd3ae60ff, 0xd3ae62ff, 0xd4b067ff, 0xd7b26cff, 
    0xefd09dff, 0xf1d4a9ff, 0xf1d5a7ff, 0xf2d6a9ff, 0xf3dbb3ff, 0xf1d7a8ff, 
    0xebc782ff, 0xeecb8dff, 0xf1d199ff, 0xf2d69eff, 0xf3d9a5ff, 0xf4dcafff, 
    0xf2d5a0ff, 0xf2d49dff, 0xebca89ff, 0xeccb8bff, 0xedcc89ff, 0xebca81ff, 
    0xebca82ff, 0xf0d49cff, 0xf0daadff, 0xf0daaeff, 0xf1dcaeff, 0xefd7a3ff, 
    0xefd6a3ff, 0xeed6a5ff, 0xeacd92ff, 0xcca05bff, 0x9e570fff, 0x994702ff, 
    0x9f4d04ff, 0xa45704ff, 0xaa5c04ff, 0xb06604ff, 0xb7720bff, 0xb8730dff, 
    0xb56406ff, 0xba7106ff, 0xbd790bff, 0xbf7a0dff, 0xc28015ff, 0xc4851cff, 
    0xc58920ff, 0xc3881bff, 0xbf7f0fff, 0xcf9a43ff, 0xe3bd81ff, 0xe6c38aff, 
    0xe8c690ff, 0xe6c48fff, 0xe1bb80ff, 0xd5a651ff, 0xd5a54dff, 0xe1b97aff, 
    0xe3bd83ff, 0xe0b469ff, 0xe6c083ff, 0xe2ba7eff, 0xdcae6bff, 0xe1b77cff, 
    0xe7c38fff, 0xe7c48dff, 0xe5be84ff, 0xe4c086ff, 0xe6c28bff, 0xe7c389ff, 
    0xe6c383ff, 0xe5bd75ff, 0xe1b25eff, 0xe2b664ff, 0xe5bd73ff, 0xe7c17eff, 
    0xe4c17dff, 0xe9c68aff, 0xe9c68cff, 0xe6c282ff, 0xe6c381ff, 0xe9c68cff, 
    0xe9c68eff, 0xe3ba71ff, 0xe6c07fff, 0xe7c38bff, 0xe4bd78ff, 0xd49e49ff, 
    0xb76d0eff, 0xb16104ff, 0xb56807ff, 0xb87012ff, 0xb1620dff, 0xb86b08ff, 
    0xb86905ff, 0xb86c0eff, 0xca8c36ff, 0xcb9240ff, 0xc68d2fff, 0xc88e2fff, 
    0xc38725ff, 0xc4872cff, 0xba7a12ff, 0xb16d04ff, 0xb26f06ff, 0xc28d32ff, 
    0xd6b071ff, 0xd9b780ff, 0xdab884ff, 0xd6b170ff, 0xd3ab58ff, 0xd7b268ff, 
    0xdbb978ff, 0xdbb874ff, 0xdab874ff, 0xdbb875ff, 0xdaba77ff, 0xdab977ff, 
    0xd9b977ff, 0xdcbb80ff, 0xdcbb80ff, 0xd8b574ff, 0xd9b777ff, 0xdebe84ff, 
    0xddbd7fff, 0xd9b878ff, 0xd8b570ff, 0xd6b16aff, 0xd5b16cff, 0xd4af64ff, 
    0xd4ae63ff, 0xd6b16bff, 0xebc889ff, 0xf0d09cff, 0xf1d3a5ff, 0xf1d8adff, 
    0xf2d7acff, 0xefd298ff, 0xedc982ff, 0xf1d39aff, 0xf3d9a9ff, 0xf2d6a2ff, 
    0xf4dcb0ff, 0xf3deb6ff, 0xf2d9aaff, 0xf4daa8ff, 0xefd296ff, 0xedcc89ff, 
    0xedce89ff, 0xeed18dff, 0xedd395ff, 0xeccf8fff, 0xefd59fff, 0xf2dcb0ff, 
    0xf2deb4ff, 0xefd9abff, 0xf0d8a6ff, 0xefd9abff, 0xefd6a7ff, 0xebce9aff, 
    0xd1a868ff, 0xa8651fff, 0xa35503ff, 0xa55704ff, 0xa95805ff, 0xb06305ff, 
    0xb56e09ff, 0xb36306ff, 0xb66505ff, 0xbb770eff, 0xc07e1fff, 0xc38529ff, 
    0xc48827ff, 0xc78c29ff, 0xc78e27ff, 0xc78c23ff, 0xc68a22ff, 0xddb06dff, 
    0xe9c591ff, 0xeecd9fff, 0xe9c89aff, 0xdfb97eff, 0xd6a859ff, 0xd4a651ff, 
    0xd9ae64ff, 0xe0b97aff, 0xe1b977ff, 0xdfb468ff, 0xe5bf86ff, 0xddb474ff, 
    0xcc9433ff, 0xd4a452ff, 0xe6c18cff, 0xe7c38dff, 0xe7c18cff, 0xe7c28aff, 
    0xe7c488ff, 0xe7c48bff, 0xe9c68dff, 0xe3b970ff, 0xdead53ff, 0xdfb05aff, 
    0xe4ba6dff, 0xe6be75ff, 0xe6c07cff, 0xe6be77ff, 0xe5bc75ff, 0xe5be7cff, 
    0xe5c17fff, 0xe6c17fff, 0xe7c588ff, 0xe3bd79ff, 0xe5be7dff, 0xeac691ff, 
    0xe8c386ff, 0xe4ba6cff, 0xd0983bff, 0xbc7106ff, 0xc37c11ff, 0xca8a2bff, 
    0xc88620ff, 0xc88421ff, 0xc68322ff, 0xc58223ff, 0xcd9040ff, 0xc88a35ff, 
    0xc6862dff, 0xc28225ff, 0xbd7b17ff, 0xba7813ff, 0xb36e07ff, 0xb06904ff, 
    0xb26d03ff, 0xbc801aff, 0xd1a75aff, 0xd8b579ff, 0xd8b47aff, 0xd6b26dff, 
    0xd9b66fff, 0xd8b466ff, 0xddbc7bff, 0xe0bf87ff, 0xdbbc7cff, 0xdebf82ff, 
    0xdebe83ff, 0xdebe85ff, 0xddbd84ff, 0xdcbb82ff, 0xdbba7eff, 0xddbb81ff, 
    0xdbbb80ff, 0xe0c18cff, 0xdfc085ff, 0xdbba7cff, 0xd8b56fff, 0xd6b167ff, 
    0xd6b26cff, 0xd5b16aff, 0xd4ae66ff, 0xd6b068ff, 0xeac57cff, 0xecc989ff, 
    0xf0d5a4ff, 0xf0d9acff, 0xf0d29dff, 0xf2d396ff, 0xf3d699ff, 0xf5dba7ff, 
    0xf4ddafff, 0xf3daaaff, 0xf2d8a8ff, 0xf3dcb2ff, 0xf4ddb8ff, 0xf4dfb8ff, 
    0xf3dfb1ff, 0xf0d59aff, 0xf0d396ff, 0xf2daa6ff, 0xf1dbadff, 0xe8c782ff, 
    0xe8c778ff, 0xeed49dff, 0xefd5a3ff, 0xeed8a6ff, 0xeed6a4ff, 0xefd6a6ff, 
    0xedd4a8ff, 0xedd4a6ff, 0xeed5aaff, 0xdbb883ff, 0xb27322ff, 0xaa5f03ff, 
    0xac6305ff, 0xae6105ff, 0xb16305ff, 0xb15d04ff, 0xb9730bff, 0xc18428ff, 
    0xc78f40ff, 0xc89345ff, 0xc99342ff, 0xca933cff, 0xc78d26ff, 0xc68d20ff, 
    0xca9030ff, 0xe0b87eff, 0xeac997ff, 0xebcb9cff, 0xe8c697ff, 0xdab16dff, 
    0xd6aa5bff, 0xe0ba7eff, 0xe1bb7fff, 0xdeb979ff, 0xdeb874ff, 0xdfb66cff, 
    0xdeb46bff, 0xe1b97aff, 0xdaaf65ff, 0xddb36aff, 0xe4bf83ff, 0xe3bd80ff, 
    0xe3bc7cff, 0xe1b871ff, 0xdeb162ff, 0xe2b872ff, 0xe5bd7cff, 0xe1b362ff, 
    0xddad55ff, 0xdeb05bff, 0xe2b565ff, 0xe2b563ff, 0xe1b562ff, 0xdfb35bff, 
    0xe0b562ff, 0xe4ba70ff, 0xe6c080ff, 0xe5bf7eff, 0xe5c382ff, 0xe6c383ff, 
    0xe5c07cff, 0xe5c07eff, 0xe6bc73ff, 0xe6ba6bff, 0xddac55ff, 0xca8b26ff, 
    0xc5831cff, 0xc88523ff, 0xc6821bff, 0xc78627ff, 0xcb8f39ff, 0xc8882eff, 
    0xc7832eff, 0xca8a3bff, 0xc4832eff, 0xc07b20ff, 0xbb7715ff, 0xb7740cff, 
    0xb5740bff, 0xb56e05ff, 0xb67105ff, 0xb77711ff, 0xcda04eff, 0xd5ae67ff, 
    0xd5b06aff, 0xdbbb80ff, 0xe0c08bff, 0xdfbe85ff, 0xddbd7fff, 0xdebf86ff, 
    0xdfc087ff, 0xe0c290ff, 0xdfc28eff, 0xdebe86ff, 0xdcbc83ff, 0xdbba80ff, 
    0xdbba80ff, 0xdcba81ff, 0xdab880ff, 0xdcbc84ff, 0xdcbc7cff, 0xdbba77ff, 
    0xd7b46aff, 0xd7b366ff, 0xd5b26aff, 0xd5b16aff, 0xd6b270ff, 0xd6b26dff, 
    0xefd08aff, 0xeed093ff, 0xf0d6a3ff, 0xefd39bff, 0xf0d394ff, 0xf4d9a0ff, 
    0xf5dba5ff, 0xf7e2b3ff, 0xf4dba7ff, 0xf2d9a4ff, 0xf0d8a5ff, 0xf0d8a8ff, 
    0xf0d5abff, 0xf0d9abff, 0xf0daa9ff, 0xf0d9a2ff, 0xeed295ff, 0xf0d7a2ff, 
    0xf0daa9ff, 0xeacd8dff, 0xe6c675ff, 0xe9cd86ff, 0xf0daa7ff, 0xedd6a5ff, 
    0xebce98ff, 0xedd39fff, 0xefd6a5ff, 0xf0daafff, 0xeed7aeff, 0xeed3a2ff, 
    0xdfbb7bff, 0xba7e28ff, 0xae6907ff, 0xae6505ff, 0xaf6204ff, 0xb26604ff, 
    0xba7912ff, 0xc58b37ff, 0xca974fff, 0xca994fff, 0xcb9a4dff, 0xcb9644ff, 
    0xc78e2eff, 0xc58b20ff, 0xc88f2dff, 0xe1ba7dff, 0xebcc9cff, 0xe5c48fff, 
    0xdfb97eff, 0xd7ab5fff, 0xe0ba82ff, 0xe5c291ff, 0xe5c38dff, 0xe0bd7eff, 
    0xe2c083ff, 0xe3c186ff, 0xdcb56eff, 0xd8ac5cff, 0xddb369ff, 0xdbae5fff, 
    0xdfb56cff, 0xdfb56bff, 0xdbad57ff, 0xd9aa54ff, 0xdbad5cff, 0xdeb467ff, 
    0xe1b971ff, 0xdfb05dff, 0xdcac53ff, 0xdfb25cff, 0xe3b96bff, 0xe5bc73ff, 
    0xe4bb70ff, 0xe0b45fff, 0xe0b15aff, 0xdfaf54ff, 0xe4b967ff, 0xe6c07dff, 
    0xe8c68cff, 0xe7c484ff, 0xe6c17eff, 0xe5bf7aff, 0xe4b96cff, 0xe5bb70ff, 
    0xe5b96dff, 0xddaf5aff, 0xc98b2bff, 0xc68019ff, 0xc78524ff, 0xc8882fff, 
    0xc88929ff, 0xc5821fff, 0xc07813ff, 0xc47e25ff, 0xbb7215ff, 0xb76d09ff, 
    0xb9750fff, 0xbd8021ff, 0xbb7e1aff, 0xb67407ff, 0xb47005ff, 0xb07009ff, 
    0xc89b45ff, 0xd4af69ff, 0xd6b26cff, 0xddbc83ff, 0xe1c08eff, 0xe1c290ff, 
    0xe0c18dff, 0xdebe84ff, 0xe0c18bff, 0xe0c291ff, 0xdec18dff, 0xddbe85ff, 
    0xdebe86ff, 0xdcbf82ff, 0xdcbd82ff, 0xdbbb83ff, 0xdec08eff, 0xdec38eff, 
    0xdcbc7eff, 0xdbb978ff, 0xdab875ff, 0xd7b36aff, 0xd5b26bff, 0xd4b36fff, 
    0xd5b26dff, 0xd5b16dff, 0xf2d79eff, 0xefd396ff, 0xeecf8fff, 0xefd290ff, 
    0xf3d99fff, 0xf3dca7ff, 0xf4dda8ff, 0xf5dfaeff, 0xf2d9a3ff, 0xf2d59cff, 
    0xf0d7a2ff, 0xefd7a2ff, 0xf0d6a3ff, 0xeed6a0ff, 0xefd8a6ff, 0xeed7a0ff, 
    0xebce8cff, 0xeed49aff, 0xf0dbabff, 0xefd79dff, 0xebce82ff, 0xedd38cff, 
    0xf0dcabff, 0xecd6a6ff, 0xedd5a0ff, 0xecd29aff, 0xf0d9a8ff, 0xeedaadff, 
    0xebd2a1ff, 0xefd7a3ff, 0xefd8a4ff, 0xdebd7aff, 0xb57a1dff, 0xb26f0bff, 
    0xb5700aff, 0xb77309ff, 0xbb7d18ff, 0xc1862cff, 0xc79446ff, 0xca994eff, 
    0xcc9d51ff, 0xcb9847ff, 0xc89134ff, 0xc3891fff, 0xc58b21ff, 0xddb573ff, 
    0xe9ca9aff, 0xe3c188ff, 0xd9b16bff, 0xd7ad62ff, 0xdeba81ff, 0xe1bd86ff, 
    0xdeb975ff, 0xe1bf7fff, 0xe4c48cff, 0xe1bf82ff, 0xd8b165ff, 0xd0a346ff, 
    0xcfa042ff, 0xd0a044ff, 0xd8ab58ff, 0xddb366ff, 0xdaae5cff, 0xd8ab54ff, 
    0xdfb46aff, 0xe1b975ff, 0xe2bb77ff, 0xe1b76fff, 0xddae56ff, 0xe1b665ff, 
    0xe4bb71ff, 0xe3bb70ff, 0xe5bb6eff, 0xe1b561ff, 0xdfb25cff, 0xdeb156ff, 
    0xe5bd70ff, 0xe6c280ff, 0xe9c68eff, 0xe7c384ff, 0xe6c17cff, 0xe6be77ff, 
    0xe5bb6bff, 0xe5b96cff, 0xe3b666ff, 0xe1b462ff, 0xcc8e30ff, 0xc57f18ff, 
    0xc88520ff, 0xcd8f36ff, 0xc98b2bff, 0xcc943cff, 0xc98e34ff, 0xc58629ff, 
    0xbb740eff, 0xb46c04ff, 0xbc7b19ff, 0xc18528ff, 0xb97c15ff, 0xb37206ff, 
    0xb17208ff, 0xaf6f06ff, 0xc49135ff, 0xd5ad67ff, 0xd9b677ff, 0xdebe86ff, 
    0xe0c18fff, 0xe0c493ff, 0xe0c696ff, 0xe0c289ff, 0xe2c58fff, 0xe2c798ff, 
    0xdfc391ff, 0xddc086ff, 0xddbd85ff, 0xdbbd7eff, 0xdcbe80ff, 0xdebe87ff, 
    0xddbf89ff, 0xdfc18bff, 0xdcbc7eff, 0xdab876ff, 0xd9b876ff, 0xd7b56eff, 
    0xd7b571ff, 0xd7b675ff, 0xd3b06aff, 0xd3ac63ff, 
};
